/* some debug functions
*/
/*
 * MIPS floating point support
 * Copyright (C) 1994-2000 Algorithmics Ltd.  All rights reserved.
 * http://www.algor.co.uk
 *
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 */

/**************************************************************************
 *
 *  December 14, 1999
 *  Modified to build and operate in Linux kernel environment. 
 *
 *  Kevin D. Kissell, kevink@acm.org
 *  Copyright (C) 1999 MIPS Technologies, Inc. All rights reserved.
 *
 *************************************************************************/

#ifdef __KERNEL__
#define printf printk
#else
#include <stdio.h>
#include <stdarg.h>
#endif

#include "ieee754.h"

#define DP_EBIAS	1023
#define DP_EMIN		(-1022)
#define DP_EMAX		1023
#define DP_FBITS	52

#define SP_EBIAS	127
#define SP_EMIN		(-126)
#define SP_EMAX		127
#define SP_FBITS	23

#define DP_MBIT(x)	((unsigned long long)1 << (x))
#define DP_HIDDEN_BIT	DP_MBIT(DP_FBITS)
#define DP_SIGN_BIT	DP_MBIT(63)


#define SP_MBIT(x)	((unsigned long)1 << (x))
#define SP_HIDDEN_BIT	SP_MBIT(SP_FBITS)
#define SP_SIGN_BIT	SP_MBIT(31)


#define SPSIGN(sp)	(sp.parts.sign)
#define SPBEXP(sp)	(sp.parts.bexp)
#define SPMANT(sp)	(sp.parts.mant)

#define DPSIGN(dp)	(dp.parts.sign)
#define DPBEXP(dp)	(dp.parts.bexp)
#define DPMANT(dp)	(dp.parts.mant)

ieee754dp ieee754dp_dump(char *m,ieee754dp x)
{
    int i;

    printf("%s",m);
    printf("<%08x,%08x>\n", (unsigned)(x.bits >> 32), (unsigned)x.bits);
    printf("\t=");
    switch(ieee754dp_class(x)) {
    case IEEE754_CLASS_QNAN:
    case IEEE754_CLASS_SNAN:
	printf("Nan %c", DPSIGN(x)?'-':'+');
	for(i=DP_FBITS-1;i>=0;i--)
	  printf("%c", DPMANT(x) & DP_MBIT(i)?'1':'0');
	break;
    case IEEE754_CLASS_INF:
	printf("%cInfinity", DPSIGN(x)?'-':'+');
	break;
    case IEEE754_CLASS_ZERO:
	printf("%cZero", DPSIGN(x)?'-':'+');
	break;
    case IEEE754_CLASS_DNORM:
	printf("%c0.",  DPSIGN(x)?'-':'+');
	for(i=DP_FBITS-1;i>=0;i--)
	  printf("%c", DPMANT(x) & DP_MBIT(i)?'1':'0');
	printf("e%d",DPBEXP(x) - DP_EBIAS);
	break;
    case IEEE754_CLASS_NORM:
	printf("%c1.",  DPSIGN(x)?'-':'+');
	for(i=DP_FBITS-1;i>=0;i--)
	  printf("%c", DPMANT(x) & DP_MBIT(i)?'1':'0');
	printf("e%d", DPBEXP(x) - DP_EBIAS);
	break;
    default:
	printf("Illegal/Unknown IEEE754 value class");
    }
    printf("\n");
    return x;
}

ieee754sp ieee754sp_dump(char *m,ieee754sp x)
{
    int i;

    printf("%s=",m);
    printf("<%08x>\n", (unsigned)x.bits);
    printf("\t=");
    switch(ieee754sp_class(x)) {
    case IEEE754_CLASS_QNAN:
    case IEEE754_CLASS_SNAN:
	printf("Nan %c", SPSIGN(x)?'-':'+');
	for(i=SP_FBITS-1;i>=0;i--)
	  printf("%c", SPMANT(x) & SP_MBIT(i)?'1':'0');
	break;
    case IEEE754_CLASS_INF:
	printf("%cInfinity", SPSIGN(x)?'-':'+');
	break;
    case IEEE754_CLASS_ZERO:
	printf("%cZero", SPSIGN(x)?'-':'+');
	break;
    case IEEE754_CLASS_DNORM:
	printf("%c0.",  SPSIGN(x)?'-':'+');
	for(i=SP_FBITS-1;i>=0;i--)
	  printf("%c", SPMANT(x) & SP_MBIT(i)?'1':'0');
	printf("e%d",SPBEXP(x) - SP_EBIAS);
	break;
    case IEEE754_CLASS_NORM:
	printf("%c1.",  SPSIGN(x)?'-':'+');
	for(i=SP_FBITS-1;i>=0;i--)
	  printf("%c", SPMANT(x) & SP_MBIT(i)?'1':'0');
	printf("e%d",SPBEXP(x) - SP_EBIAS);
	break;
    default:
	printf("Illegal/Unknown IEEE754 value class");
    }
    printf("\n");
    return x;
}
