/*
 * tx39.c: TX39 core specific MMU/cache code.
 *
 * Copyright (C) 2000 Yoichi Yuasa <yuasa@users.sourceforge.net>
 * Copyright (C) 2001 AXE, Inc.  <fuku@axe-inc.co.jp>
 *
 * This code is based on r2300.c (R2000 and R3000 specific mmu/cache code) by:
 * Copyright (C) 1996 David S. Miller (dm@engr.sgi.com)
 * Copyright (C) 1998, 2000 Harald Koerfgen
 * Copyright (C) 1998 Gleb Raiko & Vladimir Roganov
 * 
 */

#include <linux/init.h>
#include <linux/mm.h>

#include <asm/asm.h>
#include <asm/bootinfo.h>
#include <asm/cacheops.h>
#include <asm/pgtable.h>
#include <asm/mmu_context.h>
#include <asm/io.h>
#include <asm/wbflush.h>

unsigned long mips_tlb_entries;

static unsigned long dcache_size, dcache_lsize;
static unsigned long icache_size, icache_lsize;
static unsigned long scache_size, scache_lsize; /* Not used */

#include <asm/r4kcache.h>

#undef DEBUG_TLB
#undef DEBUG_CACHE
#undef ENABLE_WRITE_BACK
#undef DISABLE_ICACHE
#undef DISABLE_DCACHE

/*
 * Cache oprations.
 */
static void tx39_clear_page_i16(void *page)
{
	__asm__ __volatile__(
		".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		"addiu\t$1,%0,%2\n"
		"1:\tsw\t$0,(%0)\n\t"
		"sw\t$0,4(%0)\n\t"
		"sw\t$0,8(%0)\n\t"
		"sw\t$0,12(%0)\n\t"
		"addiu\t%0,32\n\t"
		"sw\t$0,-16(%0)\n\t"
		"sw\t$0,-12(%0)\n\t"
		"sw\t$0,-8(%0)\n\t"
		"bne\t$1,%0,1b\n\t"
		"sw\t$0,-4(%0)\n\t"
		".set\tat\n\t"
		".set\treorder"
		:"=r" (page)
		:"0" (page),
		 "I" (PAGE_SIZE)
		:"$1","memory");
}

static void tx39_clear_page_d16i16(void *page)
{
	__asm__ __volatile__(
		".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		".set\tmips3\n\t"
		"addiu\t$1,%0,%2\n"
		"1:\tcache\t%3,(%0)\n\t"
		"sw\t$0,(%0)\n\t"
		"sw\t$0,4(%0)\n\t"
		"sw\t$0,8(%0)\n\t"
		"sw\t$0,12(%0)\n\t"
		"addiu\t%0,32\n\t"
		"cache\t%3,-16(%0)\n\t"
		"sw\t$0,-16(%0)\n\t"
		"sw\t$0,-12(%0)\n\t"
		"sw\t$0,-8(%0)\n\t"
		"bne\t$1,%0,1b\n\t"
		"sw\t$0,-4(%0)\n\t"
		".set\tmips0\n\t"
		".set\tat\n\t"
		".set\treorder"
		:"=r" (page)
		:"0" (page),
		 "I" (PAGE_SIZE),
		 "i" (Hit_Writeback_Inv_D)
		:"$1","memory");
}

static void tx39_copy_page_i16(void *to, void *from)
{
	unsigned long dummy1, dummy2;
	unsigned long reg1, reg2, reg3, reg4;

	__asm__ __volatile__(
		".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		"addiu\t$1,%0,%8\n"
		"1:\tlw\t%2,(%1)\n\t"
		"lw\t%3,4(%1)\n\t"
		"lw\t%4,8(%1)\n\t"
		"lw\t%5,12(%1)\n\t"
		"sw\t%2,(%0)\n\t"
		"sw\t%3,4(%0)\n\t"
		"sw\t%4,8(%0)\n\t"
		"sw\t%5,12(%0)\n\t"
		"lw\t%2,16(%1)\n\t"
		"lw\t%3,20(%1)\n\t"
		"lw\t%4,24(%1)\n\t"
		"lw\t%5,28(%1)\n\t"
		"sw\t%2,16(%0)\n\t"
		"sw\t%3,20(%0)\n\t"
		"sw\t%4,24(%0)\n\t"
		"sw\t%5,28(%0)\n\t"
		"addiu\t%0,64\n\t"
		"addiu\t%1,64\n\t"
		"lw\t%2,-32(%1)\n\t"
		"lw\t%3,-28(%1)\n\t"
		"lw\t%4,-24(%1)\n\t"
		"lw\t%5,-20(%1)\n\t"
		"sw\t%2,-32(%0)\n\t"
		"sw\t%3,-28(%0)\n\t"
		"sw\t%4,-24(%0)\n\t"
		"sw\t%5,-20(%0)\n\t"
		"lw\t%2,-16(%1)\n\t"
		"lw\t%3,-12(%1)\n\t"
		"lw\t%4,-8(%1)\n\t"
		"lw\t%5,-4(%1)\n\t"
		"sw\t%2,-16(%0)\n\t"
		"sw\t%3,-12(%0)\n\t"
		"sw\t%4,-8(%0)\n\t"
		"bne\t$1,%0,1b\n\t"
		"sw\t%5,-4(%0)\n\t"
		".set\tat\n\t"
		".set\treorder"
		:"=r" (dummy1), "=r" (dummy2),
		 "=&r" (reg1), "=&r" (reg2), "=&r" (reg3), "=&r" (reg4)
		:"0" (to), "1" (from),
		 "I" (PAGE_SIZE));
}

static void tx39_copy_page_d16i16(void *to, void *from)
{
	unsigned long dummy1, dummy2;
	unsigned long reg1, reg2, reg3, reg4;

	__asm__ __volatile__(
		".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		".set\tmips3\n\t"
		"addiu\t$1,%0,%8\n"
		"1:\tcache\t%9,(%0)\n\t"
		"lw\t%2,(%1)\n\t"
		"lw\t%3,4(%1)\n\t"
		"lw\t%4,8(%1)\n\t"
		"lw\t%5,12(%1)\n\t"
		"sw\t%2,(%0)\n\t"
		"sw\t%3,4(%0)\n\t"
		"sw\t%4,8(%0)\n\t"
		"sw\t%5,12(%0)\n\t"
		"cache\t%9,16(%0)\n\t"
		"lw\t%2,16(%1)\n\t"
		"lw\t%3,20(%1)\n\t"
		"lw\t%4,24(%1)\n\t"
		"lw\t%5,28(%1)\n\t"
		"sw\t%2,16(%0)\n\t"
		"sw\t%3,20(%0)\n\t"
		"sw\t%4,24(%0)\n\t"
		"sw\t%5,28(%0)\n\t"
		"cache\t%9,32(%0)\n\t"
		"addiu\t%0,64\n\t"
		"addiu\t%1,64\n\t"
		"lw\t%2,-32(%1)\n\t"
		"lw\t%3,-28(%1)\n\t"
		"lw\t%4,-24(%1)\n\t"
		"lw\t%5,-20(%1)\n\t"
		"sw\t%2,-32(%0)\n\t"
		"sw\t%3,-28(%0)\n\t"
		"sw\t%4,-24(%0)\n\t"
		"sw\t%5,-20(%0)\n\t"
		"cache\t%9,-16(%0)\n\t"
		"lw\t%2,-16(%1)\n\t"
		"lw\t%3,-12(%1)\n\t"
		"lw\t%4,-8(%1)\n\t"
		"lw\t%5,-4(%1)\n\t"
		"sw\t%2,-16(%0)\n\t"
		"sw\t%3,-12(%0)\n\t"
		"sw\t%4,-8(%0)\n\t"
		"bne\t$1,%0,1b\n\t"
		"sw\t%5,-4(%0)\n\t"
		".set\tmips0\n\t"
		".set\tat\n\t"
		".set\treorder"
		:"=r" (dummy1), "=r" (dummy2),
		 "=&r" (reg1), "=&r" (reg2), "=&r" (reg3), "=&r" (reg4)
		:"0" (to), "1" (from),
		 "I" (PAGE_SIZE),
		 "i" (Hit_Writeback_Inv_D));
}

static void __init probe_dcache(u32 config)
{
	dcache_size = 1 << (10 + ((config >> 16) & 7));

	switch (mips_cputype) {
	case CPU_TX3920:
		dcache_lsize = 16;
		break;
	}

	printk("Primary data cache %ldkb, linesize %ld bytes\n",
		dcache_size >> 10, dcache_lsize);
}

static void __init probe_icache(u32 config)
{
	icache_size = 1 << (10 + ((config >> 19) & 7));

	switch (mips_cputype) {
	case CPU_TX3920:
		icache_lsize = 16;
		break;
	}

	printk("Primary instruction cache %ldkb, linesize %ld bytes\n",
		icache_size >> 10, icache_lsize);
}

#define cache16_2way_unroll16(base,op)				\
	__asm__ __volatile__("					\
		.set noreorder;					\
		.set mips3;					\
		cache %1, 0x000(%0); cache %1, 0x001(%0);	\
		cache %1, 0x010(%0); cache %1, 0x011(%0);	\
		cache %1, 0x020(%0); cache %1, 0x021(%0);	\
		cache %1, 0x030(%0); cache %1, 0x031(%0);	\
		cache %1, 0x040(%0); cache %1, 0x041(%0);	\
		cache %1, 0x050(%0); cache %1, 0x051(%0);	\
		cache %1, 0x060(%0); cache %1, 0x061(%0);	\
		cache %1, 0x070(%0); cache %1, 0x071(%0);	\
		cache %1, 0x080(%0); cache %1, 0x081(%0);	\
		cache %1, 0x090(%0); cache %1, 0x091(%0);	\
		cache %1, 0x0a0(%0); cache %1, 0x0a1(%0);	\
		cache %1, 0x0b0(%0); cache %1, 0x0b1(%0);	\
		cache %1, 0x0c0(%0); cache %1, 0x0c1(%0);	\
		cache %1, 0x0d0(%0); cache %1, 0x0d1(%0);	\
		cache %1, 0x0e0(%0); cache %1, 0x0e1(%0);	\
		cache %1, 0x0f0(%0); cache %1, 0x0f1(%0);	\
		.set mips0;					\
		.set reorder"					\
		:						\
		: "r" (base),					\
		  "i" (op));

#define load_tag_2way(base,op,way0,way1)			\
	__asm__ __volatile__("					\
		.set noreorder;					\
		.set mips3;					\
		cache %3, 0x000(%2);				\
		mfc0 %0, $20;					\
		cache %3, 0x001(%2);				\
		mfc0 %1, $20;					\
		.set mips0;					\
		.set reorder"					\
		:"=r" (way0), "=r" (way1)			\
		: "r" (base),					\
		  "i" (op));

static inline void tx39_blast_dcache16(void)
{
	unsigned long start = KSEG0;
	unsigned long end = (start + dcache_size / 2);	/* 2-way cache */

	while(start < end) {
		cache16_2way_unroll16(start,Index_Writeback_Inv_D);
		start += 0x100;
	}
}

static inline void tx39_blast_dcache16_page(unsigned long page)
{
	unsigned long start = page;
	unsigned long end = (start + dcache_size / 2);	/* 2-way cache */

	while (start < end) {
		cache16_2way_unroll16(start,Index_Writeback_Inv_D);
		start += 0x100;
	}
}

static inline void tx39_blast_dcache16_page_indexed(unsigned long page)
{
	unsigned long start = page;
	unsigned long end = (start + PAGE_SIZE);

	while(start < end) {
		cache16_2way_unroll16(start,Index_Writeback_Inv_D);
		start += 0x100;
	}
}

static inline void tx39_blast_icache16(void)
{
	unsigned long start = KSEG0;
	unsigned long end = (start + icache_size / 2);
	unsigned long dummy = 0, config;
	unsigned long flags;

	save_and_cli(flags);

	/* disable I cache and stop streaming */
	__asm__ __volatile__(
		".set	noreorder\n\t"
		"mfc0	%0, $3\n\t"
		"li	%1, ~0x00000020\n\t"
		"and	%1, %1, %0\n\t"
		"mtc0	%1, $3\n\t"
		"j	1f\n\t"
		" nop\n\t"
		"1: .set reorder"
		: "=r" (config) : "r" (dummy));

	/* flush I cache */
	while (start < end) {
		cache16_unroll32(start, Index_Invalidate_I);
		start += 0x200;
	}

	/* restore I cache enable bit*/
	__asm__ __volatile__(
		".set	noreorder\n\t"
		"mtc0	%0, $3\n\t"
		".set	reorder"
		: : "r" (config));
	restore_flags(flags);
}

static void tx39_clear_lock_icache16(void)
{
	unsigned long start = KSEG0;
	unsigned long end = (start + icache_size / 2);
	unsigned long dummy = 0, config;
	unsigned long flags;

	save_and_cli(flags);

	/* disable I cache and stop streaming */
	__asm__ __volatile__(
		".set	noreorder\n\t"
		"mfc0	%0, $3\n\t"
		"li	%1, ~0x00000020\n\t"
		"and	%1, %1, %0\n\t"
		"mtc0	%1, $3\n\t"
		"j	1f\n\t"
		" nop\n\t"
		"1: .set reorder"
		: "=r" (config) : "r" (dummy));

	while(start < end) {
		cache16_unroll32(start, Index_LockBit_Clear_I);
		start += 0x200;
	}

	/* restore I cache enable bit*/
	__asm__ __volatile__(
		".set	noreorder\n\t"
		"mtc0	%0, $3\n\t"
		".set	reorder"
		: : "r" (config));
	restore_flags(flags);
}

static void tx39_clear_lock_dcache16(void)
{
	unsigned long start = KSEG0;
	unsigned long end = (start + dcache_size / 2);

	while(start < end) {
		cache16_unroll32(start, Index_LockBit_Clear_D);
		start += 0x200;
	}
}

static inline void tx39_disable_icache(void)
{
	unsigned long dummy = 0, config;
	unsigned long flags;

	save_and_cli(flags);

	/* disable I cache and stop streaming */
	__asm__ __volatile__(
		".set	noreorder\n\t"
		"mfc0	%0, $3\n\t"
		"li	%1, ~0x00000020\n\t"
		"and	%1, %1, %0\n\t"
		"mtc0	%1, $3\n\t"
		"j	1f\n\t"
		" nop\n\t"
		"1: .set reorder"
		: "=r" (config) : "r" (dummy));
	restore_flags(flags);
}

static inline void tx39_load_tag_icache16(u32 *buf, int line)
{
	unsigned long start = KSEG0;
	unsigned long end = (start + line * 16);
	unsigned long dummy = 0, config;
	int i;
	unsigned long flags;

	save_and_cli(flags);
	/* disable I cache and stop streaming */
	__asm__ __volatile__(
		".set	noreorder\n\t"
		"mfc0	%0, $3\n\t"
		"li	%1, ~0x00000020\n\t"
		"and	%1, %1, %0\n\t"	/* disable I-cache */
		"mtc0	%1, $3\n\t"
		"j	1f\n\t"
		" nop\n\t"
		"1: .set reorder"
		: "=r" (config) : "r" (dummy));

	i = 0;
	while (start < end) {
		load_tag_2way(start, Index_Load_Tag_I, buf[i*2], buf[i*2+1]);
		start += 16;
		i++;
	}

	/* restore I-cache enable bit*/
	__asm__ __volatile__(
		".set	noreorder\n\t"
		"mtc0	%0, $3\n\t"
		".set	reorder"
		: : "r" (config));
	restore_flags(flags);
}

static inline void tx39_load_tag_dcache16(u32 *buf, int line)
{
	unsigned long start = KSEG0;
	unsigned long end = (start + line * 16);
	int i;

	i = 0;
	while (start < end) {
		load_tag_2way(start, Index_Load_Tag_D, buf[i*2], buf[i*2+1]);
		start += 16;
		i++;
	}
}

#define DEBUG_CACHE_SIZE_MAX	(16384 * 2)
static u32 tag_buf[DEBUG_CACHE_SIZE_MAX/16];

void tx39_print_icache_tag(void)
{
    u32 *bufp;
    int i, line;

    line = icache_size / 16 / 2;
    bufp = (u32 *)KSEG1ADDR((u32)tag_buf);
    tx39_load_tag_icache16(bufp, line);
    printk("icache tag:\n");
    for(i=0; i<line; i++){
	if( i%4 == 0 )
	    printk("\ntag %04x: ", i * 16);
	printk(" %08x:%08x", bufp[i*2], bufp[i*2+1]);
    }
    printk("\n");
}

void tx39_print_dcache_tag(void)
{
    u32 *bufp;
    int i, line;

    line = icache_size / 16 / 2;
    bufp = (u32 *)KSEG1ADDR((u32)tag_buf);
    tx39_load_tag_dcache16(bufp, line);
    printk("dcache tag:\n");
    for(i=0; i<line; i++){
	if( i%4 == 0 )
	    printk("\ntag %04x: ", i * 16);
	printk(" %08x:%08x", bufp[i*2], bufp[i*2+1]);
    }
    printk("\n");
}

static inline void tx39_flush_cache_all_i16(void)
{
	tx39_blast_icache16();
}

static inline void tx39_flush_cache_all_d16i16(void)
{
	tx39_blast_dcache16();
	tx39_blast_icache16();
}

static void tx39_flush_cache_range_i16(struct mm_struct *mm,
                                       unsigned long start,
                                       unsigned long end)
{
	if (mm->context != 0) {
#ifdef DEBUG_CACHE
		printk("crange[%d,%08lx,%08lx]", (int)mm->context, start, end);
#endif
		tx39_blast_icache16();
	}
}

static void tx39_flush_cache_range_d16i16(struct mm_struct *mm,
                                            unsigned long start,
                                            unsigned long end)
{
	if (mm->context != 0) {
#ifdef DEBUG_CACHE
		printk("crange[%d,%08lx,%08lx]", (int)mm->context, start, end);
#endif
		tx39_blast_dcache16();
		tx39_blast_icache16();
	}
}

static void tx39_flush_cache_mm_i16(struct mm_struct *mm)
{
	if (mm->context != 0) {

#ifdef DEBUG_CACHE
		printk("cmm[%d]", (int)mm->context);
#endif
		tx39_flush_cache_all_i16();
	}
}

static void tx39_flush_cache_mm_d16i16(struct mm_struct *mm)
{
	if (mm->context != 0) {

#ifdef DEBUG_CACHE
		printk("cmm[%d]", (int)mm->context);
#endif
		tx39_flush_cache_all_d16i16();
	}
}

static inline unsigned long get_phys_page (unsigned long addr,
                                           struct mm_struct *mm)
{
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;
	unsigned long physpage;

	addr &= PAGE_MASK;
	pgd = pgd_offset(mm, addr);
	pmd = pmd_offset(pgd, addr);
	pte = pte_offset(pmd, addr);

	if((physpage = pte_val(*pte)) & _PAGE_VALID)
		return KSEG0ADDR(physpage);
	else
		return 0;
}

static void tx39_flush_cache_page_i16(struct vm_area_struct *vma,
                                      unsigned long page)
{
	tx39_flush_cache_all_i16();
}

static void tx39_flush_cache_page_d16i16(struct vm_area_struct *vma,
                                         unsigned long page)
{
	struct mm_struct *mm = vma->vm_mm;
	unsigned long physpage;

	if (mm->context == 0)
		return;

#ifdef DEBUG_CACHE
	printk("cpage[%d,%08lx]", (int)mm->context, page);
#endif

	physpage = get_phys_page(page, mm);
	if (physpage) {
		if (mm == current->active_mm) {
			tx39_blast_dcache16_page(page);
		}
		else {
			page = KSEG0 + (page & (dcache_size - 1));
			tx39_blast_dcache16_page_indexed(page);
		}
	}
}

static void tx39_flush_page_to_ram_i16(struct page *page)
{
	/*
	 * Nothing to be done
	 */
}

static void tx39_flush_page_to_ram_d16i16(struct page *page)
{
	tx39_blast_dcache16_page((unsigned long)page_address(page));
}

static void tx39_flush_cache_sigtramp_i16(unsigned long addr)
{
	tx39_flush_cache_all_i16();
}

static void tx39_flush_cache_sigtramp_d16i16(unsigned long addr)
{
	unsigned long daddr, iaddr; 

#ifdef DEBUG_CACHE
	printk("csigtramp[%08lx]", addr);
#endif

	daddr = addr & ~(dcache_lsize - 1);
	protected_writeback_dcache_line(daddr);
	protected_writeback_dcache_line(daddr + dcache_lsize);
	iaddr = addr & ~(icache_lsize - 1);
	protected_flush_icache_line(iaddr);
	protected_flush_icache_line(iaddr + icache_lsize);
}

static void tx39_dma_cache_wback_inv_i16(unsigned long start,
                                         unsigned long size)
{
	/*
	 * Nothing to be done
	 */
}

static void tx39_dma_cache_wback_inv_d16i16(unsigned long start,
                                            unsigned long size)
{
	unsigned long end, addr;

	if (size >= dcache_size) {
		flush_cache_all();
	}
	else {
		addr = start & ~(dcache_lsize - 1);
                end = (start + size) & ~(dcache_lsize - 1);
                while (1) {
                        flush_dcache_line(addr); /* Hit_Writeback_Inv_D */
                        if (addr == end) break;
                        addr += dcache_lsize;
                }
	}
}

/*
 *	TLB operations.
 */
void flush_tlb_all(void)
{
	unsigned long flags;
	unsigned long old_ctx;
	int entry;

#ifdef DEBUG_TLB
	printk("[tlball]");
#endif

	save_and_cli(flags);
	old_ctx = (get_entryhi() & 0xfc0);
	write_32bit_cp0_register(CP0_ENTRYLO0, 0);
	for(entry = 8; entry < mips_tlb_entries; entry++) {
		write_32bit_cp0_register(CP0_INDEX, entry << 8);
		write_32bit_cp0_register(CP0_ENTRYHI, ((entry | 0x80000) << 12));
		__asm__ __volatile__("tlbwi");
	}
	set_entryhi(old_ctx);
	restore_flags(flags);
}

void flush_tlb_mm(struct mm_struct *mm)
{
	if (mm->context != 0) {
		unsigned long flags;

#ifdef DEBUG_TLB
		printk("[tlbmm<%d>]", mm->context);
#endif
		save_and_cli(flags);
		get_new_mmu_context(mm, asid_cache);
		if (mm == current->active_mm)
			set_entryhi(mm->context & 0xfc0);
		restore_flags(flags);
	}
}

void flush_tlb_range(struct mm_struct *mm, unsigned long start,
				  unsigned long end)
{
	if (mm->context != 0) {
		unsigned long flags;
		int size;

#ifdef DEBUG_TLB
		printk("[tlbrange<%02x,%08lx,%08lx>]", (mm->context & 0xfc0),
		       start, end);
#endif
		save_and_cli(flags);
		size = (end - start + (PAGE_SIZE - 1)) >> PAGE_SHIFT;
		if(size <= mips_tlb_entries) {
			int oldpid = (get_entryhi() & 0xfc0);
			int newpid = (mm->context & 0xfc0);

			start &= PAGE_MASK;
			end += (PAGE_SIZE - 1);
			end &= PAGE_MASK;
			while(start < end) {
				int idx;

				set_entryhi(start | newpid);
				start += PAGE_SIZE;
				tlb_probe();
				idx = get_index();
				set_entrylo0(0);
				set_entryhi(KSEG0);
				if(idx < 0)
					continue;
				tlb_write_indexed();
			}
			set_entryhi(oldpid);
		} else {
			get_new_mmu_context(mm, asid_cache);
			if (mm == current->active_mm)
				set_entryhi(mm->context & 0xfc0);
		}
		restore_flags(flags);
	}
}

void flush_tlb_page(struct vm_area_struct *vma, unsigned long page)
{
	if(vma->vm_mm->context != 0) {
		unsigned long flags;
		int oldpid, newpid, idx;

#ifdef DEBUG_TLB
		printk("[tlbpage<%d,%08lx>]", vma->vm_mm->context, page);
#endif
		newpid = (vma->vm_mm->context & 0xfc0);
		page &= PAGE_MASK;
		save_and_cli(flags);
		oldpid = (get_entryhi() & 0xfc0);
		set_entryhi(page | newpid);
		tlb_probe();
		idx = get_index();
		set_entrylo0(0);
		set_entryhi(KSEG0);
		if(idx < 0)
			goto finish;
		tlb_write_indexed();

finish:
		set_entryhi(oldpid);
		restore_flags(flags);
	}
}

/*
 * Initialize new page directory with pointers to invalid ptes
 */
void pgd_init(unsigned long page)
{
	unsigned long dummy1, dummy2;

	/*
	 * The plain and boring version for the R3000.  No cache flushing
	 * stuff is implemented since the R3000 has physical caches.
	 */
	__asm__ __volatile__(
		".set\tnoreorder\n"
		"1:\tsw\t%2,(%0)\n\t"
		"sw\t%2,4(%0)\n\t"
		"sw\t%2,8(%0)\n\t"
		"sw\t%2,12(%0)\n\t"
		"sw\t%2,16(%0)\n\t"
		"sw\t%2,20(%0)\n\t"
		"sw\t%2,24(%0)\n\t"
		"sw\t%2,28(%0)\n\t"
		"subu\t%1,1\n\t"
		"bnez\t%1,1b\n\t"
		"addiu\t%0,32\n\t"
		".set\treorder"
		:"=r" (dummy1),
		 "=r" (dummy2)
		:"r" ((unsigned long) invalid_pte_table),
		 "0" (page),
		 "1" (USER_PTRS_PER_PGD/8));
}

void update_mmu_cache(struct vm_area_struct * vma,
                      unsigned long address, pte_t pte)
{
	unsigned long flags;
	pgd_t *pgdp;
	pmd_t *pmdp;
	pte_t *ptep;
	int idx, pid;

	/*
	 * Handle debugger faulting in for debugee.
	 */
	if (current->active_mm != vma->vm_mm)
		return;

	pid = get_entryhi() & 0xfc0;

#ifdef DEBUG_TLB
	if((pid != (vma->vm_mm->context & 0xfc0)) || (vma->vm_mm->context == 0)) {
		printk("update_mmu_cache: Wheee, bogus tlbpid mmpid=%d tlbpid=%d\n",
		       (int) (vma->vm_mm->context & 0xfc0), pid);
	}
#endif

	save_and_cli(flags);
	address &= PAGE_MASK;
	set_entryhi(address | (pid));
	pgdp = pgd_offset(vma->vm_mm, address);
	tlb_probe();
	pmdp = pmd_offset(pgdp, address);
	idx = get_index();
	ptep = pte_offset(pmdp, address);
	set_entrylo0(pte_val(*ptep));
	set_entryhi(address | (pid));
	if(idx < 0) {
		tlb_write_random();
#if 0
		printk("[MISS]");
#endif
	} else {
		tlb_write_indexed();
#if 0
		printk("[HIT]");
#endif
	}
#if 0
	if(!strcmp(current->comm, "args")) {
		printk("<");
		for(idx = 8; idx < mips_tlb_entries; idx++) {
			set_index(idx);
			tlb_read();
			address = get_entryhi();
			if((address & 0xfc0) != 0)
				printk("[%08lx]", address);
		}
		printk(">\n");
	}
#endif
	set_entryhi(pid);
	restore_flags(flags);
}

void show_regs(struct pt_regs * regs)
{
	/*
	 * Saved main processor registers
	 */
	printk("$0 : %08x %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
	       0, (unsigned long) regs->regs[1], (unsigned long) regs->regs[2],
	       (unsigned long) regs->regs[3], (unsigned long) regs->regs[4],
	       (unsigned long) regs->regs[5], (unsigned long) regs->regs[6],
	       (unsigned long) regs->regs[7]);
	printk("$8 : %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
	       (unsigned long) regs->regs[8], (unsigned long) regs->regs[9],
	       (unsigned long) regs->regs[10], (unsigned long) regs->regs[11],
               (unsigned long) regs->regs[12], (unsigned long) regs->regs[13],
	       (unsigned long) regs->regs[14], (unsigned long) regs->regs[15]);
	printk("$16: %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
	       (unsigned long) regs->regs[16], (unsigned long) regs->regs[17],
	       (unsigned long) regs->regs[18], (unsigned long) regs->regs[19],
               (unsigned long) regs->regs[20], (unsigned long) regs->regs[21],
	       (unsigned long) regs->regs[22], (unsigned long) regs->regs[23]);
	printk("$24: %08lx %08lx                   %08lx %08lx %08lx %08lx\n",
	       (unsigned long) regs->regs[24], (unsigned long) regs->regs[25],
	       (unsigned long) regs->regs[28], (unsigned long) regs->regs[29],
               (unsigned long) regs->regs[30], (unsigned long) regs->regs[31]);

	/*
	 * Saved cp0 registers
	 */
	printk("epc  : %08lx\nStatus: %08x\nCause : %08x\n",
	       (unsigned long) regs->cp0_epc, (unsigned int) regs->cp0_status,
	       (unsigned int) regs->cp0_cause);
}

static unsigned long wired = 0;

#if 1
static void __flush_tlb_all_entry(void)
{
	unsigned long flags;
	unsigned long old_ctx;
	int entry;

	save_and_cli(flags);
	old_ctx = (get_entryhi() & 0xfc0);
	write_32bit_cp0_register(CP0_ENTRYLO0, 0);
	for(entry = 0; entry < mips_tlb_entries; entry++) {
		write_32bit_cp0_register(CP0_INDEX, entry << 8);
		write_32bit_cp0_register(CP0_ENTRYHI, ((entry | 0x80000) << 12));
		__asm__ __volatile__("tlbwi");
	}
	set_entryhi(old_ctx);
	restore_flags(flags);
}
#endif

void add_wired_entry(unsigned long entrylo0, unsigned long entrylo1,
                     unsigned long entryhi, unsigned long pagemask)
{
	unsigned long flags;
	unsigned long old_ctx;
	
	if (wired < 8) {
		save_and_cli(flags);
		old_ctx = get_entryhi() & 0xfc0;
		set_entrylo0(entrylo0);
		set_entryhi(entryhi);
		set_index((wired << 8) & 0x700);
		wired++;
		tlb_write_indexed();
		set_entryhi(old_ctx);
	        flush_tlb_all();    
		restore_flags(flags);
	}
}

void __init ld_mmu_tx39(void)
{
	int dcache_writeback;
	u32 config;

	config= read_32bit_cp0_register(CP0_CONF);

	printk("CPU revision is: %08x\n", read_32bit_cp0_register(CP0_PRID));

	probe_icache(config);
	probe_dcache(config);

#ifdef ENABLE_WRITE_BACK
	if( !(config & (1<<CONFB_WBON)) ){
	    /* enable write-back mode under HC-AJ2 */
	    config |= (1<<CONFB_WBON) | (1<<CONFB_CWFON);
	    write_32bit_cp0_register(CP0_CONF, config);
	}
#else
	if( config & (1<<CONFB_WBON) ){
	    /* disable write-back mode under HC-VJ1C */
	    tx39_blast_dcache16();
	    config &= ~((1<<CONFB_WBON) | (1<<CONFB_CWFON));
	    write_32bit_cp0_register(CP0_CONF, config);
	}
#endif

#ifdef DISABLE_ICACHE
	if( config & (1<<CONFB_ICE) ){
	    config &= ~(1<<CONFB_ICE);
	    tx39_disable_icache();
	}
#else
	if( !(config & (1<<CONFB_ICE)) ){
	    tx39_blast_icache16();
	    config |= (1<<CONFB_ICE);
	    write_32bit_cp0_register(CP0_CONF, config);
	}
#endif

#ifdef DISABLE_DCACHE
	if( config & (1<<CONFB_DCE) ){
	    config &= ~(1<<CONFB_DCE);
	    write_32bit_cp0_register(CP0_CONF, config);
	}
#else
	if( !(config & (1<<CONFB_DCE)) ){
	    config |= (1<<CONFB_DCE);
	    write_32bit_cp0_register(CP0_CONF, config);
	}
#endif

	// tx39_print_icache_tag();
	// tx39_print_dcache_tag();

	dcache_writeback = read_32bit_cp0_register(CP0_CONF) & 0x2000 ? 1 : 0;

	switch (mips_cputype) {
	case CPU_TX3920:
		mips_tlb_entries = 64;
		break;
	}


	if (dcache_writeback) {
		switch (dcache_lsize) {
		case 16:
			_clear_page = tx39_clear_page_d16i16;
			_copy_page = tx39_copy_page_d16i16;
			_flush_cache_all = tx39_flush_cache_all_d16i16;
			_flush_cache_mm = tx39_flush_cache_mm_d16i16;
			_flush_cache_range = tx39_flush_cache_range_d16i16;
			_flush_cache_page = tx39_flush_cache_page_d16i16;
			_flush_cache_sigtramp = tx39_flush_cache_sigtramp_d16i16;
			_flush_page_to_ram = tx39_flush_page_to_ram_d16i16;

			_dma_cache_wback_inv = tx39_dma_cache_wback_inv_d16i16;
			break;
		}
	}
	else {
		_clear_page = tx39_clear_page_i16;
		_copy_page = tx39_copy_page_i16;
		_flush_cache_all = tx39_flush_cache_all_i16;
		_flush_cache_mm = tx39_flush_cache_mm_i16;
		_flush_cache_range = tx39_flush_cache_range_i16;
		_flush_cache_page = tx39_flush_cache_page_i16;
		_flush_cache_sigtramp = tx39_flush_cache_sigtramp_i16;
		_flush_page_to_ram = tx39_flush_page_to_ram_i16;

		_dma_cache_wback_inv = tx39_dma_cache_wback_inv_i16;
	}

	write_32bit_cp0_register(CP0_WIRED, 8)
	write_32bit_cp0_register(CP0_PAGEMASK, PM_4K);

	tx39_clear_lock_icache16(); // WindowsCE use lock bit. (maybe)
	flush_cache_all();

	__flush_tlb_all_entry();
	// flush_tlb_all();
}
