/*
 * $Id: int-handler.S,v 1.14 2000/08/26 21:16:19 pavel Exp $
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2000 Jim Pick <jim@jimpick.com>
 * Copyright (C) 1999 Harald Koerfgen
 */
	
#include <asm/asm.h>
#include <asm/regdef.h>
#include <asm/mipsregs.h>
#include <asm/stackframe.h>
#include <asm/r39xx.h>


		.data
	        .globl  HighPriVect
HighPriVect:
	        .word   spurious        # ?
	        .word   io_posnegint0   # IOPOSINT(0) or IONEGINT(0)
	        .word   spurious        # CHIDMACNTINT
	        .word   spurious        # TELDMACNTINT
	        .word   spurious        # SNDDMACNTINT
	        .word   spurious        # Reserved
	        .word   io_negint56     # IONEGINT(6) or IONEGINT(5)
	        .word   spurious        # Reserved
	        .word   io_posint56     # IOPOSINT(6) or IOPOSINT(5)
	        .word   spurious        # Reserved
	        .word   spurious        # UARTBRXINT
	        .word   uarta_rx        # UARTARXINT
	        .word   spurious        # Reserved
	        .word   periodic_timer  # PERINT
	        .word   spurious        # ALARMINT (handle?)
	        .word   spurious        # POSPWROKINT or NEGPWROKINT (handle?)
	

		.text
		.set	noreorder
		.align	5
		NESTED(r39xx_handle_int, PT_SIZE, ra)
		.set	noat
		SAVE_ALL
		CLI
		.set	at
	
		/*
		 * Get pending Interrupts
		 */
		mfc0	t0,CP0_CAUSE		# get pending interrupts

#if 0
	
		mfc0	t2,CP0_STATUS
		and	t0,t2			# isolate allowed ones

#endif

#ifdef CONFIG_VTECH_HELIO_EMULATOR
        /* The emulator doesn't support high priority interrupts. */
        /* OK, this sucks.  Synthesize irq0 if it's the non-priority periodic timer. */
                lui     t0, %hi(IntStatus5)
                lw      t1, %lo(IntStatus5)(t0)
                li t0, INT5_PERIODICINT
                and    t1, t0
                bnez    t1, handle_it
                li      a0,0
#endif

		andi	t2, t0, IE_IRQ4		# irq 4 (high priority)
		bne	t2, IE_IRQ4, low_priority
		nop


/*
 *  Ok, we've got a high priority interrupt (a.k.a. an external interrupt).
 *  Read Interrupt Status Register 6 to get vector.
 */

high_priority:
	
		lui	t0, %hi(IntStatus6)
		lw	t1, %lo(IntStatus6)(t0)
		andi	t1, INT6_INTVECT

		la	t2, HighPriVect
		addu	t1, t1, t2
		lw	t2, 0(t1)
		jr	t2
		nop

/* 
 *  Periodic timer interrupt - map to IRQ 0
 */
	
periodic_timer:
	
		j	handle_it
		li	a0, 0

/* 
 *  UARTA recieve interrupts - map to IRQ 3
 */

uarta_rx:

		j	handle_it
		li	a0, 3


/* IO Interrupts */

/* 
 *  Fired when button 1 on the Helio is pressed or released.   Map to IRQ 10.
 *  Read IO interrupts or status register to determine which button
 *  was pressed. 
 */

io_posnegint0:

		j	handle_it
		li	a0, 10

/* 
 *  Fired when one of three buttons on the back of the Helio are pressed or
 *  released.   Map to IRQ 11.  Read IO interrupts or status register to
 *  determine which button was pressed. 
 */

io_posint56:
	
		j	handle_it
		li	a0, 11

/*
 *  Not expected - map to IRQ 12
 */
	
io_negint56:

		j	handle_it
		li	a0, 12


/*
 * Ok, we got one of the possible 143 Interrupt Module interrupts, oh well.
 *
 * Instead of assigning a single number to each of them, Ill try to form
 * groups. For example: The Interrupt Status Register 2 contains
 * 20 valid status bits all related to the two serial ports on the R3912.
 * Use a single interrupt number and let the serial driver figure out what
 * has happened.
 */

low_priority:
/*
 *  SIB Subframe Interrupts - Map to IRQ 40
 */

		lui	t0, %hi(IntStatus1)
		lw	t1, %lo(IntStatus1)(t0)
#ifndef CONFIG_VTECH_HELIO_EMULATOR
        /* stuck on in emulator */
		li	t0, INT1_SIBSF0INT
		and	t2, t1, t0
		bnez	t2, handle_it
		li	a0, 40
#endif

/*
 * Mobilon IDE interrupts - Map to IRQ 1
 */

#ifdef CONFIG_SHARP_MOBILON
		lui	t0, %hi(IntStatus3)
		lw	t1, %lo(IntStatus3)(t0)
		andi	t1, 0x4
		bnez	t1, handle_it
		 li	a0, 1
#endif
#ifdef CONFIG_VTECH_HELIO_EMULATOR        
/*
 * Rx Serial port interrupts (including transmit) mapped to IRQ 3.  Again, high pri ints seem hosed.
 */
#define UARTA_RX_INTS (INT2_UARTARXINT | INT2_UARTARXOVERRUN | INT2_UARTAFRAMEINT | INT2_UARTABREAKINT)
#define UARTB_RX_INTS (INT2_UARTBRXINT | INT2_UARTBRXOVERRUN | INT2_UARTBFRAMEINT | INT2_UARTBBREAKINT)        
        
		lui	t0, %hi(IntStatus2)
		lw	t1, %lo(IntStatus2)(t0)
		li	t0, (UARTA_RX_INTS | UARTB_RX_INTS)
		and	t2, t1, t0
		bnez	t2, handle_it
		 li	a0, 3
#endif
                			
/*
 *  Unknown Interrupt - Map to IRQ 20
 */

	
		j	handle_it
		 li	a0, 20
	
spurious:	
		
		j	spurious_interrupt
		nop

	
handle_it:	jal	do_IRQ
		move	a1,sp
		j	ret_from_irq
		nop
		
		END(r39xx_handle_int)
