/*
 * Keyboard driver for philips velo -- gross hack based on wild guess
 *
 * Copyright (C) 2000 Pavel Machek <pavel@ucw.cz>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * This hack uses "Power" key in very special way.
 */

#include <linux/tty.h>
#include <linux/mm.h>
#include <linux/kbd_kern.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kbd_ll.h>
#include <linux/sysrq.h>

#include <asm/keyboard.h>
#include <asm/r39xx.h>
#include <asm/betty.h>
#include <asm/bootinfo.h>


void kbd_scan(void);
void kbd_kick(void);
static void kbd_timer_handler(unsigned long);

extern int console_loglevel;
extern int (*console_blank_hook)(int);

unsigned int orig_MFIOInput = 0;

void kbd_leds(unsigned char leds)
{
	return;
}

int kbd_setkeycode(unsigned int scancode, unsigned int keycode)
{
	return (scancode == keycode) ? 0 : -EINVAL;
}

int kbd_getkeycode(unsigned int scancode)
{
	return scancode;
}

int kbd_translate(unsigned char scancode, unsigned char *keycode,
	char raw_mode)
{
	*keycode = scancode;
	return 1;
}

char kbd_unexpected_up(unsigned char keycode)
{
	printk(__FUNCTION__": unexpected up, keycode 0x%x\n", keycode);
	return 0x80;
}

static int old_spircv = 0xfffffff, power_down = 0, current_mfio = 1, spi_command = 255, kbd_ready = 0;

static void handle(int key)
{
	extern int voltage_input;
	if (key == 0x5e)
		MFIOOutput ^= MFIO_PIN_BACKLIGHT;
	if (power_down) {
		power_down = 2;
		switch (key) {
#ifdef CONFIG_MAGIC_SYSRQ
		case 0x13:	handle_sysrq('1', NULL, NULL, NULL); break;
		case 0x3f:	handle_sysrq('9', NULL, NULL, NULL); break;
		case 0x2b:	handle_sysrq('b', NULL, NULL, NULL); break;
		case 0x24:	handle_sysrq('s', NULL, NULL, NULL); break;
		case 0x3d:	handle_sysrq('u', NULL, NULL, NULL); break;
		case 0x2d:	handle_sysrq('t', NULL, NULL, NULL); break;
		case 0x3b:	handle_sysrq('m', NULL, NULL, NULL); break;
		case 0x4d:	handle_sysrq('p', NULL, NULL, NULL); break;
		case 0x44:	handle_sysrq('k', NULL, NULL, NULL); break;
#endif
		case 0x21:	/* space */
			MFIOOutput ^= current_mfio;
			printk( "MFIOOutput now %08x, MFIOInput is %08x\n", MFIOOutput, MFIOInput );
			break;
		case 0x01:	/* control */
			current_mfio <<= 1;
			if (!current_mfio)
				current_mfio = 1;
			printk( "current_mfio = %08x\n", current_mfio );
			break;
		case 0x19:	/* alt */
			spi_command++;
			spi_command &= 255;
			printk( "Sending SPI command %d\n", spi_command );
			SPIData = spi_command;
			break;
		case 0x41:	/* left */
			voltage_input = BETTY_ADC_INPUT_AD0;
			printk( "Voltage:\n" );
			get_battery_voltage();
			break;
		case 0x49:	/* down */
			voltage_input = BETTY_ADC_INPUT_AD1;
			printk( "Voltage:\n" );
			get_battery_voltage();
			break;
		case 0x32:	/* right */
			voltage_input = BETTY_ADC_INPUT_AD2;
			printk( "Voltage:\n" );
			get_battery_voltage();
			break;
		case 0x4a:	/* up */
			voltage_input = BETTY_ADC_INPUT_AD3;
			printk( "Voltage:\n" );
			get_battery_voltage();
			break;
		case 0x29:	/* esc */
			printk("IntStatus1 = %8x\n", IntStatus1 );
			printk("IntStatus2 = %8x\n", IntStatus2 );
			printk("IntStatus3 = %8x\n", IntStatus3 );
			printk("IntStatus4 = %8x\n", IntStatus4 );
			printk("IntStatus5 = %8x\n", IntStatus5 );
			printk("IntStatus6 = %8x\n", IntStatus6 );
			printk("orig_MFIOInput = %8x\n", orig_MFIOInput );
			break;
		}
	}
}

static int counter = 0;

void kbd_scan(void)
{
	int new_spircv;
	if (!kbd_ready)
		return;
	new_spircv = SPIData & 0xff;
	if ((old_spircv != new_spircv) && (new_spircv != 0xff)) {
//	        printk( "SPIData changed: %x\n", new_spircv );
		if (power_down || (new_spircv == 0x5e))
			handle(new_spircv);
		else
			handle_scancode(new_spircv & 0x7f, !(new_spircv & 0x80));
	}
	old_spircv = new_spircv;
	SPIData = 0;
}


static void power_down_interrupt(int irq, void *dev_id,
					 struct pt_regs * regs)
{
	IntClear5 = INT5_POSONBUTINT;
	power_down = 1;
}

static void power_up_interrupt(int irq, void *dev_id,
					 struct pt_regs * regs)
{
	IntClear5 = INT5_NEGONBUTINT;
	if (power_down != 2) {	/* Someone pressed and released power without pressing any action key */
		kbd_kick();
	}
	power_down = 0;
}

static void mic_down_interrupt(int irq, void *dev_id,
					 struct pt_regs * regs)
{
	IntClear1 = ~0;
	IntClear2 = ~0;
	IntClear3 = ~0;
	IntClear4 = ~0;
	IntClear5 = ~0;
	IntClear6 = ~0;
	console_loglevel = 9;
	printk( "System is still alive\n" );
}

static void mic_up_interrupt(int irq, void *dev_id,
					 struct pt_regs * regs)
{
/*
	console_loglevel = 9;
	printk( "Microphone up\n" );
	IntClear3 = ~0;
	IntClear4 = ~0;
*/
}

static void kbd_down_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	IntClear3 = MFIO_PIN_KEYBOARD;
	kbd_scan();
}

static void kbd_up_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	IntClear4 = MFIO_PIN_KEYBOARD;
	kbd_scan();
}

void kbd_kick(void)
{
	ClockControl |= CLK_ENSPICLK;
	SPIControl |= 1;
	mdelay(100);
	MFIOOutput &= ~MFIO_PIN_RESET_MOTOROLA;
	mdelay(100);
	MFIOOutput |= MFIO_PIN_RESET_MOTOROLA;
	kbd_ready = 1;
}

static void velo_console_blank(int blank)
{
	if (blank) {
		MFIOOutput &= ~MFIO_PIN_BACKLIGHT;
		MFIOOutput |= MFIO_PIN_LCD_POWER;
	} else {
		MFIOOutput |= MFIO_PIN_BACKLIGHT;
		MFIOOutput &= ~MFIO_PIN_LCD_POWER;
	}
}

void __init kbd_init_hw(void)
{
	int i;

	printk( "Initializing hackish velo 1 keyboard driver...\n" );
	orig_MFIOInput = MFIOInput;
	kbd_kick();
	console_blank_hook = velo_console_blank;
	if ((i = request_irq(16, power_down_interrupt, SA_INTERRUPT,
			"power_down", NULL))) {
		printk(KERN_ERR "tpanel: Cannot allocate irq. (%d)\n", i);
	}
	if ((i = request_irq(17, power_up_interrupt, SA_INTERRUPT,
			"power_up", NULL))) {
		printk(KERN_ERR "tpanel: Cannot allocate irq. (%d)\n", i);
	}
	if ((i = request_irq(18, mic_down_interrupt, SA_INTERRUPT,
			"mic_down", NULL))) {
		printk(KERN_ERR "tpanel: Cannot allocate irq. (%d)\n", i);
	}
	if ((i = request_irq(19, mic_up_interrupt, SA_INTERRUPT,
			"mic_up", NULL))) {
		printk(KERN_ERR "tpanel: Cannot allocate irq. (%d)\n", i);
	}
#if 1
	if ((i = request_irq(21, kbd_down_interrupt, SA_INTERRUPT,
			"kbd_down", NULL))) {
		printk(KERN_ERR "kbd: Cannot allocate irq. (%d)\n", i);
	}
	if ((i = request_irq(22, kbd_up_interrupt, SA_INTERRUPT,
			"kbd_up", NULL))) {
		printk(KERN_ERR "kbd: Cannot allocate irq. (%d)\n", i);
	}
#else
	disable_irq(21);
	disable_irq(22);
#endif
}
