/*
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * IDE routines for Toshiba TMPRR392x processors.
 *
 * Copyright (C) 2000 by Yoichi Yuasa <yuasa@users.sourceforge.net>
 * Copyright (C) 1999 by Harald Koerfgen
 *
 * Massively based on ide-std.c, Copyright (C) 1998, 1999 by Ralf Baechle
 */

#include <linux/sched.h>
#include <linux/ide.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>
#include <asm/io.h>
#include <asm/ptrace.h>
#include <asm/hdreg.h>
#include <asm/kseg2.h>

static int tx3920_ide_default_irq(ide_ioreg_t base)
{
	return 10;
}

static ide_ioreg_t tx3920_ide_default_io_base(int index)
{
	switch (index) {
		case 0: return 0x1f0;
		case 1: return 0x170;
		default: return 0;
	}
}

static void tx3920_ide_init_hwif_ports (hw_regs_t *hw, ide_ioreg_t data_port,
                                     ide_ioreg_t ctrl_port, int *irq)
{
	ide_ioreg_t reg = data_port;
	int i;

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg;
		reg += 1;
	}
	if (ctrl_port) {
		hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	} else {
		hw->io_ports[IDE_CONTROL_OFFSET] = hw->io_ports[IDE_DATA_OFFSET] + 0x206;
	}
	if (irq != NULL)
		*irq = 0;
}

static int tx3920_ide_request_irq(unsigned int irq,
                                void (*handler)(int,void *, struct pt_regs *),
                                unsigned long flags, const char *device,
                                void *dev_id)
{
	return request_irq(irq, handler, flags, device, dev_id);
}			

static void tx3920_ide_free_irq(unsigned int irq, void *dev_id)
{
	free_irq(irq, dev_id);
}

static int tx3920_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
	return check_region(from, extent);
}

static void tx3920_ide_request_region(ide_ioreg_t from, unsigned int extent,
                                    const char *name)
{
	request_region(from, extent, name);
}

static void tx3920_ide_release_region(ide_ioreg_t from, unsigned int extent)
{
	release_region(from, extent);
}

struct ide_ops tx3920_ide_ops = {
	&tx3920_ide_default_irq,
	&tx3920_ide_default_io_base,
	&tx3920_ide_init_hwif_ports,
	&tx3920_ide_request_irq,
	&tx3920_ide_free_irq,
	&tx3920_ide_check_region,
	&tx3920_ide_request_region,
	&tx3920_ide_release_region
};

void tx3920ide_init(void)
{
	kseg2_page_alloc(0x6c700000, 0x100000, KSEG2_FLAGS_ALL,
	                 &mips_io_port_base);
}
