/*
 * Interrupt handler for Toshiba TMPR3922 and PLUM2 Interrupt Module.
 *
 * Copyright (C) 1992 Linus Torvalds
 * Copyright (C) 1999 Harald Koerfgen
 * Copyright (C) 2000 Yoichi Yuasa <yuasa@users.sorceforge.net>
 * Copyright (C) 2001 AXE, Inc     <fuku@axe-inc.co.jp>
 *
 */
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/malloc.h>
#include <linux/random.h>

#include <asm/bitops.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/tx3920.h>
#include <asm/plum2.h>

#define DEBUG_IRQ
#undef LAZY_READ_STATUS

extern asmlinkage void tx3920_handle_int(void);

unsigned long spurious_count = 0;

irq_cpustat_t irq_stat [NR_CPUS];

/* NR_IRQS = 64 on MIPS */
#if NR_IRQS != 64
#error "NR_IRQS != 64"
#endif

/*
 * Per-IRQ information, by not initializing, this gets filled with NULLs:
 */
typedef struct {
	struct irqaction* irq_action;	// info on low-level handler
	int depth;			// < 0: enabled, >= 0: disabled
} irq_info_t;

static irq_info_t irq_info[NR_IRQS];

static inline void mask_irq_with_plum2_clear(unsigned int irq_nr)
{
    switch (irq_nr) {
    case 0:  /* Periodic Timer Interrupt */ 
	//	IntEnable5 &= ~INT5_PERINT;
	break;

    case 1:  /* Keyboard Interrupt */
	break;

    case 3:  /* Serial port A interrupt */
	break;

    case 4:  /* Serial port B interrupt */
	break;

    case 5:
	IntEnable5 &= ~INT5_PWROKINT;
	break;

    case 6:  /* alarm interrupt */
	break;

#ifdef CONFIG_SHARP_TELIOS
#ifdef CONFIG_SOUND_TELIOS
#define IRQ_CHI1_0INT 18
#define IRQ_CHI0_5INT 17

    case IRQ_CHI1_0INT:	/* SOUND DMA */
	IntEnable1 &= ~INT1_CHI1_0INT;
	break;
    case IRQ_CHI0_5INT:	/* SOUND DMA */
	IntEnable1 &= ~INT1_CHI0_5INT;
	break;
#endif
    case 19:	/* SIB start of subframe 0 Interrupt */
	break;

    case 20:	/* SIB IRQ */
	break;

    case 21:	/* SPI IRQ */
	break;

    case 29:	/* Optpoint */
	IntEnable4 &= ~(INT4_MFIONEGINT(9));
	break;
    case 30:	/* Button  */
	IntEnable8 &= ~(INT8_IOPOSINT8 | INT8_IOPOSINT7);
	break;
    case 31:	/* DC JACK IN */
	IntEnable8 &= ~(INT8_IOPOSINT6 | INT8_IONEGINT6);
	break;
#endif

#ifdef CONFIG_PLUM2
    case 32:
	Plum2_IntEnable = 0;
	IntClear8 = INT8_IOPOSINT5;
	/*	Plum2_IntEnable = PLUM2_INT_ENABLE; */
	break;

    case 33:
	Plum2_IntEnable = 0;
	Plum2_PccIntEnable &= ~PLUM2_PCC1_INT_ENABLE;
	Plum2_PccIntClear = PLUM2_PCC1_INT_CLEAR;
	IntClear8 = INT8_IOPOSINT5;
	Plum2_IntEnable = PLUM2_INT_ENABLE;
	break;

    case 34:
	Plum2_IntEnable = 0;
	Plum2_PccIntEnable &= ~PLUM2_PCC2_INT_ENABLE;
	Plum2_PccIntClear = PLUM2_PCC2_INT_CLEAR;
	IntClear8 = INT8_IOPOSINT5;
	Plum2_IntEnable = PLUM2_INT_ENABLE;
	break;
#endif

    default:
    }
}

static inline void mask_irq(unsigned int irq_nr)
{
    switch (irq_nr) {
    case 0:  /* Periodic Timer Interrupt */ 
	//	IntEnable5 &= ~INT5_PERINT;
	break;

    case 1:  /* Keyboard Interrupt */
	break;

    case 3:  /* Serial port A interrupt */
	break;

    case 4:  /* Serial port B interrupt */
	break;

    case 5:
	IntEnable5 &= ~INT5_PWROKINT;
	break;

    case 6:  /* alarm interrupt */
	break;

#ifdef CONFIG_SHARP_TELIOS
#ifdef CONFIG_SOUND_TELIOS
#define IRQ_CHI1_0INT 18
#define IRQ_CHI0_5INT 17

    case IRQ_CHI1_0INT:	/* SOUND DMA */
	IntEnable1 &= ~INT1_CHI1_0INT;
	break;
    case IRQ_CHI0_5INT:	/* SOUND DMA */
	IntEnable1 &= ~INT1_CHI0_5INT;
	break;
#endif
    case 19:	/* SIB start of subframe 0 Interrupt */
	break;

    case 20:	/* SIB IRQ */
	break;

    case 21:	/* SPI IRQ */
	break;

    case 29:	/* Optpoint */
	IntEnable4 &= ~(INT4_MFIONEGINT(9));
	break;
    case 30:	/* Button  */
	IntEnable8 &= ~(INT8_IOPOSINT8 | INT8_IOPOSINT7);
	break;
    case 31:	/* DC JACK IN */
	IntEnable8 &= ~(INT8_IOPOSINT6 | INT8_IONEGINT6);
	break;
#endif

#ifdef CONFIG_PLUM2
    case 32:
//	IntEnable8 &= ~INT8_IOPOSINT5;
//	Plum2_IntEnable = 0;
//	IntClear8 = INT8_IOPOSINT5;
//	Plum2_IntEnable = PLUM2_INT_ENABLE;
	break;

    case 33:
//	IntEnable8 &= ~INT8_IOPOSINT5;
//	Plum2_IntEnable = 0;
	Plum2_PccIntEnable &= ~PLUM2_PCC1_INT_ENABLE;
	//	Plum2_PccIntClear = PLUM2_PCC1_INT_CLEAR;
	//	IntClear8 = INT8_IOPOSINT5;
	//	Plum2_IntEnable = PLUM2_INT_ENABLE;
	break;

    case 34:
//	IntEnable8 &= ~INT8_IOPOSINT5;
//	Plum2_IntEnable = 0;
	Plum2_PccIntEnable &= ~PLUM2_PCC2_INT_ENABLE;
	//	Plum2_PccIntClear = PLUM2_PCC2_INT_CLEAR;
	//	IntClear8 = INT8_IOPOSINT5;
	//	Plum2_IntEnable = PLUM2_INT_ENABLE;
	break;
#endif

    default:
    }
}

static inline void unmask_irq(unsigned int irq_nr)
{
    switch (irq_nr) {
    case 0:
	//	IntEnable5 |= INT5_PERINT;
	break;

    case 1:  /* Keyboard Interrupt */
	break;

    case 3:  /* Serial port A interrupt */
	break;

    case 4:  /* Serial port B interrupt */
	break;

    case 5:
	IntEnable5 |= INT5_PWROKINT;
	break;

    case 6:  /* alarm interrupt */
	break;

    case 19:	/* SIB start of subframe 0 Interrupt */
	break;

    case 20:	/* SIBIRQ */
	break;

#ifdef CONFIG_SHARP_TELIOS
#ifdef CONFIG_SOUND_TELIOS
    case IRQ_CHI1_0INT:	/* SOUND DMA */
	IntEnable1 |= INT1_CHI1_0INT;
	break;
    case IRQ_CHI0_5INT:	/* SOUND DMA */
	IntEnable1 &= ~INT1_CHI0_5INT;
	break;
#endif
    case 29:	/* Optpoint */
	IntEnable4 |= INT4_MFIONEGINT(9);
	break;
    case 30:	/* Button */
	IntEnable8 |= INT8_IOPOSINT8 | INT8_IOPOSINT7;
	break;
    case 31:	/* DC JACK IN */
	IntEnable8 |= INT8_IOPOSINT6 | INT8_IONEGINT6;
	break;
#endif

#ifdef CONFIG_PLUM2
    case 32:
//	IntEnable8 |= INT8_IOPOSINT5;
//	Plum2_IntEnable = PLUM2_INT_ENABLE;
	Plum2_IntEnable = 0;
	IntClear8 = INT8_IOPOSINT5;
	Plum2_IntEnable = PLUM2_INT_ENABLE;
 	break;

    case 33:
//	IntEnable8 |= INT8_IOPOSINT5;
	Plum2_IntEnable = 0;
 	Plum2_PccIntEnable |= PLUM2_PCC1_INT_ENABLE;
	Plum2_IntEnable = PLUM2_INT_ENABLE;
	break;

    case 34:
//	IntEnable8 |= INT8_IOPOSINT5;
	Plum2_IntEnable = 0;
 	Plum2_PccIntEnable |= PLUM2_PCC2_INT_ENABLE;
	Plum2_IntEnable = PLUM2_INT_ENABLE;
	break;
#endif

    default:
    }
}

void disable_irq(unsigned int irq_nr)
{
    unsigned long flags;

    save_and_cli(flags);
    if (++irq_info[irq_nr].depth >= 0)
	mask_irq(irq_nr);
    restore_flags(flags);
}

void enable_irq(unsigned int irq_nr)
{
    unsigned long flags;

    save_and_cli(flags);
    if (--irq_info[irq_nr].depth < 0)
	unmask_irq(irq_nr);
    restore_flags(flags);
}

/*
 * Pointers to the low-level handlers: first the general ones, then the
 * fast ones, then the bad ones.
 */
extern void interrupt(void);

int get_irq_list(char *buf)
{
    int i, len = 0;
    struct irqaction *action;

    for (i = 0; i < NR_IRQS; i++) {
	action = irq_info[i].irq_action;
	if (!action)
	    continue;
	len += sprintf(buf + len, "%2d: %8d %c %s",
		       i, kstat.irqs[0][i],
		       (action->flags & SA_INTERRUPT) ? '+' : ' ',
		       action->name);
	for (action = action->next; action; action = action->next) {
	    len += sprintf(buf + len, ",%s %s",
			   (action->flags & SA_INTERRUPT) ? " +" : "",
			   action->name);
	}
	len += sprintf(buf + len, "\n");
    }
    return len;
}

#ifdef CONFIG_PLUM2
static int plum2_interrupt(void)
{
    u32 status;

    status = Plum2_IntStatus & 0xff;
    if(!status)
	return 63;

    if (status & PLUM2_CSC_INT) return PLUM2_CSC_IRQ;
    if (status & PLUM2_PCC_INT) {
	u32 pcc_status = Plum2_PccIntStatus;
	if (pcc_status & PLUM2_PCC1_INT) return PLUM2_PCC1_IRQ;
	if (pcc_status & PLUM2_PCC2_INT) return PLUM2_PCC2_IRQ;
    }
    if (status & PLUM2_DISP_INT) return PLUM2_DISP_IRQ;
    if (status & PLUM2_USB_INT) return PLUM2_USB_IRQ;
    if (status & PLUM2_USBWAKE_INT) return PLUM2_USBWAKE_IRQ;
    if (status & PLUM2_SM_INT) return PLUM2_SM_IRQ;
    return PLUM2_EXT_IRQ;
}
#endif

static inline void dispatch_irq_handler(int irq, struct pt_regs *regs)
{
    struct irqaction *action;
    int do_random, cpu;
    
    cpu = smp_processor_id();
    irq_enter(cpu);
    kstat.irqs[cpu][irq]++;

    mask_irq_with_plum2_clear(irq);

    action = irq_info[irq].irq_action;

    if (action) {
	if (!(action->flags & SA_INTERRUPT))
	    __sti();

	do_random = 0;
	do {
	    do_random |= action->flags;
	    action->handler(irq, action->dev_id, regs);
	    action = action->next;
	} while (action);
	if (do_random & SA_SAMPLE_RANDOM)
	    add_interrupt_randomness(irq);

	unmask_irq(irq);
	__cli();
    }
    irq_exit(cpu);
}

struct tx3920_irq_table{
    int irq;
    int regno;
    u32 status_bit;
    int priority;
};

static struct tx3920_irq_table irq_table_high_dummy[] ={
    {-1, 0, 0},
};
static struct tx3920_irq_table irq_table_high_10[] ={
    {7, INT8_REG, INT8_IOPOSINT5, 10},
#ifdef CONFIG_SHARP_TELIOS
    {31, INT8_REG, INT8_IONEGINT6},
#endif
    {-1, 0, 0},
};
static struct tx3920_irq_table irq_table_high_12[] ={
    {3, INT2_REG, INT2_UARTARXINT, 12},
    {4, INT2_REG, INT2_UARTBRXINT, 12},
    {-1, 0, 0},
};
static struct tx3920_irq_table irq_table_high_13[] ={
    {0, INT5_REG, INT5_PERINT, 13},
    {-1, 0, 0},
};
static struct tx3920_irq_table irq_table_high_14[] ={
    {6, INT5_REG, INT5_ALARMINT, 14},
    {-1, 0, 0},
};

static struct tx3920_irq_table *irq_table_high_vec[16] ={
    irq_table_high_dummy,		/* not use */
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_dummy,
    irq_table_high_10,
    irq_table_high_dummy,
    irq_table_high_12,
    irq_table_high_13,
    irq_table_high_14,
    irq_table_high_dummy,
};

static struct tx3920_irq_table irq_table_low[] ={
    {19, INT1_REG, INT1_SIBSF0INT, 0},
    {20, INT1_REG, INT1_SIBIRQPOSINT, 0},
    {3, INT2_REG, INT2_UARTA_INT, 0},
    {4, INT2_REG, INT2_UARTB_INT, 0},
#ifdef CONFIG_SHARP_TELIOS
# ifdef CONFIG_SOUND_TELIOS
    {IRQ_CHI1_0INT, INT1_REG, INT1_CHI1_0INT, 0},
    {IRQ_CHI0_5INT, INT1_REG, INT1_CHI0_5INT, 0},
# endif
    {1, INT3_REG, INT3_MFIOPOSINT(12), 0},
    {21, INT5_REG, INT5_SPIBUFAVAILINT|INT5_SPIERRINT|INT5_SPIRCVINT|INT5_SPIEMPTYINT, 0},
    {29, INT4_REG, INT4_MFIONEGINT(9), 0},
    {30, INT8_REG, INT8_IOPOSINT8 | INT8_IOPOSINT7, 0},
    {31, INT8_REG, INT8_IONEGINT6 | INT8_IOPOSINT6, 0},
#endif
    {-1, 0, 0, 0},
};

static u32 tx39int_status_low[TX39INT_REGS];
static u32 tx39int_status[TX39INT_REGS];

u32 tx39uart_status;

#ifdef DEBUG_IRQ
static int tx3920_search_irq_table(int irq, struct tx3920_irq_table *table)
{
    int i;

    for( i=0; table[i].irq != -1; i++ )
	if( table[i].irq == irq )
	    return i;
    return -1;
}

static int tx3920_check_irq(int irq)
{
    int i;
    if( tx3920_search_irq_table(irq, irq_table_low) >= 0 )
	return 0;
    for(i=0; i<16; i++)
	if( tx3920_search_irq_table(irq, irq_table_high_vec[i]) >= 0 )
	    return 0;
    if( PLUM2_IRQ_BASE <= irq && irq < PLUM2_IRQ_BASE + PLUM2_IRQ_MAX )
	return 0;
    return -EINVAL;
}

static void print_unknown_interrupt(int irqlow)
{
#ifdef LAZY_READ_STATUS
    static inline u32 tx39int_get_status_raw(int regno);
    static inline u32 tx39int_get_status(int regno);

    extern void tx3920_print_intreg(void);
    int i;
    IntClear1 = 0xffffffff;
    IntClear2 = 0xffffffff;
    IntClear3 = 0xffffffff;
    IntClear4 = 0xffffffff;
    IntClear5 = 0xffffffff;
    //	IntClear6 dose not exist
    IntClear7 = 0xffffffff;
    IntClear8 = 0xffffffff;
    for( i=0; i<TX39INT_REGS; i++ ){
	tx39int_status[i] = tx39int_get_status_raw(i);
	tx39int_status_low[i] = tx39int_get_status(i);
    }

    if( irqlow == 0 ){
	printk("irq.c:do_IRQ: unknown interrupt (irqhigh:%d)\n", INTVECT(tx39int_status[INT6_REG]));
	printk("Stat1: %08x  Stat2: %08x  Stat3: %08x  Stat4: %08x\n",
	       tx39int_status[INT1_REG],
	       tx39int_status[INT2_REG],
	       tx39int_status[INT3_REG],
	       tx39int_status[INT4_REG]);
	printk("Stat5: %08x  Stat6: %08x  Stat7: %08x  Stat8: %08x\n",
	       tx39int_status[INT5_REG],
	       tx39int_status[INT6_REG],
	       tx39int_status[INT7_REG],
	       tx39int_status[INT8_REG]);
	tx3920_print_intreg();

    }else{
	printk("irq.c:do_IRQ: unknown interrupt (irqlow)\n");
	printk("Stat1: %08x  Stat2: %08x  Stat3: %08x  Stat4: %08x\n",
	       tx39int_status[INT1_REG],
	       tx39int_status[INT2_REG],
	       tx39int_status[INT3_REG],
	       tx39int_status[INT4_REG]);
	printk("Stat5: %08x  Stat6: %08x  Stat7: %08x  Stat8: %08x\n",
	       tx39int_status[INT5_REG],
	       tx39int_status[INT6_REG],
	       tx39int_status[INT7_REG],
	       tx39int_status[INT8_REG]);
	printk("Masked Stat1: %08x  Stat2: %08x  Stat3: %08x  Stat4: %08x\n",
	       tx39int_status_low[INT1_REG],
	       tx39int_status_low[INT2_REG],
	       tx39int_status_low[INT3_REG],
	       tx39int_status_low[INT4_REG]);
	printk("Masked Stat5: %08x  Stat6: %08x  Stat7: %08x  Stat8: %08x\n",
	       tx39int_status_low[INT5_REG],
	       tx39int_status_low[INT6_REG],
	       tx39int_status_low[INT7_REG],
	       tx39int_status_low[INT8_REG]);
    }
#else
    extern void tx3920_print_intreg(void);
    IntClear1 = 0xffffffff;
    IntClear2 = 0xffffffff;
    IntClear3 = 0xffffffff;
    IntClear4 = 0xffffffff;
    IntClear5 = 0xffffffff;
    //	IntClear6 dose not exist
    IntClear7 = 0xffffffff;
    IntClear8 = 0xffffffff;
    if( irqlow == 0 ){
	printk("irq.c:do_IRQ: unknown interrupt (irqhigh:%d)\n", INTVECT(tx39int_status[INT6_REG]));
	printk("Stat1: %08x  Stat2: %08x  Stat3: %08x  Stat4: %08x\n",
	       tx39int_status[INT1_REG],
	       tx39int_status[INT2_REG],
	       tx39int_status[INT3_REG],
	       tx39int_status[INT4_REG]);
	printk("Stat5: %08x  Stat6: %08x  Stat7: %08x  Stat8: %08x\n",
	       tx39int_status[INT5_REG],
	       tx39int_status[INT6_REG],
	       tx39int_status[INT7_REG],
	       tx39int_status[INT8_REG]);
	tx3920_print_intreg();

    }else{
	printk("irq.c:do_IRQ: unknown interrupt (irqlow)\n");
	printk("Stat1: %08x  Stat2: %08x  Stat3: %08x  Stat4: %08x\n",
	       tx39int_status[INT1_REG],
	       tx39int_status[INT2_REG],
	       tx39int_status[INT3_REG],
	       tx39int_status[INT4_REG]);
	printk("Stat5: %08x  Stat6: %08x  Stat7: %08x  Stat8: %08x\n",
	       tx39int_status[INT5_REG],
	       tx39int_status[INT6_REG],
	       tx39int_status[INT7_REG],
	       tx39int_status[INT8_REG]);
	printk("Masked Stat1: %08x  Stat2: %08x  Stat3: %08x  Stat4: %08x\n",
	       tx39int_status_low[INT1_REG],
	       tx39int_status_low[INT2_REG],
	       tx39int_status_low[INT3_REG],
	       tx39int_status_low[INT4_REG]);
	printk("Masked Stat5: %08x  Stat6: %08x  Stat7: %08x  Stat8: %08x\n",
	       tx39int_status_low[INT5_REG],
	       tx39int_status_low[INT6_REG],
	       tx39int_status_low[INT7_REG],
	       tx39int_status_low[INT8_REG]);
    }
#endif
}
#endif

#ifdef LAZY_READ_STATUS
static u32 tx39int_read_flags;

static struct tx39int_buf{
    u32 status_raw;
    u32 status;
    volatile unsigned long *status_reg;
    volatile unsigned long *enable_reg;
}tx39int_buf[TX39INT_REGS];

static void tx39int_get_status_init(void)
{
    tx39int_buf[INT1_REG].status_reg = (volatile unsigned long *)&IntStatus1;
    tx39int_buf[INT2_REG].status_reg = (volatile unsigned long *)&IntStatus2;
    tx39int_buf[INT3_REG].status_reg = (volatile unsigned long *)&IntStatus3;
    tx39int_buf[INT4_REG].status_reg = (volatile unsigned long *)&IntStatus4;
    tx39int_buf[INT5_REG].status_reg = (volatile unsigned long *)&IntStatus5;
    tx39int_buf[INT6_REG].status_reg = (volatile unsigned long *)&IntStatus6;
    tx39int_buf[INT7_REG].status_reg = (volatile unsigned long *)&IntStatus7;
    tx39int_buf[INT8_REG].status_reg = (volatile unsigned long *)&IntStatus8;
    tx39int_buf[INT1_REG].enable_reg = (volatile unsigned long *)&IntEnable1;
    tx39int_buf[INT2_REG].enable_reg = (volatile unsigned long *)&IntEnable2;
    tx39int_buf[INT3_REG].enable_reg = (volatile unsigned long *)&IntEnable3;
    tx39int_buf[INT4_REG].enable_reg = (volatile unsigned long *)&IntEnable4;
    tx39int_buf[INT5_REG].enable_reg = (volatile unsigned long *)&IntEnable5;
    tx39int_buf[INT6_REG].enable_reg = (volatile unsigned long *)&IntEnable6;
    tx39int_buf[INT7_REG].enable_reg = (volatile unsigned long *)&IntEnable7;
    tx39int_buf[INT8_REG].enable_reg = (volatile unsigned long *)&IntEnable8;
}

static inline u32 tx39int_get_status_raw(int regno)
{
    if( tx39int_read_flags & (1 << regno) ){
	return tx39int_buf[regno].status_raw;
    }else{
	tx39int_read_flags |= 1 << regno;
	tx39int_buf[regno].status_raw = *(tx39int_buf[regno].status_reg);
	return tx39int_buf[regno].status_raw;
    }
}

static inline u32 tx39int_get_status(int regno)
{
    if( tx39int_read_flags & (1 << (regno + TX39INT_REGS)) ){
	return tx39int_buf[regno].status;
    }else{
	tx39int_read_flags |= 1 << (regno + TX39INT_REGS);
	tx39int_buf[regno].status = tx39int_get_status_raw(regno) &
	    *(tx39int_buf[regno].enable_reg);
	return tx39int_buf[regno].status;
    }
}
#endif

/*
 * do_IRQ handles IRQ's that have been installed without the
 * SA_INTERRUPT flag: it uses the full signal-handling return
 * and runs with other interrupts enabled. All relatively slow
 * IRQ's should use this format: notably the keyboard/timer
 * routines.
 */

asmlinkage void do_IRQ(int irq, struct pt_regs *regs)
{
    int pendint=0, plum2irq;
    struct tx3920_irq_table *table;

#define CONFIG_IRQ_STACK_CHECK
#ifdef CONFIG_IRQ_STACK_CHECK
    {
	long off;

	off = regs->regs[29] - (long)current;	/* current - sp */
	if( off < sizeof(struct task_struct) + sizeof(struct pt_regs) * 4 ){
	    mask_irq(33);
	    mask_irq(34);
	    print_unknown_interrupt(irq);
	    printk("kernel stack overflow!! (irq=%d, current=%08x, sp=%08x)\n", irq, current, regs->regs[29]);
	    while(1)
		;
	}
    }
#endif

#ifdef LAZY_READ_STATUS
    tx39int_read_flags = 0;
    if (irq == 0){
	table = irq_table_high_vec[INTVECT(IntStatus6)];
	while( table->irq != -1 ){
	    if(tx39int_get_status_raw(table->regno) & table->status_bit){
		pendint = 1;
		if( table->irq == 3 || table->irq == 4 )
		    tx39uart_status = tx39int_get_status_raw(INT2_REG);
#ifdef CONFIG_PLUM2
		if( table->irq == 7 ){
		    plum2irq=plum2_interrupt();
		    dispatch_irq_handler(plum2irq, regs);
		}else
#endif
		    dispatch_irq_handler(table->irq, regs);
		break;
	    }
	    table++;
	}
    }else{
	table = irq_table_low;
	while( table->irq != -1 ){
	    if(tx39int_get_status(table->regno) & table->status_bit){
		pendint = 1;
		if( table->irq == 3 || table->irq == 4 )
		    tx39uart_status = tx39int_get_status_raw(INT2_REG);
		dispatch_irq_handler(table->irq, regs);
		break;
	    }
	    table++;
	}
    }
    if( !pendint ){
#ifdef DEBUG_IRQ
	print_unknown_interrupt(irq);
#endif
	dispatch_irq_handler(63, regs);
    }
#else
    tx39int_status[INT6_REG] = IntStatus6;	/* read first !! */
    tx39int_status[INT1_REG] = IntStatus1;
    tx39int_status[INT2_REG] = IntStatus2;
    tx39int_status[INT3_REG] = IntStatus3;
    tx39int_status[INT4_REG] = IntStatus4;
    tx39int_status[INT5_REG] = IntStatus5;
    tx39int_status[INT7_REG] = IntStatus7;
    tx39int_status[INT8_REG] = IntStatus8;

    if (irq == 0){
	table = irq_table_high_vec[INTVECT(tx39int_status[INT6_REG])];
	while( table->irq != -1 ){
	    if(tx39int_status[table->regno] & table->status_bit){
		pendint = 1;
		if( table->irq == 3 || table->irq == 4 )
		    tx39uart_status = tx39int_status[INT2_REG];
#ifdef CONFIG_PLUM2
		if( table->irq == 7 ){
		    plum2irq=plum2_interrupt();
		    dispatch_irq_handler(plum2irq, regs);
		}else
#endif
		    dispatch_irq_handler(table->irq, regs);
		break;
	    }
	    table++;
	}
    }else{
	tx39int_status_low[INT1_REG] = tx39int_status[INT1_REG] & IntEnable1;
	tx39int_status_low[INT2_REG] = tx39int_status[INT2_REG] & IntEnable2;
	tx39int_status_low[INT3_REG] = tx39int_status[INT3_REG] & IntEnable3;
	tx39int_status_low[INT4_REG] = tx39int_status[INT4_REG] & IntEnable4;
	tx39int_status_low[INT5_REG] = tx39int_status[INT5_REG] & IntEnable5;
	tx39int_status_low[INT6_REG] = tx39int_status[INT6_REG];
	tx39int_status_low[INT7_REG] = tx39int_status[INT7_REG] & IntEnable7;
	tx39int_status_low[INT8_REG] = tx39int_status[INT8_REG] & IntEnable8;

	table = irq_table_low;
	while( table->irq != -1 ){
	    if(tx39int_status_low[table->regno] & table->status_bit){
		pendint = 1;
		if( table->irq == 3 || table->irq == 4 )
		    tx39uart_status = tx39int_status[INT2_REG];
		dispatch_irq_handler(table->irq, regs);
		break;
	    }
	    table++;
	}
    }
    if( !pendint ){
#ifdef DEBUG_IRQ
	print_unknown_interrupt(irq);
#endif
	dispatch_irq_handler(63, regs);
    }
#endif
    return;
}

/*
 * Idea is to put all interrupts
 * in a single table and differenciate them just by number.
 */
int setup_tx3920_irq(int irq, struct irqaction *new)
{
    int shared = 0;
    struct irqaction *old, **p;
    unsigned long flags;

#ifdef DEBUG_IRQ
    if( tx3920_check_irq(irq) ){
	printk("setup_tx3920_irq: unknown irq number (%d)\n", irq);
	return -EINVAL;
    }
#endif

    p = &irq_info[irq].irq_action;
    if ((old = *p) != NULL) {
	/* Can't share interrupts unless both agree to */
	if (!(old->flags & new->flags & SA_SHIRQ))
	    return -EBUSY;

	/* Can't share interrupts unless both are same type */
	if ((old->flags ^ new->flags) & SA_INTERRUPT)
	    return -EBUSY;

	/* add new interrupt at end of irq queue */
	do {
	    p = &old->next;
	    old = *p;
	} while (old);
	shared = 1;
    }
    if (new->flags & SA_SAMPLE_RANDOM)
	rand_initialize_irq(irq);

    save_and_cli(flags);
    *p = new;

    if (!shared) {
	irq_info[irq].depth = -1;
	unmask_irq(irq);
    }
    restore_flags(flags);
    return 0;
}

int request_irq(unsigned int irq,
		void (*handler) (int, void *, struct pt_regs *),
		unsigned long irqflags,
		const char *devname,
		void *dev_id)
{
    int retval;
    struct irqaction *action;

    if (irq >= NR_IRQS)
	return -EINVAL;
    if (!handler)
	return -EINVAL;

    action = (struct irqaction *) kmalloc(sizeof(struct irqaction), GFP_KERNEL);
    if (!action)
	return -ENOMEM;

    action->handler = handler;
    action->flags = irqflags;
    action->mask = 0;
    action->name = devname;
    action->next = NULL;
    action->dev_id = dev_id;

    retval = setup_tx3920_irq(irq, action);

    if (retval)
	kfree(action);
    return retval;
}

void free_irq(unsigned int irq, void *dev_id)
{
    struct irqaction *action, **p;
    unsigned long flags;

    if (irq >= NR_IRQS) {
	printk("Trying to free IRQ%d\n", irq);
	return;
    }
    for (p = &irq_info[irq].irq_action; (action = *p) != NULL; p = &action->next) {
	if (action->dev_id != dev_id)
	    continue;

	/* Found it - now free it */
	save_and_cli(flags);
	*p = action->next;
	if (!irq_info[irq].irq_action){
	    irq_info[irq].depth = 0;
	    mask_irq(irq);
	}
	restore_flags(flags);
	kfree(action);
	return;
    }
    printk("Trying to free free IRQ%d\n", irq);
}

unsigned long probe_irq_on(void)
{
    /* TODO */
    return 0;
}

int probe_irq_off(unsigned long irqs)
{
    /* TODO */
    return 0;
}

#ifdef CONFIG_PLUM2
static struct irqaction irq7 = { NULL, SA_INTERRUPT, 0,
                                 "dummy", NULL, NULL };
#endif /* CONFIG_PLUM2 */

int (*irq_cannonicalize)(int irq);

static int tmpr3922_irq_cannonicalize(int irq)
{
	return ((irq == 2) ? 9 : irq);
}

void __init init_IRQ(void)
{
#ifdef LAZY_READ_STATUS
	tx39int_get_status_init();
#endif
	irq_cannonicalize = tmpr3922_irq_cannonicalize;
	/*
	 * Turn all interrupts off
	 */
	IntEnable1 = 0;
	IntEnable2 = 0;
	IntEnable3 = 0;
	IntEnable4 = 0;
	IntEnable5 = 0;
	IntEnable6 = 0;
	IntEnable7 = 0;
	IntEnable8 = 0;

	IntClear1 = 0xffffffff;
	IntClear2 = 0xffffffff;
	IntClear3 = 0xffffffff;
	IntClear4 = 0xffffffff;
	IntClear5 = 0xffffffff;
//	IntClear6 dose not exist
	IntClear7 = 0xffffffff;
	IntClear8 = 0xffffffff;

	IntEnable6 = INT6_GLOBALEN | INT6_PERINT | /* INT6_UARTARXINT |*/
	             INT6_IOPOSINT56 | INT6_ALARMINT;

#ifdef CONFIG_PLUM2
	IntEnable8 = INT8_IOPOSINT5;

	Plum2_IntEnable = 0;
	Plum2_ExtIntEnable = 0;
	Plum2_PccIntEnable = PLUM2_PCC1_INT_INPUT_ENABLE |
	    		     PLUM2_PCC2_INT_INPUT_ENABLE;
	Plum2_UsbIntEnable = 0;
	Plum2_SmIntEnable = 0;

	Plum2_McsEnable = PLUM2_RMCS_ENABLE;
	Plum2_ExtIntClear = 0xffffffff;
	Plum2_PccIntClear = 0xffffffff;
	Plum2_IntEnable = PLUM2_INT_ENABLE;
#endif

	set_except_vector(0, tx3920_handle_int);

	set_cp0_status(ST0_BEV, 0);
	set_cp0_status(ST0_IM,
	               read_32bit_cp0_register(CP0_STATUS) | IE_IRQ2 | IE_IRQ4);
	setup_tx3920_irq(7, &irq7);
}

EXPORT_SYMBOL(irq_cannonicalize);
