/*
 * optpoint - Routines for optpoint (Pointing device) on SHARP Telios HC-AJ2
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001 Hirokazu FUKUI <fuku@axe-inc.co.jp>
 *
 * Based on arch/mips/tx3920/tpanel.c
 *
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/init.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/tx3920.h>
#include <linux/delay.h>
#include "spi.h"

#ifdef CONFIG_SHARP_TELIOS
#include "sharp_telios.h"
#endif

#define STATUS_RESET	0
#define STATUS_SEND	1
#define STATUS_RECV	2
#define STATUS_RECV_F	3
static int optp_status;

#define PROTO_PS2	0	/* Musmice of the ps/2 series   */
#define PROTO_IMPS2	1	/* Microsoft Intellimouse (ps2) */
#define PROTO_MSC	2	/* Microsoft Intellimouse (ps2) */
static int device_proto = PROTO_PS2;


#if 0
static void optpoint_poll_handler(unsigned long dev_addr);

static struct timer_list optpoint_poll_timer = {
    {NULL, NULL}, 0UL, 0UL, optpoint_poll_handler
};
#endif

static wait_queue_head_t optpoint_wait;
static struct fasync_struct *fasyncptr;


struct optpoint_recv {
    unsigned char bytes[3];
    char ptr;
};

#define RECV_BUFSIZE 16
#define RECV_BUF_NEXT(x) (((x) == recvbuf + RECV_BUFSIZE - 1) ? recvbuf : (x)+1)
static struct optpoint_recv recvbuf[RECV_BUFSIZE];
static struct optpoint_recv *recvbuf_head = recvbuf;
static struct optpoint_recv *recvbuf_tail = recvbuf;

static int optpoint_put_recvbuf(unsigned char data)
{
    /* check first byte */
    if( recvbuf_head->ptr == 0 ){
	if( (data & 0xcc) != 0x08 ){
	    printk("optpoint: Bad first byte (0x%02x)\n", data);
	    return 0;
	}
    }

    recvbuf_head->bytes[(int)recvbuf_head->ptr] = data;
    recvbuf_head->ptr++;

    if( recvbuf_head->ptr < 3 )
	return 0;
      
    recvbuf_head->ptr = 0;

    recvbuf_head = RECV_BUF_NEXT(recvbuf_head);

    if (recvbuf_head == recvbuf_tail) {
	recvbuf_tail = RECV_BUF_NEXT(recvbuf_tail);
	if( recvbuf_head->ptr > 0 )
	    *recvbuf_tail = *recvbuf_head;	/* overwrite */
    }

    recvbuf_head->ptr = 0;

    wake_up_interruptible(&optpoint_wait);
    if (fasyncptr)
	kill_fasync(&fasyncptr, SIGIO, POLL_IN);
    return 1;
}


#define SEND_BUFSIZE 64
#define SEND_BUF_NEXT(x) (((x) == sendbuf + SEND_BUFSIZE - 1) ? sendbuf : (x)+1)
static unsigned char sendbuf[SEND_BUFSIZE];
static unsigned char *sendbuf_head = sendbuf;
static unsigned char *sendbuf_tail = sendbuf;

static int optpoint_put_sendbuf(unsigned char data)
{
    unsigned char *next_head;

    next_head = SEND_BUF_NEXT(sendbuf_head);
    if( next_head == sendbuf_tail )
	return -1;

    *sendbuf_head = data;
    sendbuf_head = next_head;
    return 0;
}


/***********************************************************************/

static int fasync_optpoint(int fd, struct file *filp, int on)
{
  int retval;

  retval = fasync_helper(fd, filp, on, &fasyncptr);
  if (retval < 0)
    return retval;
  return 0;
}

static int open_optpoint(struct inode * inode, struct file * file)
{
  recvbuf_tail = recvbuf_head;

  return 0;
}

static int close_optpoint(struct inode * inode, struct file * file)
{
  fasync_optpoint(-1, file, 0);
  return 0;
}

static ssize_t read_optpoint(struct file * file,
			   char * buffer, size_t count, loff_t *ppos)
{
    unsigned char data;
    int temp;
    size_t retcnt = 0;

    if (count < 1)
	return -EINVAL;

    // We need to access the circular buffer atomic to
    // anything else that will read or write it

    cli();

    init_waitqueue_head(&optpoint_wait);

    while (recvbuf_head == recvbuf_tail) {
	if (file->f_flags & O_NONBLOCK) {
	    sti();
	    return -EAGAIN;
	}
	interruptible_sleep_on(&optpoint_wait);
	// sleep_on will sti for us
	if (signal_pending(current))
	    return -ERESTARTSYS;
	cli();
    }

    do {
	switch( device_proto ){
	case PROTO_IMPS2:
	    if( recvbuf_tail->ptr < 3 ){
		data = recvbuf_tail->bytes[(int)(recvbuf_tail->ptr)];
		recvbuf_tail->ptr++;
	    }else{
		recvbuf_tail = RECV_BUF_NEXT(recvbuf_tail);
		data = 0x00;
	    }
	case PROTO_MSC:
	    switch(recvbuf_tail->ptr++){
	    case 0:
		temp = recvbuf_tail->bytes[0];
		data = 0x80
		    | ((((temp & 0x01) << 2)		/* left button  */
		       | ((temp & 0x02) >> 1)) ^ 0x7);	/* right button */
		break;
	    case 1:
		temp = recvbuf_tail->bytes[1];
		if( recvbuf_tail->bytes[0] & 0x10 )
		    temp = temp - 0x100;
		data = temp/2;
		break;
	    case 2:
		temp = recvbuf_tail->bytes[2];
		if( recvbuf_tail->bytes[0] & 0x20 )
		    temp = temp - 0x100;
		data = temp/2;
		break;
	    case 3:
		temp = recvbuf_tail->bytes[1];
		if( recvbuf_tail->bytes[0] & 0x10 )
		    temp = temp - 0x100;
		data = temp - temp/2;
		break;
	    case 4:
		temp = recvbuf_tail->bytes[2];
		if( recvbuf_tail->bytes[0] & 0x20 )
		    temp = temp - 0x100;
		data = temp - temp/2;
		recvbuf_tail = RECV_BUF_NEXT(recvbuf_tail);
		break;
	    default:
		/* not reached */
		data = 0;
	    }
	    break;
	case PROTO_PS2:
	default:
	    if( recvbuf_tail->ptr < 2 ){
		data = recvbuf_tail->bytes[(int)(recvbuf_tail->ptr)];
		recvbuf_tail->ptr++;
	    }else{
		data = recvbuf_tail->bytes[(int)(recvbuf_tail->ptr)];
		recvbuf_tail->ptr++;
		recvbuf_tail = RECV_BUF_NEXT(recvbuf_tail);
	    }
	}
		
	sti();
		
	if (!access_ok(VERIFY_WRITE, buffer, 3))
	    return -EFAULT;

	__put_user(data, (unsigned char*)(buffer + retcnt));

	retcnt ++;
	if (retcnt >= count )
	    return retcnt;

	cli();
    } while (recvbuf_head != recvbuf_tail);

    sti();

    return retcnt;
}

static unsigned int poll_optpoint(struct file *file, poll_table * wait)
{
  int comp;

  poll_wait(file, &optpoint_wait, wait);
  cli();
  comp = (recvbuf_head != recvbuf_tail);
  sti();
  if (comp)
    return POLLIN | POLLRDNORM;
  return 0;
}

static void optpoint_line_ctrl(int c_req, int t_rdy)
{
    unsigned flags, temp;

    save_and_cli(flags);
    temp = MFIOOutput;
    temp &= ~(TELIOS_MFIO_OPTP_T_RDY | TELIOS_MFIO_OPTP_C_REQ);
    temp |= ((t_rdy ? TELIOS_MFIO_OPTP_T_RDY : 0)
	           | (c_req ? TELIOS_MFIO_OPTP_C_REQ : 0));
    MFIOOutput = temp;
    restore_flags(flags);
}

static int optpoint_get_s_enb_ne(void)
{
    return MFIOInput & TELIOS_MFIO_OPTP_S_ENB_N;
}

static void optpoint_stanby_mode(void)
{
    optpoint_put_sendbuf(0xf5);		/* Disable */
    optpoint_put_sendbuf(0xf2);		/* Set Stream Mode */
    optpoint_put_sendbuf(0xf8);		/* Stanby */
    optpoint_put_sendbuf(0xf4);		/* Enable */

    if( optp_status == STATUS_RECV ){
	if( recvbuf_head->ptr == 0 ){
	    optp_status = STATUS_SEND;
	    optpoint_line_ctrl(1, 1);
	}else
	    optp_status = STATUS_RECV_F;
    }
}

#define RECV_BYTE 0
#define SEND_BYTE 1
static unsigned char optpoint_send_a_byte(unsigned char data, int send_flag)
{
    unsigned char ret;

    while( optpoint_get_s_enb_ne() )
	;

    spi_put_word(data);

    if( send_flag )
	optpoint_line_ctrl( 0, 1 );	
    
    while( !optpoint_get_s_enb_ne() )
	;

    ret = spi_get_word() & 0xff;

    return ret;
}

#if 0
static void optpoint_poll_handler(unsigned long dev_addr)
{
    unsigned char bytes[3];
    optpoint_line_ctrl(0, 1);
    bytes[0] = optpoint_send_a_byte(0x00);
    bytes[1] = optpoint_send_a_byte(0x00);
    bytes[2] = optpoint_send_a_byte(0x00);
    optpoint_put_recvbuf(bytes);
    optpoint_line_ctrl(0, 0);

    wake_up_interruptible(&optpoint_wait);
    if (fasyncptr)
      kill_fasync(&fasyncptr, SIGIO, POLL_IN);


    mod_timer(&optpoint_poll_timer, jiffies + HZ/10);
}
#endif

static void optpoint_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
    extern void tx3920_print_spireg(void);
    unsigned char data;

    switch( optp_status ){
    case STATUS_SEND:
	data = *sendbuf_tail;
	sendbuf_tail = SEND_BUF_NEXT(sendbuf_tail);
	optpoint_send_a_byte(data, SEND_BYTE);

	if( sendbuf_head == sendbuf_tail )
	    optp_status = STATUS_RECV;
	else
	    optpoint_line_ctrl(1,1);
	break;

    case STATUS_RECV:
	data = optpoint_send_a_byte(0x00, RECV_BYTE);
	optpoint_put_recvbuf(data);
	break;

    case STATUS_RECV_F:
	data = optpoint_send_a_byte(0x00, RECV_BYTE);
	if( optpoint_put_recvbuf(data) ){
	    optp_status = STATUS_SEND;
	    optpoint_line_ctrl(1, 1);
	}
	break;
	
    case STATUS_RESET:
    default:
	printk("optpoint_interrupt: bad status\n");
    }
    IntClear4 = INT4_MFIONEGINT(9);
}

struct file_operations telios_optpoint_fops = {
  read:           read_optpoint,
  poll:           poll_optpoint,
  /*  ioctl:	  ioctl_optpoint, */
  open:           open_optpoint,
  release:        close_optpoint,
  fasync:	  fasync_optpoint,
};

static struct miscdevice telios_optpoint = {
  PSMOUSE_MINOR, "telios_optpoint", &telios_optpoint_fops
};

static int __init telios_optpoint_init(void)
{
  unsigned flags;
  int i;

  init_waitqueue_head(&optpoint_wait);

  save_and_cli(flags);

  spi_init();
  spi_enable(1);

  optpoint_line_ctrl(0,0);

  if ((i = request_irq(29, optpoint_interrupt, SA_INTERRUPT,
		       "optpoint", NULL))) {
      printk(KERN_ERR "optpoint: Cannot allocate irq. (%d)\n", i);
      return i;
  }

  fasyncptr = NULL;

  optp_status = STATUS_RECV;
  recvbuf_head->ptr = 0;

  optpoint_stanby_mode();

  restore_flags(flags);

  misc_register(&telios_optpoint);

#if 0
  init_timer(&optpoint_poll_timer);
  optpoint_poll_timer.expires = jiffies + HZ / 10;
  optpoint_poll_timer.function = optpoint_poll_handler;
  add_timer(&optpoint_poll_timer);
#endif

  printk("Optpoint driver initialized\n");
  return 0;
}

__initcall(telios_optpoint_init);
