/*
 * linux/arch/mips/tx3920/plum2lcd.c
 *
 * Toshiba PLUM2 LCD/CRT Controller support.
 *
 * Copyright (C) 2001 AXE,Inc.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/delay.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/plum2.h>
#include <asm/kseg2.h>
#include <asm/tx3920.h>

#ifdef CONFIG_SHARP_TELIOS
#include "sharp_telios.h"
#endif
static int backlight, lcdpower;

#define BKLIGHT_MASTER_MAX 7
static int backlight_master_val = BKLIGHT_MASTER_MAX - 1;

extern int plum2lcd_backlight(int n);

void plum2lcd_init_backlight(void)
{
    backlight = 8;
    plum2lcd_backlight(backlight);
}

int get_plum2lcd_backlight(void)
{
    return backlight;
}

int plum2lcd_backlight(int n)
{
#ifdef CONFIG_SHARP_TELIOS
    int flags;

    backlight = n;
    if(n <= 0) {
	save_and_cli(flags);
	BackLight_Period &= ~BKLGT_PERIOD_EN;
	BackLight_Reset = 0;
	restore_flags(flags);
    } else {
	n -= BKLIGHT_MASTER_MAX - backlight_master_val;
	if( n <= 0 )
	    n = 1;
	save_and_cli(flags);
	BackLight_Period |= BKLGT_PERIOD_EN;
	BackLight_Reset = BKLGT_LIGHTVAL(n);
	restore_flags(flags);
    }
#else
    backlight = n;
#endif
    return 0;
}

int plum2lcd_lcdpower(int n)
{
    int flags;

    lcdpower = n;
    if(n == 0) {
	save_and_cli(flags);
	Plum2_PowerCtrl &=  ~PLUM2_POWER_BKLIGHT;
	restore_flags(flags);
    } else {
	save_and_cli(flags);
	Plum2_PowerCtrl |=  PLUM2_POWER_BKLIGHT;
	restore_flags(flags);
    }
    return 0;
}

void plum2lcd_backlight_dump(void)
{
#ifdef CONFIG_SHARP_TELIOS
    BackLight_Delay ^= BKLGT_DELAY_IO2;
    printk("Key GA: base=%08lx\n", bklight_base);
    printk(" VR    =%02x\n", BackLight_VR);
    printk(" Period=%02x\n", BackLight_Period);
    printk(" Reset =%02x\n", BackLight_Reset);
    printk(" delay =%02x\n", BackLight_Delay);

    printk(" PowCtl =%02x\n", Plum2_PowerCtrl);
    printk(" LCDLUM =%02x\n", Plum2_DispLcdLuminance);
    printk("\n");
#endif
}


static void set_backlight_master(int n)
{
    if( 0 <= n  && n <= BKLIGHT_MASTER_MAX ){
	backlight_master_val = n;
	plum2lcd_backlight(get_plum2lcd_backlight());
    }
}

void backlight_master_inc(unsigned char input_kcode, int down)
{
    if( down )
	set_backlight_master(backlight_master_val + 1);
}

void backlight_master_dec(unsigned char input_kcode, int down)
{
    if( down )
	set_backlight_master(backlight_master_val - 1);
}

void plum2lcd_setup(void)
{
    sharp_telios_map_bklgt_io();
    lcdpower = 1; /* boot time always on */
    plum2lcd_init_backlight();
    plum2lcd_lcdpower(lcdpower);
    plum2lcd_backlight(backlight);
}
