/*
 * memory.c: memory initialisation code.
 *
 * Copyright (C) Harald Koerfgen
 * Copyright (C) Jim Pick
 * Copyright (C) Yuasa Yoichi <yuasa@users.sourceforge.net>
 * Copyright (C) Hirokazu FUKUI <fuku@axe-inc.co.jp>
 *
 * $Id: memory.c,v 1.11 2000/05/02 14:28:47 nop Exp $
 */
#include <asm/addrspace.h>
#include <asm/tx3920.h>
#include <asm/bootinfo.h>
#include <asm/page.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/bootmem.h>

#undef PROM_DEBUG

#ifdef PROM_DEBUG
extern int prom_printf(char * fmt, ...);
#endif

volatile unsigned long mem_err = 0;

extern char _end;
extern char _fdata;

#define PFN_UP(x)	(((x) + PAGE_SIZE-1) >> PAGE_SHIFT)
#define PFN_DOWN(x)	((x) >> PAGE_SHIFT)
#define PFN_ALIGN(x)	(((unsigned long)(x) + (PAGE_SIZE - 1)) & PAGE_MASK)

#define TX39_DRAM_MAGIC	0xa5a5a5a5

#define TX39_DRAM_PAGE_SHIFT	20
#define TX39_DRAM_PAGE_SIZE	(1UL << TX39_DRAM_PAGE_SHIFT)
#define TX39_DRAM_PAGE_MASK	((~(TX39_DRAM_PAGE_SIZE - 1)) & 0x3fffffff)

#if 0
static unsigned long __init tx39_get_memory_size(void)
{
	unsigned long memory_size;
	u32 *page;

	memory_size = ((unsigned long)&_end + (TX39_DRAM_PAGE_SIZE - 1)) &
	              TX39_DRAM_PAGE_MASK;

#ifdef PROM_DEBUG
	prom_printf("kernel_end: 0x%08x\n", memory_size);
#endif

	while (memory_size < TX39_SYSADDR_DRAM_BANK_SIZE) {
		page = (u32 *)KSEG1ADDR(memory_size);
		*(volatile u32 *)page = TX39_DRAM_MAGIC;
		*(volatile u32 *)(page + 4) = ~TX39_DRAM_MAGIC;
		barrier();

		if ((*(volatile u32 *)page) != TX39_DRAM_MAGIC ||
		    (*(volatile u32 *)(page + 4)) != ~TX39_DRAM_MAGIC)
			break;

		/* add 1MB */
		memory_size += TX39_DRAM_PAGE_SIZE;
	}

	
#ifdef PROM_DEBUG
	prom_printf("memory_size: 0x%08x\n", memory_size);
#endif
	return memory_size;
}
#endif

void __init prom_meminit(void)
{

	unsigned long bootmem_start, bootmem_end, start_pfn, bootmap_size;
	unsigned long bank1_start, bank1_end, bank2_start, bank2_end;
	unsigned long tx39_memconfig0;

	tx39_memconfig0 = MemConfig0 & 0x000ffff0;
	if( tx39_memconfig0 == 0x00031060 ){		/* HC-VJ1C, maybe */
	    bank1_start = 0x00000000;
	    bank1_end   = 0x02000000;
	    bank2_start = bank2_end = 0;
	}else if( tx39_memconfig0 == 0x000f5550 ){	/* HC-AJ2, maybe */
	    bank1_start = 0x00000000;
	    bank1_end   = 0x01000000;
	    bank2_start = 0x02000000;
	    bank2_end   = 0x03000000;
	}else if( tx39_memconfig0 == 0x0005f440 ){	/* HC-AJ1, maybe */
	    bank1_start = 0x00000000;
	    bank1_end   = 0x01000000;
	    bank2_start = 0x02000000;
	    bank2_end   = 0x03000000;
	}else{						/* unknown */
	    bank1_start = 0x00000000;
	    bank1_end   = 0x01000000;
	    bank2_start = bank2_end = 0;
	}

	bootmem_start = PHYSADDR(PFN_ALIGN(&_end));
	if( bank2_start == bank2_end )
	    bootmem_end = bank1_end;
	else
	    bootmem_end = bank2_end;

	start_pfn = PFN_UP(bootmem_start);

	/* Register all the contiguous memory with the bootmem allocator
	   and free it.  Be careful about the bootmem freemap.  */
	bootmap_size = init_bootmem(start_pfn, bootmem_end >> PAGE_SHIFT);
	free_bootmem(bootmem_start + bootmap_size,
		     bank1_end - (bootmem_start + bootmap_size));
	if( bank2_start != bank2_end )
	    free_bootmem(bank2_start, bank2_end - bank2_start);
}

#if 0
void __init prom_meminit(void)
{

	unsigned long free_start, free_end, start_pfn, bootmap_size;
	unsigned long mem_size = 0;

	mem_size = tx39_get_memory_size();

	free_start = PHYSADDR(PFN_ALIGN(&_end));
	free_end = mem_size;
	start_pfn = PFN_UP(PHYSADDR((unsigned long)&_end));

#ifdef PROM_DEBUG
	prom_printf("free_start: 0x%08x\n", free_start);
	prom_printf("free_end: 0x%08x\n", free_end);
	prom_printf("start_pfn: 0x%08x\n", start_pfn);
#endif

	/* Register all the contiguous memory with the bootmem allocator
	   and free it.  Be careful about the bootmem freemap.  */
	bootmap_size = init_bootmem(start_pfn, mem_size >> PAGE_SHIFT);
	free_bootmem(free_start + bootmap_size, free_end - free_start - bootmap_size);
}
#endif

int __init page_is_ram(unsigned long pagenr)
{
        return 1;
}

/* Called from mem_init() to fixup the mem_map page settings. */
void __init prom_fixup_mem_map(unsigned long start, unsigned long end)
{
}

void prom_free_prom_memory (void)
{
}
