/*
 * Setup the Toshiba TMPR3922.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001 AXE, Inc.    <fuku@axe-inc.co.jp>
 * Copyright (C) 2000 Yoichi Yuasa <yuasa@users.sourceforge.net>
 * Copyright (C) 1999 Harald Koerfgen
 */
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/mc146818rtc.h>
#include <linux/param.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/ide.h>
#include <asm/mipsregs.h>
#include <asm/bootinfo.h>
#include <asm/irq.h>
#include <asm/reboot.h>
#include <asm/tx3920.h>
#include <asm/plum2.h>
#include <asm/wbflush.h>
#include <asm/addrspace.h>
#ifdef CONFIG_KSEG2
#include <asm/kseg2.h>
#endif

extern void r39xx_machine_restart(char *command);
extern void r39xx_machine_halt(void);
extern void r39xx_machine_power_off(void);

extern struct ide_ops tx3920_ide_ops;
extern struct rtc_ops tx3920_rtc_ops;

extern int setup_tx3920_irq(int, struct irqaction *);

void (*board_time_init) (struct irqaction * irq);
void (*__wbflush) (void);

#ifdef CONFIG_SHARP_TELIOS
extern void sharp_telios_setup(void);
#endif

#define PRINT_CONFIG

#ifdef PRINT_CONFIG
void debug_registers(void)
{
    /*
     * Debugging aid: print some configuration registers 
     */
	printk("MemConfig0:    0x%08x\n", MemConfig0);
	printk("MemConfig1:    0x%08x\n", MemConfig1);
	printk("MemConfig2:    0x%08x\n", MemConfig2);
	printk("MemConfig3:    0x%08x\n", MemConfig3);
	printk("MemConfig4:    0x%08x\n", MemConfig4);
	printk("\n");
	printk("IOcontrol:     0x%08x\n", IOControl);
	printk("MFIOOutput:    0x%08x\n", MFIOOutput);
	printk("MFIODirection: 0x%08x\n", MFIODirection);
	printk("MFIOInput:     0x%08x\n", MFIOInput);
	printk("MFIOSelect:    0x%08x\n", MFIOSelect);
	printk("IOPowerDown:   0x%08x\n", IOPowerDown);
	printk("MFIOPowerDown: 0x%08x\n", MFIOPowerDown);
	printk("\n");
	printk("IntEnable1:    0x%08x\n", IntEnable1);
	printk("IntEnable2:    0x%08x\n", IntEnable2);
	printk("IntEnable3:    0x%08x\n", IntEnable3);
	printk("IntEnable4:    0x%08x\n", IntEnable4);
	printk("IntEnable5:    0x%08x\n", IntEnable5);
	printk("IntEnable6:    0x%08x\n", IntEnable6);
	printk("IntEnable7:    0x%08x\n", IntEnable7);
	printk("IntEnable8:    0x%08x\n", IntEnable8);
	printk("\n");
	printk("CP0_STATUS:    0x%08x\n",  read_32bit_cp0_register(CP0_STATUS));
	printk("CP0_CAUSE:     0x%08x\n",  read_32bit_cp0_register(CP0_CAUSE));
	printk("CP0_CONF:      0x%08x\n",  read_32bit_cp0_register(CP0_CONF));
	printk("\n");
	printk("ClockControl: 0x%08x\n", ClockControl);
	printk("PeriodTimer:  0x%08x\n", PeriodTimer);
	printk("TimerControl: 0x%08x\n", TimerControl);
	printk("\n");

#ifdef CONFIG_PLUM2
	printk("Plum2_IntEnable   :    0x%08x\n", Plum2_IntEnable);
	printk("Plum2_ExtIntEnable:    0x%08x\n", Plum2_ExtIntEnable);
	printk("Plum2_PccIntEnable:    0x%08x\n", Plum2_PccIntEnable);
	printk("Plum2_PccClkSelect:    0x%08x\n", Plum2_PccClkSelect);
	printk("Plum2_UsbIntEnable:    0x%08x\n", Plum2_UsbIntEnable);
	printk("Plum2_SmIntEnable :    0x%08x\n", Plum2_SmIntEnable);
#endif

	printk("\n");
}
#endif


#if 0
static void __init tx3920_irq_setup(void)
{

#ifdef PRINT_CONFIG
	debug_registers();
#endif

	/*
	 * Turn all interrupts off
	 */
	IntEnable1 = 0;
	IntEnable2 = 0;
	IntEnable3 = 0;
	IntEnable4 = 0;
	IntEnable5 = 0;
	IntEnable6 = 0;
	IntEnable7 = 0;
	IntEnable8 = 0;

	IntClear1 = 0xffffffff;
	IntClear2 = 0xffffffff;
	IntClear3 = 0xffffffff;
	IntClear4 = 0xffffffff;
	IntClear5 = 0xffffffff;
	IntClear7 = 0xffffffff;
	IntClear8 = 0xffffffff;

	IntEnable6 = INT6_GLOBALEN | INT6_PERINT | INT6_UARTARXINT |
	             INT6_IOPOSINT56;
	IntEnable8 = INT8_IOPOSINT5;

#ifdef CONFIG_PLUM2
	Plum2_IntEnable = 0;
	Plum2_ExtIntEnable = 0;
	Plum2_PccIntEnable = 0;
	Plum2_UsbIntEnable = 0;
	Plum2_SmIntEnable = 0;

	Plum2_McsEnable = PLUM2_RMCS_ENABLE;
	Plum2_ExtIntClear = 0xffffffff;
	Plum2_PccIntClear = 0xffffffff;
	Plum2_IntEnable = PLUM2_INT_ENABLE;
#endif

	set_except_vector(0, tx3920_handle_int);

	set_cp0_status(ST0_BEV, 0);
	set_cp0_status(ST0_IM,
	               read_32bit_cp0_register(CP0_STATUS) | IE_IRQ2 | IE_IRQ4);
}
#endif

/*
 * enable the periodic interrupts
 */
static __init void tx3920_time_init(struct irqaction *irq)
{
	/*
	 * Here we go, enable periodic interrupts.
	 */
	setup_tx3920_irq(0, irq);

	ClockControl |= CLK_ENTIMERCLK;
	PeriodTimer = PER_TIMER_COUNT;
	TimerControl = TIM_ENPERTIMER;
}

/*
 * Oh, well
 */
void tx3920_wbflush(void)
{
	/*
	 * The status of the writeback buffer is available
	 * via the Coprocessor 0 condition
	 */
	__asm__ __volatile__("
		.set	push
		.set	noreorder
		.set	mips2
		sync
		nop
	1:	bc0f	1b
		nop
		.set pop"
	);
}

#ifdef CONFIG_CPU_TMPR3922
#define TX3920_ALARM_IRQ	6

struct tx3920_alarm_box{
    u32 low;
    u32 high;
    void (*handler)(void *arg, struct pt_regs *regs);
    void *arg;
};

static struct tx3920_alarm_box tx3920_alarm_box;

static void
tx3920_alarm_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    void (*function)(void *arg, struct pt_regs *regs);

    IntEnable5 &= ~INT5_ALARMINT; 	/* I think no need this line
					   when ALARMINT is IRQHIGH */
    IntClear5 = INT5_ALARMINT;

    function = tx3920_alarm_box.handler;
    if( function ){
	tx3920_alarm_box.handler = NULL;
	(*function)(tx3920_alarm_box.arg, regs);
    }
}

int
tx3920_alarm_set_handler(u32 count,
			 void (*function)(void *arg, struct pt_regs *regs),
			 void *arg)
{
    u32 low, high, cur_low;
    unsigned flags;

    if( function == NULL )
	return -EINVAL;

    save_and_cli(flags);
    tx3920_read_rtc(high, cur_low);
    low = cur_low + count;
    if( low < cur_low )
	high = (high + 1) & RTC_HIGH_MASK;

    tx3920_alarm_box.low = low;
    tx3920_alarm_box.high = high;
    tx3920_alarm_box.handler = function;
    tx3920_alarm_box.arg = arg;

    AlarmLow = tx3920_alarm_box.low;
    AlarmHigh = tx3920_alarm_box.high;
    IntClear5 = INT5_ALARMINT;
    IntEnable5 |= INT5_ALARMINT; 	 /* I think no need this line
					   when ALARMINT is IRQHIGH */
    restore_flags(flags);
    return 0;
}

static struct irqaction irq6 = {tx3920_alarm_interrupt, SA_INTERRUPT, 0,
 			 "alarm", NULL, NULL};

static int __init tx3920_alarm_init(void)
{
    unsigned flags;
    int ret;

    save_and_cli(flags);
    ret = setup_tx3920_irq(TX3920_ALARM_IRQ, &irq6);
    IntEnable5 &= ~INT5_ALARMINT; 	/* IntEnable5 is of no effect
					   because IRQHIGH */
    restore_flags(flags);
    return ret;
}
/* __initcall(tx3920_alarm_init); */
#endif /* CONFIG_CPU_TMPR3922 */

void __init r39xx_setup(void)
{
	board_time_init = tx3920_time_init;

	isa_slot_offset = KSEG2;

	_machine_restart = r39xx_machine_restart;
	_machine_halt = r39xx_machine_halt;
	_machine_power_off = r39xx_machine_power_off;

#ifdef CONFIG_BLK_DEV_IDE
	ide_ops = &tx3920_ide_ops;
#endif

	rtc_ops = &tx3920_rtc_ops;

#ifdef CONFIG_FB
	conswitchp = &dummy_con;
#endif

#ifdef CONFIG_BLK_DEV_INITRD
	ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
#endif

	__wbflush = tx3920_wbflush;

#ifdef CONFIG_PLUM2PCC
	/* Shutdown PC Card */
	plum2pcc_pccard_shutdown();
#endif

#ifdef CONFIG_KSEG2
	kseg2_init();
#endif

#ifdef CONFIG_CPU_TMPR3922
    tx3920_alarm_init();
#endif

#ifdef CONFIG_SHARP_TELIOS
	sharp_telios_setup();
#endif
}

