/*
 * linux/arch/mips/tx3920/sharp_telios.c
 *
 * Platform support for Sharp Telios.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001 AXE, Inc.   <fuku@axe-inc.co.jp>
 *
 */

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/plum2.h>
#include <asm/kseg2.h>
#include <asm/tx3920.h>
#include "sharp_telios.h"

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <asm/pgalloc.h>

#define DEBUG_TELIOS
#define DEBUG_PRINT
#define DEBUG_CT

extern int setup_tx3920_irq(int, struct irqaction *);

unsigned long bklight_base=0;

#ifdef DEBUG_PRINT
void tx3920_print_CP0REG(void)
{
	printk("Status: %08x   Cause: %08x  EPC: %08x   Context:%08x\n",
	       read_32bit_cp0_register(CP0_STATUS),
	       read_32bit_cp0_register(CP0_CAUSE),
	       read_32bit_cp0_register(CP0_EPC),
	       read_32bit_cp0_register(CP0_CONTEXT));
	printk("BadVAddr: %08x   PRID: %08x  Config: %08x   Cache:%08x\n",
	       read_32bit_cp0_register(CP0_BADVADDR),
	       read_32bit_cp0_register(CP0_PRID),
	       read_32bit_cp0_register(CP0_CONF),
	       read_32bit_cp0_register($7)); /* CP0_CACHE */
	printk("TagLO: %08x\n",
	       read_32bit_cp0_register(CP0_XCONTEXT));
	printk("\n");
	printk("EntryHi: %08x   EntryLo: %08x  Index: %08x   Random:%08x\n",
	       read_32bit_cp0_register(CP0_ENTRYHI),
	       read_32bit_cp0_register(CP0_ENTRYLO0),
	       read_32bit_cp0_register(CP0_INDEX),
	       read_32bit_cp0_register(CP0_RANDOM));
	printk("PageMask: %08x   Wired: %08x\n",
	       read_32bit_cp0_register(CP0_PAGEMASK),
	       read_32bit_cp0_register(CP0_WIRED));
	printk("\n");
	printk("Debug: %08x   DEPC: %08x\n",
	       read_32bit_cp0_register(CP0_CONFIG),
	       read_32bit_cp0_register(CP0_LLADDR));
}

void tx3920_print_intreg(void)
{
    printk("tx3920 intreg:\n");
    printk("Ena1: %08x  Ena2: %08x  Ena3: %08x  Ena4: %08x \n",
	   IntEnable1, IntEnable2, IntEnable3, IntEnable4);
    printk("Ena5: %08x  Ena6: %08x  Ena7: %08x  Ena8: %08x \n",
	   IntEnable5, IntEnable6, IntEnable7, IntEnable8);
    printk("Stat1: %08x  Stat2: %08x  Stat3: %08x  Stat4: %08x \n",
	   IntStatus1, IntStatus2, IntStatus3, IntStatus4);
    printk("Stat5: %08x  Stat6: %08x  Stat7: %08x  Stat8: %08x \n",
	   IntStatus5, IntStatus6, IntStatus7, IntStatus8);
    printk("\n");
}

void tx3920_print_biureg(void)
{
    printk("tx3920 biureg:\n");
    printk("MemConfig0-4: %08x  %08x  %08x  %08x  %08x\n",
	   MemConfig0, MemConfig1, MemConfig2, MemConfig3, MemConfig4);
    printk("MemConfig5-8: %08x  %08x  %08x  %08x\n",
	   MemConfig5, MemConfig6, MemConfig7, MemConfig8);
    printk("\n");
}

void tx3920_print_ioreg(void)
{
    printk("tx3920 ioreg:\n");
    printk("IOCnt: %08x  DataInOut: %08x IOPD: %08x  MFIOPD: %08x\n",
	   IOControl, IODataInOut, IOPowerDown, MFIOPowerDown);
    printk("MFIO Out: %08x  Direc: %08x  In: %08x  Sel: %08x\n",
	   MFIOOutput, MFIODirection, MFIOInput, MFIOSelect);
    printk("\n");
}

void plum2_print_intreg(void)
{
    printk("plum2 intreg:\n");
    printk("IntStatus: %08x   IntEna: %08x\n",
	   Plum2_IntStatus, Plum2_IntEnable);
    printk("ExtInput - Status: %08x Status(M): %08x  Ena: %08x\n",
	   Plum2_ExtIntStatusRaw,
	   Plum2_ExtIntStatus,
	   Plum2_ExtIntEnable);
    printk("PCCard - Status: %08x Status(M): %08x  Ena: %08x Detect: %08x\n",
	   Plum2_PccIntStatusRaw,
	   Plum2_PccIntStatus,
	   Plum2_PccIntEnable,
	   Plum2_PccClkSelect);
    printk("UsbEna: %08x  SmartMediaEna: %08x  DispIntEna: %08x\n",
	   Plum2_UsbIntEnable,
	   Plum2_SmIntEnable,
	   Plum2_DispIntEnable);
    printk("\n");
}

void plum2_print_pcc(void)
{
#define PLUM2_PCC1REG_BASE	KSEG1ADDR(0x10805000)
#define PLUM2_PCC2REG_BASE	KSEG1ADDR(0x10805800)
#define PLUM2_PCC_POWERCTRL	0x008	/* Power Control */
#define PLUM2_PCC_GCTRL		0x078	/* Global Control */
#define PLUM2_PCC_FUNCCTRL	0x0f8	/* PCMCIA Function Control Register */
#define PLUM2_PCC_SLOTCTRL	0x100	/* Slot Control Register */
/* Slot Control Register */
#define PLUM2_PCC_SLOT_ON	BIT(7)
#define PLUM2_PCC_ID_LOCK	BIT(5)
#define PLUM2_PCC_ID_WP		BIT(4)

    printk("plum2 pccreg:\n");
    printk("GCtl: %08x\n",
	   *((volatile u32*)(PLUM2_PCC1REG_BASE + PLUM2_PCC_GCTRL)));
    printk("card0: PowCtrl: %08x, FuncCtrl: %08x, SlogCtrl: %08x\n",
	   *((volatile u32*)(PLUM2_PCC1REG_BASE + PLUM2_PCC_POWERCTRL)),
	   *((volatile u32*)(PLUM2_PCC1REG_BASE + PLUM2_PCC_FUNCCTRL)),
	   *((volatile u32*)(PLUM2_PCC1REG_BASE + PLUM2_PCC_SLOTCTRL)));
    printk("card1: PowCtrl: %08x, FuncCtrl: %08x, SlogCtrl: %08x\n",
	   *((volatile u32*)(PLUM2_PCC2REG_BASE + PLUM2_PCC_POWERCTRL)),
	   *((volatile u32*)(PLUM2_PCC2REG_BASE + PLUM2_PCC_FUNCCTRL)),
	   *((volatile u32*)(PLUM2_PCC2REG_BASE + PLUM2_PCC_SLOTCTRL)));
}

void plum2_print_power(void)
{
    printk("plum2 power:\n");
    printk("PowerCtrl: %08x, ClockCtrl: %08x\n",
	   Plum2_PowerCtrl,Plum2_ClockCtrl);
}

void tx3920_print_uartreg(void)
{
    int base;
    base = 0xb0;
    printk("tx3920 uartreg:\n");
    printk("UartA Ctrl1: %08x  RHR: %08x\n", REG_AT(base), REG_AT(base+0x14));
    base = 0xc8;
    printk("UartB Ctrl1: %08x   RHR: %08x\n", REG_AT(base), REG_AT(base+0x14));
    printk("\n");
}

void tx3920_print_IrDAreg(void)
{
    int base;
    base = 0xb0;
    printk("tx3920 IrDAreg:\n");
    printk("RingPtr: %08x  BaseAddr: %08x RingSize: %08x  Cfg0: %08x  Cfg1: %08x\n",
	   REG_AT(0x28), REG_AT(0x2c), REG_AT(0x30), REG_AT(0x38), 
	   REG_AT(0x3c));
    printk("Ena: %08x  RvCnt: %08x Status: %08x\n",
	   REG_AT(0x44), REG_AT(0x54), REG_AT(0x5c));
    printk("\n");
}

void tx3920_print_sibreg(void)
{
    printk("tx3920 sibreg:\n");
    printk("SIBControl: %08x  SIBDMAControl: %08x\n",
           SIBControl, SIBDMAControl);
    printk("\n");
}

void tx3920_print_chireg(void)
{
    printk("tx3920 chireg:\n");
    printk("CHIControl: %08x\n",
           CHIControl);
    printk("\n");
}

void tx3920_print_powreg(void)
{
    int base;
    base = 0xb0;
    printk("tx3920 PowCtrlreg:\n");
    printk("PowerCtroll: %08x\n", PowerControl);
    printk("\n");
}

void tx3920_print_clockreg(void)
{
    int base;
    base = 0xb0;
    printk("tx3920 clockreg:\n");
    printk("ClockCtrol: %08x\n", ClockControl);
    printk("\n");
}

void tx3920_print_spireg(void)
{
    printk("tx3920 spireg:\n");
    printk("SPICnt: %08x\n", SPIControl );
    printk("\n");
}

void print_mm_vma(struct mm_struct *mm)
{
    struct vm_area_struct *vma;

    vma = mm->mmap;
    printk("print_vma:\n");
    while( vma ){
	printk("%08lx-%08lx: vm_page_prot=%08lx vm_flags=%08lx\n",
	       vma->vm_start, vma->vm_end,
	       vma->vm_page_prot.pgprot, vma->vm_flags);
	vma = vma->vm_next;
    }
}

void print_vma(struct vm_area_struct *vma)
{
    printk("print_vma:\n");
    while( vma ){
	printk("%08lx-%08lx: vm_page_prot=%08lx vm_flags=%08lx\n",
	       vma->vm_start, vma->vm_end,
	       vma->vm_page_prot.pgprot, vma->vm_flags);
	vma = vma->vm_next;
    }
}

void print_vmlist(struct vm_struct *vmlist)
{
    printk("print_vmlist:\n");
    while( vmlist ){
	printk("%08x-%08x: flags=%08lx\n",
	       (unsigned)vmlist->addr, (unsigned)(vmlist->addr + vmlist->size),
	       vmlist->size);
	vmlist = vmlist->next;
    }
}

void print_pte(struct mm_struct *mm, unsigned start, unsigned end)
{
    unsigned i;
    pgd_t * dir;
    pmd_t *pmd;
    pte_t *pte;
    unsigned none, bad, pte_ent;
	
    if( !mm ){
	printk("mm=NULL, Use init_mm\n");
	mm = &init_mm;
    }


    start = PAGE_ALIGN(start);
    end = PAGE_ALIGN(end);

    printk("pte: %08x-%08x\n", start, end);
    for( i=start; i<end; i+=PAGE_SIZE ){
	dir = pgd_offset(mm, i);
	pmd = pmd_alloc(dir, i);

#if 0
	printk("%08x: %08x\n", i, pmd);
#else	
	if( pmd ){
	    bad = pmd_bad(*pmd);
	    none = pmd_none(*pmd);
	    pte = pte_offset(pmd, i);
	    if( pte )
		pte_ent = pte->pte;
	    else
		pte_ent = -2;
	    if( !none )
		printk("%08x: %08x%c%c [pgd head:%08x, pte head:%08x]\n",
		       i, pte_ent, (bad ? 'b' : '-'), (none ? 'n' : '-'),
		       (unsigned int)mm->pgd, (unsigned int)pmd_val(*pmd));
	}else{
	    bad = 1;
	    none = 1;
	    pte = 0;
	    pte_ent = 0;
	}
#endif
#if 0
	if( (i - start) % 6 == 0 )
	    printk("\n%08x:", i);
	printk(" %08x%c%c", pte_ent, (bad ? 'b' : '-'), (none ? 'n' : '-'));
#endif
    }
    printk("\n");
}
#endif /* DEBUG_PRINT */

#ifdef DEBUG_CT
#define TRACE_VECSIZ 128
#define STR_MAX 2
static int ct_ptr = 0;
static struct code_trace_box{
    int no;
    char str[STR_MAX];
}ct_vec[TRACE_VECSIZ];

void ct_set(int no, char *name)
{
    unsigned flags;

    save_and_cli(flags);
    ct_vec[ct_ptr].no = no;
    ct_vec[ct_ptr].str[0] = name[0];
    ct_vec[ct_ptr].str[1] = name[1];
    ct_ptr = (ct_ptr + 1) % TRACE_VECSIZ;
    restore_flags(flags);
}

void ct_print(void)
{
    static struct code_trace_box temp[TRACE_VECSIZ];
    int i, p;

    printk("ct_print:\n");
    do{
	p = ct_ptr;
	for( i=0; i<TRACE_VECSIZ; i++ ){
	    temp[i] = ct_vec[(p+i)%TRACE_VECSIZ];
	}
    }while(p != ct_ptr);
    for( i=0; i<TRACE_VECSIZ; i++ ){
	ct_vec[i].no = 0;
	ct_vec[i].str[0] = 0;
	ct_vec[i].str[1] = 0;
    }
    for( i=0; i<TRACE_VECSIZ; i++ ){
	if( i % 8 == 0 )
	    printk("\n");
#if 0
	printk(" [%2d:%2d-%c%c]", i, temp[i].no,
	       temp[i].str[0] ? temp[i].str[0] : '-',
	       temp[i].str[1] ? temp[i].str[1] : '-');
#else
	printk(" [%2d:%08x-%c%c]", i, temp[i].no,
	       temp[i].str[0] ? temp[i].str[0] : '-',
	       temp[i].str[1] ? temp[i].str[1] : '-');
#endif
    }
    printk("\n");
}
#endif /* DEBUG_CT */

#ifdef DEBUG_TELIOS
static void led_blink_handler(unsigned long dev_addr);

static struct timer_list led_blink = {
    {NULL, NULL}, 0UL, 0UL, led_blink_handler
};

static unsigned long led_count = 0;

void telios_led1(int sw)
{
	unsigned flags;

	save_and_cli(flags);
	if( sw )
		IODataInOut = (IODataInOut | 0x00100000);
	else
		IODataInOut = (IODataInOut & ~0x00100000);
	restore_flags(flags);
}

void telios_led2(int sw)
{
	unsigned flags;

	save_and_cli(flags);
	if( sw )
		IODataInOut = (IODataInOut | 0x00080000);
	else
		IODataInOut = (IODataInOut & ~0x00080000);
	restore_flags(flags);
}

static void led_blink_handler(unsigned long dev_addr)
{
	telios_led1((led_count++)&1);
	mod_timer(&led_blink, jiffies + HZ);
}

static int __init telios_led_blink_init(void)
{
	init_timer(&led_blink);
	led_blink.expires = jiffies + HZ;
	led_blink.function = led_blink_handler;
	add_timer(&led_blink);
	printk("Debug: LED start\n");
	return 0;
}
__initcall(telios_led_blink_init);
#endif

#ifdef DEBUG_TELIOS
#define TELIOS_BUTTON_IRQ	30

#if 0
static void debug_alarm_handler(void *arg, struct pt_regs *regs)
{
    ct_print();
    tx3920_print_intreg();
}
#endif

static void
telios_button_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    u32 status;

    status = IntStatus8;

    if(status & INT8_IOPOSINT7){
	IntClear8 = INT8_IOPOSINT7;
	printk("Pushed Surfing SW\n");
	{
	    extern void show_stack(unsigned int *sp);
	    extern void show_stack(unsigned int *sp);
	    extern void show_trace(unsigned int *sp);
	    extern void show_code(unsigned int *pc);

	    show_regs(regs);
	    printk("Process %s (pid: %d, stackpage=%08lx)\n",
		   current->comm, current->pid, (unsigned long) current);
	    show_stack((unsigned int *) regs->regs[29]);
	    show_trace((unsigned int *) regs->regs[29]);
	    show_code((unsigned int *) regs->cp0_epc);
	    printk("\n");
	    {
		extern struct vm_struct * vmlist;
		struct task_struct *tsk = current;
		printk("init_mm:\n");
		print_mm_vma(&init_mm);
		if( vmlist ){
		    printk("vmlist:\n");
		    print_vmlist(vmlist);
		}
		if( tsk->mm ){
		    printk("current->mm:\n");
		    print_mm_vma(tsk->mm);
		}
		/*		print_pte(tsk->mm, 0xc0000000, 0xe0000000);*/
	    }
	}
    }
    if(status & INT8_IOPOSINT8){
	IntClear8 = INT8_IOPOSINT8;
	printk("Pushed Mail recipt SW\n");
#ifdef DEBUG_PRINT
	ct_print();
	// REG_AT(0x3c) = 0x0000;	/* clear  IrDA config1 register */
	tx3920_print_intreg();
	//	tx3920_print_CP0REG();
	tx3920_print_ioreg();
	// tx3920_print_spireg();
	// plum2_print_power();
	tx3920_print_powreg();
	tx3920_print_clockreg();
	tx3920_print_uartreg();
	// tx3920_print_IrDAreg();
	// tx3920_print_sibreg();
	// IntClear3 |= BIT(12);
	// IntClear4 |= BIT(12);
	// plum2_print_intreg();
	// plum2_print_pcc();
	// tx3920_print_CP0REG();
	//	BackLight_VR ^= BKLGT_VR_VR0;
	// plum2lcd_backlight_dump();
#endif
#if 0
	tx3920_print_IrDAreg();
	REG_AT(0x44) = 0x0000;	/* clear PhyAndClock*/
	REG_AT(0x50) = 0x0400;	/* dummy (MaximumPacketLength)*/
	// REG_AT(0x38) = 0x04a0;	/* dummy (config0) */
	REG_AT(0x38) = 0x04a0;	/* dummy (config0) */
	REG_AT(0x4c) = 0x0000;	/* dummy (InfraerdPhyConfig) */
	REG_AT(0x44) = 0x8000;	/* set   PhyAndClock*/
	tx3920_print_IrDAreg();
	tx3920_print_sibreg();
	tx3920_print_chireg();
	tx3920_print_uartreg();
#endif
    }
}


static struct irqaction irq30 = {telios_button_interrupt, SA_INTERRUPT, 0,
 			 "Telios Button", NULL, NULL};

static int __init telios_button_init(void)
{
    printk("telios button initialized.\n");

    return setup_tx3920_irq(TELIOS_BUTTON_IRQ, &irq30);
}
__initcall(telios_button_init);
#endif

#ifdef DEBUG_TELIOS
#define TELIOS_DC_JACK_IRQ	31
static void
telios_dc_jack_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    u32 status;

    status = IntStatus8;

    IntClear8 = INT8_IONEGINT6 | INT8_IOPOSINT6;

    if( status & INT8_IONEGINT6 ){
	if( status & INT8_IOPOSINT6 )
	    printk("DC JACK IN/OUT\n");
	else
	    printk("DC JACK IN\n");
    }else if( status & INT8_IOPOSINT6 )
	printk("DC JACK OUT\n");
}

static struct irqaction irq31 = {telios_dc_jack_interrupt, SA_INTERRUPT, 0,
 			 "Telios DC JACK", NULL, NULL};

static int __init telios_dc_jack_init(void)
{
    
    return setup_tx3920_irq(TELIOS_DC_JACK_IRQ, &irq31);
}
__initcall(telios_dc_jack_init);
#endif

void sharp_telios_map_bklgt_io(void)
{
    if( !bklight_base )
	if( kseg2_page_alloc(BKLGT_BASE, 0x200, KSEG2_FLAGS_ALL, &bklight_base) ){
	    printk("Not alloc kseg2 page for KeyGA.\n");
	    return;
	}
}

void __init sharp_telios_setup(void)
{
    MemConfig0 |= MEM0_ENCS1USER;		/* enable KUSEG CS1 */
}
