/* $Id: sib.c,v 1.1.2.3 2001/05/07 05:23:58 fuku Exp $
 *
 * SIB - Routines for Serial Interconnect Bus on r39xx
 *       (used for sound, telecom, touchscreen, and other interfaces)
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2000 Jim Pick <jim@jimpick.com>
 */

#include <linux/wait.h>
#include <linux/sched.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/tx3920.h>
#include <asm/betty.h>
#include "sib.h"


#define SIB_WRITEBUFFER_SIZE   256
static struct sib_write_struct writebuffer[ SIB_WRITEBUFFER_SIZE ];
static struct sib_write_struct *writebuffer_head = writebuffer; /* Write here */
static struct sib_write_struct *writebuffer_tail = writebuffer; /* Read here */
static int writebuffer_count = 0; /* Number of writes in buffer */
static void (*next_read_callback)(void) = NULL;

/*
 *  Interrupt routine
 */

static void sib_interrupt_sf0(int irq, void *dev_id,
					 struct pt_regs * regs)
{
      	IntEnable1 &= ~INT1_SIBSF0INT;
	IntClear1 = INT1_SIBSF0INT;

	//    printk(__FUNCTION__"()\n");
	
	if (next_read_callback != NULL) {
	    /*		printk("next_read_callback: %p\n", next_read_callback); */
		next_read_callback();
	}

	if (writebuffer_count == 0) {
               	SIBControl &= ~(SIB_ENSIB | SIB_ENSF0);
		IntEnable1 &= ~INT1_SIBSF0INT;
		IntClear1 = INT1_SIBSF0INT;

		next_read_callback = NULL;
	}
	else {
	  //		printk( "Out: 0x%08x\n", writebuffer_tail->data ); 
		SIBSubFrame0Control = writebuffer_tail->data;
		next_read_callback = writebuffer_tail->next_read_callback;
		writebuffer_tail++;
		if (writebuffer_tail == writebuffer + SIB_WRITEBUFFER_SIZE)
			writebuffer_tail = writebuffer;
		writebuffer_count--;
/*
		if (writebuffer_count == SIB_WRITEBUFFER_SIZE - 1)
			wake_up_interruptible(&sib_write_wait);
*/
		
		IntEnable1 |= INT1_SIBSF0INT;  
		IntClear1 = INT1_SIBSF0INT;
	}

	//   printk(__FUNCTION__"() is succesfull\n");
}


/*
 *  Initialization routine
 */

void sib_init(void)
{
	int i;
	int flags;

	save_and_cli(flags);
#ifdef CONFIG_VTECH_HELIO
	/* Reset Betty (wired to IR RXPWR pin?) */

	IRControl1 |= IR_RXPWR;      /* On */
	for (i = 0; i < 16; i++);

	IRControl1 &= ~IR_RXPWR;     /* Off */
	for (i = 0; i < 256; i++);

	IRControl1 |= IR_RXPWR;      /* On */
	for (i = 0; i < 16; i++);
#endif
	/*	PowerControl |= PWR_DIVMOD;*/  /* Peripheral clock speeds are
                                        independent from CPU speed */

	SIBControl &= ~SIB_SIBSCLKDIV_MASK;

	ClockControl &= ~CLK_SIBMCLKDIV_MASK;
	ClockControl |= CLK_SIBMCLKDIR | CLK_ENSIBMCLK |
		(2 << CLK_SIBMCLKDIV_SHIFT) | CLK_CSERSEL;

      	SIBControl &= ~(SIB_ENSIB | SIB_ENSF0);
	
	IntEnable1 &= ~INT1_SIBSF0INT;
	IntClear1 = INT1_SIBSF0INT;

	i = request_irq(19, sib_interrupt_sf0, SA_INTERRUPT, "sib sf0", NULL);
	if ( i != 0 ) {
	  printk(KERN_ERR "sib: Cannot allocate irq for SF0. (%d)\n",
		       i);
	}
	restore_flags(flags);
}


/*
 *  Send data on SIB in subframe 0 (subframe 1 is unused?)
 */

static int sib_command(int reg, unsigned int data, void (*next_read_callback)(void))
{
	 unsigned flags;

	 save_and_cli(flags);
	/* If the buffer is full, then return failure */
        if (writebuffer_count == SIB_WRITEBUFFER_SIZE){
	  printk("buffer is full!\n");
	  return -1;
        }
	
        writebuffer_head->data = data | 
		((reg << SIB_ADDRESS_SHIFT) & SIB_ADDRESS_MASK);
	writebuffer_head->next_read_callback = next_read_callback;
	writebuffer_head++;
	if (writebuffer_head == writebuffer + SIB_WRITEBUFFER_SIZE)
		writebuffer_head = writebuffer;
	writebuffer_count++;
#if 0
	SIBControl |= SIB_ENSIB | SIB_ENSF0 ;
	
	IntEnable1 |= INT1_SIBSF0INT;
	IntClear1 = INT1_SIBSF0INT;
#else
	if( !(SIBControl & SIB_ENSF0) ){
	    IntClear1 = INT1_SIBSF0INT;
	    IntEnable1 |= INT1_SIBSF0INT;
	    SIBControl |= SIB_ENSIB | SIB_ENSF0 ;
	}
#endif
	restore_flags(flags);
	return 0;
}

/*
 *  Send data on SIB in subframe 0 (subframe 1 is unused?)
 */

int sib_write(int reg, unsigned int data)
{
  /*	printk( "W: %d, 0x%04x\n", reg, data ); */
  //  printk(__FUNCTION__"()\n");
	return sib_command( reg, data | SIB_WRITE, NULL );
}

void sib_write_check(int reg, unsigned int data)
{
  //  printk(__FUNCTION__"()\n");

	if (sib_write( reg, data ))
		printk("SIB buffer full - command lost.\n");
}


/*
 *  Send data on SIB in subframe 0 and wait for returned data
 */

int sib_write_read(int reg, unsigned int data,
			      void (*next_read_callback)(void))
{
  //  printk(__FUNCTION__"()\n");
  /*	printk( "WR: %d, 0x%04x %p\n", reg, data, next_read_callback ); */
	return sib_command( reg, data | SIB_WRITE, next_read_callback );
}

int sib_read(int reg, void (*next_read_callback)(void))
{
  //  printk(__FUNCTION__"()\n");
  /*	printk( "R: %d %p\n", reg, next_read_callback ); */
	return sib_command( reg, 0, next_read_callback );
}

void sib_read_check(int reg, void (*next_read_callback)(void))
{
  //  printk(__FUNCTION__"()\n");
       if (sib_read( reg, next_read_callback ))
		printk("SIB buffer full - command lost.\n");
}
