/*
 * $Id: time.c,v 1.5 2000/07/14 18:07:04 harald Exp $
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1999  Harald Koerfgen
 *
 * This file contains the time handling details for TMPR3912/pr31700
 * based PDAs
 *
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/timex.h>

#include <asm/tx3920.h>

/* XXX */
#include <linux/delay.h>

extern volatile unsigned long wall_jiffies;
extern rwlock_t xtime_lock;

#define USECS_PER_JIFFY (1000000/HZ)

/* This function must be called with interrupts disabled 
 * It was inspired by Steve McCanne's microtime-i386 for BSD.  -- jrs
 * 
 * If you are really that interested, you should be reading
 * comp.protocols.time.ntp!
 */

#define TICK_SIZE tick

/*
 * Poll the Interrupt Status Registers
 */
#undef POLL_STATUS

static unsigned long do_gettimeoffset(void)
{
    /*
     * This is a kludge
     */
    return 0;
}

/*
 * This version of gettimeofday has near microsecond resolution.
 */
void do_gettimeofday(struct timeval *tv)
{
    unsigned long flags;

    read_lock_irqsave(&xtime_lock, flags);
    *tv = xtime;
    tv->tv_usec += do_gettimeoffset();

    /*
     * xtime is atomically updated in timer_bh. lost_ticks is
     * nonzero if the timer bottom half hasnt executed yet.
     */
    if (jiffies - wall_jiffies)
	tv->tv_usec += USECS_PER_JIFFY;

    read_unlock_irqrestore(&xtime_lock, flags);

    if (tv->tv_usec >= 1000000) {
	tv->tv_usec -= 1000000;
	tv->tv_sec++;
    }
}

void do_settimeofday(struct timeval *tv)
{
    write_lock_irq(&xtime_lock);
    /* This is revolting. We need to set the xtime.tv_usec
     * correctly. However, the value in this location is
     * is value at the last tick.
     * Discover what correction gettimeofday
     * would have done, and then undo it!
     */
    tv->tv_usec -= do_gettimeoffset();

    if (tv->tv_usec < 0) {
	tv->tv_usec += 1000000;
	tv->tv_sec--;
    }
    xtime = *tv;
    time_state = TIME_BAD;
    time_maxerror = MAXPHASE;
    time_esterror = MAXPHASE;
    write_unlock_irq(&xtime_lock);
}

static int set_rtc_mmss(unsigned long nowtime)
{
    int retval = 0;

    return retval;
}

/* last time the cmos clock got updated */
static long last_rtc_update = 0;

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick
 */

int do_write = 1;

static void
timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    IntClear5 = INT5_PERINT;			/* clear the interrupt */

#ifdef POLL_STATUS
    static unsigned long old_IntStatus1 = 0;
    static unsigned long old_IntStatus3 = 0;
    static unsigned long old_IntStatus4 = 0;
    static unsigned long old_IntStatus5 = 0;
    static int counter = 0;
    int i;

#if 0
    if (IntStatus1 != old_IntStatus1) {
	printk("IntStatus1 changed: 0x%08x\n", old_IntStatus1 ^ IntStatus1);
	IntClear1 = 0xffffffff;
	old_IntStatus1 = IntStatus1;
    }
    if (IntStatus4 != old_IntStatus4) {
	printk("IntStatus4 changed: 0x%08x\n", old_IntStatus4 ^ IntStatus4);
	IntClear4 = old_IntStatus4 ^ IntStatus4;
	old_IntStatus4 = IntStatus4;
    }
    if ((IntStatus3 & 0xb800204d) != old_IntStatus3) {
	printk("IntStatus3 changed: 0x%08x\n", old_IntStatus3 ^ (IntStatus3 & 0xb800204d));
	old_IntStatus3 = (IntStatus3 & 0xb800204d);
    }
    if ((IntStatus5 & 0x3fff) != old_IntStatus5) {
	printk("IntStatus5 changed: 0x%08x\n", old_IntStatus5 ^ (IntStatus5 & 0x3fff));
	old_IntStatus5 = (IntStatus5 & 0x3fff);
    }
#endif
    new_spircv = SPIData & 0xff;
    if ((old_spircv != new_spircv) && (new_spircv != 0xff)) {
	    printk( "SPIData changed: %x\n", new_spircv );
    }
    old_spircv = new_spircv;
    if (do_write)
	    SPIData = 0;
#endif
#ifdef CONFIG_PHILIPS_VELO
    kbd_scan();
#endif

    if (!user_mode(regs)) {
	if (prof_buffer && current->pid) {
	    extern int _stext;
	    unsigned long pc = regs->cp0_epc;

	    pc -= (unsigned long) &_stext;
	    pc >>= prof_shift;
	    /*
	     * Dont ignore out-of-bounds pc values silently,
	     * put them into the last histogram slot, so if
	     * present, they will show up as a sharp peak.
	     */
	    if (pc > prof_len - 1)
		pc = prof_len - 1;
		atomic_inc((atomic_t *) & prof_buffer[pc]);
	    }
    }

    /*
     * aaaand... action!
     */
    do_timer(regs);

    /*
     * If we have an externally syncronized Linux clock, then update
     * CMOS clock accordingly every ~11 minutes. Set_rtc_mmss() has to be
     * called as close as possible to 500 ms before the new second starts.
     */
    if (time_state != TIME_BAD && xtime.tv_sec > last_rtc_update + 660 &&
	xtime.tv_usec > 500000 - (tick >> 1) &&
	xtime.tv_usec < 500000 + (tick >> 1))
	if (set_rtc_mmss(xtime.tv_sec) == 0)
	    last_rtc_update = xtime.tv_sec;
	else
	    last_rtc_update = xtime.tv_sec - 600;	/* do it again in 60 s */
}

/* Converts Gregorian date to seconds since 1970-01-01 00:00:00.
 * Assumes input in normal date format, i.e. 1980-12-31 23:59:59
 * => year=1980, mon=12, day=31, hour=23, min=59, sec=59.
 *
 * [For the Julian calendar (which was used in Russia before 1917,
 * Britain & colonies before 1752, anywhere else before 1582,
 * and is still in use by some communities) leave out the
 * -year/100+year/400 terms, and add 10.]
 *
 * This algorithm was first published by Gauss (I think).
 *
 * WARNING: this function will overflow on 2106-02-07 06:28:16 on
 * machines were long is 32-bit! (However, as time_t is signed, we
 * will already get problems at other places on 2038-01-19 03:14:08)
 */
static inline unsigned long mktime(unsigned int year, unsigned int mon,
				   unsigned int day, unsigned int hour,
				   unsigned int min, unsigned int sec)
{
    if (0 >= (int) (mon -= 2)) {	/* 1..12 -> 11,12,1..10 */
	mon += 12;		/* Puts Feb last since it has leap day */
	year -= 1;
    }
    return ((((unsigned long) (year / 4 - year / 100 + year / 400 + 367 * mon / 12 + day) +
		 year * 365 - 719499
	     ) * 24 + hour	/* now have hours */
	    ) * 60 + min	/* now have minutes */
	) * 60 + sec;		/* finally seconds */
}

struct irqaction irq0 = {timer_interrupt, SA_INTERRUPT, 0,
 			 "timer", NULL, NULL};

void (*board_time_init) (struct irqaction * irq);

void __init time_init(void)
{
    unsigned int year, mon, day, hour, min, sec;

	sec = 0;
	min = 0;
	hour = 0;
	day = 1;
	mon = 1;
	year = 2000;

    xtime.tv_sec = mktime(year, mon, day, hour, min, sec);
    xtime.tv_usec = 0;

    board_time_init(&irq0);
}
