/* $Id: tpanel.c,v 1.1.2.8 2002/01/09 14:01:25 fuku Exp $
 *
 * tpanel - Routines for Touch Panel on tx39xx
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2000 Jim Pick <jim@jimpick.com>
 * Copyright (C) 2001 AXE, Inc.  (e-mail: <fuku@axe-inc.co.jp>)
 *
 * Based on arch/mips/r39xx/tpanel.c
 * Some parts taken from arch/mips/vr41xx/tpanel.c
 *
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/init.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/tx3920.h>
#include <asm/betty.h>
#include <linux/tpanel.h>
#include <linux/delay.h>
#include "sib.h"

#ifdef CONFIG_SHARP_TELIOS
#include "sharp_telios.h"
#endif

/* #define SKIP_MEAS_PRESSURE */
#define ONLY_PXPY	1
#define XMYM		2
#define MEAS3		3
#define MEASURE_MODE	ONLY_PXPY
/* #define MEASURE_MODE	XMYM */
/* #define MEASURE_MODE	MEAS3 */

#if MEASURE_MODE == ONLY_PXPY
#define SAMPLE_MAX 1
#elif MEASURE_MODE == XMYM
#define SAMPLE_MAX 2
#elif MEASURE_MODE == MEAS3
#define SAMPLE_MAX 3
#endif

#define DEBUG_USE_IOCTL
#ifdef DEBUG_USE_IOCTL
#define TPGET_ADCWAIT	0x4680
#define TPSET_ADCWAIT	0x4681
#define TPGET_ADCWAITNT	0x4682
#define TPSET_ADCWAITNT	0x4683
#define TPGET_PRESSMIN	0x4684
#define TPSET_PRESSMIN	0x4685
#define TPGET_PRESSCHK	0x4686
#define TPSET_PRESSCHK	0x4687
#define TPGET_SAMPRATE	0x4688
#define TPSET_SAMPRATE	0x4689
#endif

/* pen_state */
#define MEASURE_PENDOWN         0
#define MEASURE_PX              1
#define MEASURE_MX              2
#define MEASURE_PY              3
#define MEASURE_MY              4
#define MEASURE_PRESSURE        5
#define MEASURE_PRESS           6

static int pen_state = MEASURE_PENDOWN;
static int sample_counter;

static wait_queue_head_t tpanel_wait;

#define BUFFSIZE 128

/* #define TPADC_WAIT_FRAME 72 */

#define TPADC_WAIT_FRAME	95		/* 9.5msec */
#define TPADC_WAIT_NEXT_FRAME	TPADC_WAIT_FRAME
#define TP_VALID_PRESSURE_MIN	0
#define TP_SAMPLE_RATE_MIN	0		/* min */

extern int tx3920_alarm_set_handler(u32 count,
				    void (*function)(void *arg, struct pt_regs *regs),
				    void *arg);

struct tpanel_config {
    unsigned int adc_wait;
    unsigned int adc_wait_next;
    unsigned int pressure_adc_wait;
    unsigned int press_check_wait;
    unsigned int pressure_min;
    unsigned int pressure_check;
    unsigned int sample_rate;
}tpanel_conf ={
    TPADC_WAIT_FRAME,
    TPADC_WAIT_NEXT_FRAME,
    50,
    30,
    TP_VALID_PRESSURE_MIN,
    1,
    TP_SAMPLE_RATE_MIN,
};


struct tpanel_sample {
  unsigned int state;
  unsigned int x;
  unsigned int y;
  unsigned int mx;
  unsigned int my;
  unsigned int pressure;
  unsigned int lost_count;
};

static struct tpanel_sample samplebuf[BUFFSIZE];
static struct tpanel_sample *samplebuf_head = samplebuf;
static struct tpanel_sample *samplebuf_tail = samplebuf;

static struct fasync_struct *fasyncptr;

static u32 last_sample_time;

/* Prototypes */
static void tpanel_getxy(void);

static void enable_sibposint(void)
{
  IntClear1 = INT1_SIBIRQPOSINT;
  IntEnable1 |= INT1_SIBIRQPOSINT;
}

static void tpanel_interrupt_mode(void)
{
  sib_read_check(BETTY_REG_ID, enable_sibposint);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 
		  BETTY_TSPX_INT|BETTY_TSMX_INT);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
  sib_write_check(BETTY_REG_RISING_INT_ENABLE, 
		  BETTY_TSPX_RIS_INT | BETTY_TSMX_RIS_INT);

  sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
		  BETTY_TSMX_POW | BETTY_TSPX_POW |
		  BETTY_TSMY_GND | BETTY_TSPY_GND |
		  (0 << BETTY_TSC_MODE_SHIFT));
}

static void tpanel_check_press_mode(void)
{
  sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
		  BETTY_TSMX_POW | BETTY_TSPX_POW |
		  BETTY_TSMY_GND | BETTY_TSPY_GND |
		  (0 << BETTY_TSC_MODE_SHIFT));
}

static void tpanel_disable_interrupt(void)
{
  sib_write_check(BETTY_REG_RISING_INT_ENABLE, 0);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, ~0);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
}

static void adc_disable(void)
{
  sib_write_check(BETTY_REG_ADC_CTRL, 0);
}

static void (*alarm_sib_tail_fun)(void *arg, struct pt_regs *regs)=NULL;
static int alarm_count;

static void set_alarm_sib_tail(void)
{
    int ret;

    ret = tx3920_alarm_set_handler(alarm_count, alarm_sib_tail_fun, NULL);
    alarm_sib_tail_fun = NULL;
    /*
    if( ret != 0 ){
	printk("tpanel.c: tx3920_alarm_set_handler() {frame} error (%08x)\n", ret);
    }
    */
}

#define conv_30u_to_rtc(c)	((c) * RTC_PERSEC_COUNT / 33333)
#define conv_100u_to_rtc(c)	((c) * RTC_PERSEC_COUNT / 10000)
#define conv_100u_to_sib(c)	((c) * 72000 / 10000)
#define conv_sib_to_rtc(frame)  ((frame) * RTC_PERSEC_COUNT / 72000)
    				/* SIB frame rate = 72kHz (maybe) */
#define conv_rtc_to_30u(rtc)	((rtc) * 33333 / RTC_PERSEC_COUNT)
#define conv_rtc_to_100u(rtc)	((rtc) * 10000 / RTC_PERSEC_COUNT)
#define conv_rtc_to_sib(rtc)    ((rtc) * 72000 / RTC_PERSEC_COUNT)
#define conv_sib_to_100u(rtc)	((rtc) * 10000 / 72000)

static void set_alarm_by_100u(unsigned c, void (*callback)(void))
{
    alarm_count = conv_100u_to_rtc(c);
    alarm_sib_tail_fun = (void (*)(void *arg, struct pt_regs *regs))callback;
    sib_read_check(BETTY_REG_ID, set_alarm_sib_tail);
}

static void tpanel_readx_adcstart(void)
{
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_TSPY <<
		   BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA |
		  BETTY_ADC_START);
  sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
}

static void tpanel_readx_mode(void)
{
  sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
		  BETTY_TSPX_POW | BETTY_TSMX_GND |
		  (2 << BETTY_TSC_MODE_SHIFT) |
		  BETTY_TSC_BIAS_ENA);
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_TSPY <<
		   BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA);

  set_alarm_by_100u(tpanel_conf.adc_wait, tpanel_readx_adcstart);
}

static void tpanel_readx_adc_mode(void)
{
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_TSPY <<
		   BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA);

  set_alarm_by_100u(tpanel_conf.adc_wait_next, tpanel_readx_adcstart);
}

static void tpanel_ready_adcstart(void)
{
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_TSPX << BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA |
		  BETTY_ADC_START);
  sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
}

static void tpanel_ready_mode(void)
{
  sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
		  BETTY_TSPY_POW | BETTY_TSMY_GND |
		  (2 << BETTY_TSC_MODE_SHIFT) |
		  BETTY_TSC_BIAS_ENA);
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_TSPX <<
		   BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA);

  set_alarm_by_100u(tpanel_conf.adc_wait, tpanel_ready_adcstart);
}

static void tpanel_ready_adc_mode(void)
{
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_TSPX <<
		   BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA);

  set_alarm_by_100u(tpanel_conf.adc_wait_next, tpanel_ready_adcstart);
}

static void tpanel_readp_adcstart(void)
{
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_AD0 <<
		   BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA |
		  BETTY_ADC_START);
  sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
}


static void tpanel_readp_mode(void)
{
  sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
		  BETTY_TSPX_POW | BETTY_TSMX_POW |
		  (2 << BETTY_TSC_MODE_SHIFT) |
		  BETTY_TSC_BIAS_ENA);
  sib_write_check(BETTY_REG_ADC_CTRL,
		  (BETTY_ADC_INPUT_AD0 <<
		   BETTY_ADC_INPUT_SHIFT) |
		  BETTY_ADC_SYNC_ENA |
		  BETTY_ADC_ENA);

  set_alarm_by_100u(tpanel_conf.pressure_adc_wait, tpanel_readp_adcstart);
}

static void tpanel_timer_handler(void *arg, struct pt_regs *regs)
{
  if(pen_state == MEASURE_PRESS)
    sib_read_check(BETTY_REG_TOUCHSCREEN_CTRL, tpanel_getxy);
  else
      printk("tpanel_timer_handler: bad pen state = %d\n", pen_state);
}


static inline int get_xy(int *lst)
{
    int a, b, c;

    /*    return lst[0];*/
    if( lst[0] > lst[1] )
	a = lst[0] - lst[1];
    else
	a = lst[1] - lst[0];
    if( lst[1] > lst[2] )
	b = lst[1] - lst[2];
    else
	b = lst[2] - lst[1];
    if( lst[0] > lst[2] )
	c = lst[0] - lst[2];
    else
	c = lst[2] - lst[0];
    if( a < b ){
	if( a < c)
	    return (lst[0] + lst[1]) / 2;
	else
	    return (lst[0] + lst[2]) / 2;
    }else{
	if( b < c)
	    return (lst[1] + lst[2]) / 2;
	else
	    return (lst[0] + lst[2]) / 2;
    }
}

#ifdef CONFIG_TPEV
static void send_vrtpanel_raw_event(struct tpanel_sample *sample)
{
    extern void vrtp_report_event(int pen_state, int lost,
				  int x, int y, int mx, int my,
				  int pressure);
    int lost, state;

    state = sample->state;
    switch(state){
    case 0:	/* pen up */
    case 1:	/* pen down */
	vrtp_report_event(state, 0, 0, 0, 0, 0, 0);
	break;
    case 2:
	lost = sample->lost_count;
	if( lost > 0xff )
		lost = 0xff;
	vrtp_report_event(state, lost,
			      sample->x, sample->y, sample->mx, sample->my,
			      sample->pressure);
      break;
    }
}
#endif

static void tpanel_put_sample(int state, int *lstx, int *lsty, int pressure,
			      int lost)
{
  samplebuf_head->state = state;
  samplebuf_head->lost_count = lost;

#if MEASURE_MODE == ONLY_PXPY
  samplebuf_head->x = lstx[0];
  samplebuf_head->y = lsty[0];
  samplebuf_head->pressure = pressure;
  samplebuf_head->mx = 1023 - samplebuf_head->x;
  samplebuf_head->my = 1023 - samplebuf_head->y;
#elif MEASURE_MODE == XMYM
  samplebuf_head->x = lstx[0];
  samplebuf_head->y = lsty[0];
  samplebuf_head->pressure = pressure;
  samplebuf_head->mx = 1023 - lstx[1];
  samplebuf_head->my = 1023 - lsty[1];
#elif MEASURE_MODE == MEAS3
  samplebuf_head->x = get_xy(lstx);
  samplebuf_head->y = get_xy(lsty);
  samplebuf_head->pressure = pressure;
  samplebuf_head->mx = 1023 - samplebuf_head->x;
  samplebuf_head->my = 1023 - samplebuf_head->y;
#endif

#ifdef CONFIG_TPEV
  send_vrtpanel_raw_event(samplebuf_head);
#endif
      
  samplebuf_head++;
  if (samplebuf_head == samplebuf + BUFFSIZE)
    samplebuf_head = samplebuf;
      
  if (samplebuf_head == samplebuf_tail) {	
    samplebuf_tail++;
    if (samplebuf_tail == samplebuf + BUFFSIZE)
      samplebuf_tail = samplebuf;
  }
}

static void tpanel_getxy(void)
{
  unsigned long value;
  static int sample_x[SAMPLE_MAX], sample_y[SAMPLE_MAX];
  static int panel_press, sample_index, sample_pressure;
  static int lost_count=0;
  

  switch( pen_state ) {

  case MEASURE_PENDOWN:
    tpanel_disable_interrupt();

    /* set X position measurement mode*/
    tpanel_readx_mode();

    sample_index = 0;
    sample_counter = 0;
    panel_press = 0;

    tx3920_read_rtc(value, last_sample_time);
    pen_state = MEASURE_PX;
    break;

  case MEASURE_PX:
    value = SIBSubFrame0Status;

    if (! (value & BETTY_ADC_DAT_VAL)) {
	if (++sample_counter < 60){
	sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
      }else {				
	/* Put touch screen back into
	   "interrupt mode" */
	tpanel_interrupt_mode();
	
	pen_state = MEASURE_PENDOWN;
      }
      break;
    }

    sample_x[sample_index] = (value & BETTY_ADC_DATA_MASK) >>
				BETTY_ADC_DATA_SHIFT;
    /*    adc_disable();*/
    if( sample_index++ < SAMPLE_MAX ){
	tpanel_readx_adc_mode();

	sample_counter = 0;
	break;
    }

    /* set Y position measurement mode*/
    tpanel_ready_mode();

    sample_index = 0;
    sample_counter = 0;

    pen_state = MEASURE_PY;
    break;

  case MEASURE_PY:
    value = SIBSubFrame0Status;

    if (! (value & BETTY_ADC_DAT_VAL)) {
      if (++sample_counter < 60){
	  sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
      }else {				
	/* Put touch screen back into
	   "interrupt mode" */
	tpanel_interrupt_mode();
	
	pen_state = MEASURE_PENDOWN;
      }
      break;
    }

    sample_y[sample_index] = (value & BETTY_ADC_DATA_MASK) >>
				BETTY_ADC_DATA_SHIFT;
    /*    adc_disable();*/
    if( sample_index++ < SAMPLE_MAX ){
	tpanel_ready_adc_mode();
	sample_counter = 0;
	//	printk("y:%d\n", sample_index);
	break;
    }

#ifdef SKIP_MEAS_PRESSURE
    sample_pressure = 900;

    if ( panel_press ){
      tpanel_put_sample(2, sample_x, sample_y, sample_pressure, 0);
    } else {
      tpanel_put_sample(1, sample_x, sample_y, sample_pressure, 0);
      panel_press = 1;
    }

    wake_up_interruptible(&tpanel_wait);
      
    if (fasyncptr)
      kill_fasync(&fasyncptr, SIGIO, POLL_IN);

    adc_disable();
    tpanel_check_press_mode();

    pen_state = MEASURE_PRESS;
    {
	int ret;
	u32 tm_low, cnt, rate_min, tp_wait;

	tx3920_read_rtc(value, tm_low);
	rate_min = conv_100u_to_rtc(tpanel_conf.sample_rate);
	tp_wait = conv_100u_to_rtc(tpanel_conf.press_check_wait);
	cnt = tm_low - last_sample_time;
	if( cnt >= rate_min - tp_wait ){
	    cnt = tp_wait;
	}else{
	    cnt = rate_min - cnt;
	}
	ret = tx3920_alarm_set_handler(cnt, tpanel_timer_handler, NULL);
	if( ret != 0 )
	    printk("tpanel.c: tx3920_alarm_set_handler() {press} error (%08x)\n", ret);
	last_sample_time = tm_low;
    }

#else /* !SKIP_MEAS_PRESSURE */
    /* set Pressure measurement mode*/
#ifdef CONFIG_SHARP_TELIOS
    if( tpanel_conf.pressure_check )
      BackLight_Delay |= BKLGT_DELAY_IO2;
#endif
    tpanel_readp_mode();
    
    pen_state = MEASURE_PRESSURE;
#endif
    break;

  case MEASURE_PRESSURE:
    value = SIBSubFrame0Status;

    if (! (value & BETTY_ADC_DAT_VAL)) {
      if (++sample_counter < 60){
	sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
      }else {				
#ifdef CONFIG_SHARP_TELIOS
	if( tpanel_conf.pressure_check )
	  BackLight_Delay &= ~BKLGT_DELAY_IO2;
#endif
	tpanel_interrupt_mode();
	
	pen_state = MEASURE_PENDOWN;
      }
      break;
    }

#ifdef CONFIG_SHARP_TELIOS
    if( tpanel_conf.pressure_check )
      BackLight_Delay &= ~BKLGT_DELAY_IO2;
#endif
    sample_pressure = (value & BETTY_ADC_DATA_MASK) >> BETTY_ADC_DATA_SHIFT;

    /* emulate pressure value of VR41xx touch panel driver */
    sample_pressure = (sample_pressure >> 1) + 550;

    if( sample_pressure < tpanel_conf.pressure_min ){
	lost_count++;
    }else{
	if ( panel_press ){
	    tpanel_put_sample(2, sample_x, sample_y, sample_pressure,
			      lost_count);
	} else {
	    tpanel_put_sample(1, sample_x, sample_y, sample_pressure,
			      lost_count);
	    panel_press = 1;
	}
	lost_count = 0;

	wake_up_interruptible(&tpanel_wait);
      
	if (fasyncptr)
	    kill_fasync(&fasyncptr, SIGIO, POLL_IN);
    }

    adc_disable();
    tpanel_check_press_mode();

    pen_state = MEASURE_PRESS;
    {
	int ret;
	u32 tm_low, cnt, rate_min, tp_wait;

	tx3920_read_rtc(value, tm_low);
	rate_min = conv_100u_to_rtc(tpanel_conf.sample_rate);
	tp_wait = conv_100u_to_rtc(tpanel_conf.press_check_wait);
	cnt = tm_low - last_sample_time;
	if( rate_min <= tp_wait ){
	    cnt = tp_wait;
	}else if( cnt >= rate_min - tp_wait ){
	    cnt = tp_wait;
	}else{
	    cnt = rate_min - cnt;
	}
	ret = tx3920_alarm_set_handler(cnt, tpanel_timer_handler, NULL);
	if( ret != 0 )
	    printk("tpanel.c: tx3920_alarm_set_handler() {press} error (%08x)\n", ret);
    }

    break;

  case MEASURE_PRESS:
    tx3920_read_rtc(value, last_sample_time);
    value = SIBSubFrame0Status;

    if ( !(value & BETTY_TSPX_LOW) && !(value & BETTY_TSMX_LOW)){
      /* pen up! */
	// printk("Pen Up\n");
		  
      tpanel_put_sample(0, sample_x, sample_y, sample_pressure, lost_count);
      lost_count = 0;

      wake_up_interruptible(&tpanel_wait);
      if (fasyncptr)
	kill_fasync(&fasyncptr, SIGIO, POLL_IN);

      tpanel_interrupt_mode();

      pen_state = MEASURE_PENDOWN;

    } else {
      /* Pen still down */
      //            printk("Pen still down\n");

      tpanel_readx_mode();
      sample_index = 0;

      sample_counter = 0;

      pen_state = MEASURE_PX;
    }

    break;

  }
}

static void do_sib_IRQ(void)
{
  unsigned long value;

  value = SIBSubFrame0Status;

  if(value & (BETTY_TSPX_INT|BETTY_TSMX_INT)){
    if(pen_state == MEASURE_PENDOWN){
      tpanel_getxy();
    }else{
      printk("do_sib_IRQ: press: bad pen state = %d\n", pen_state);
    }
  }else if(value & BETTY_ADC_INT){
    printk("do_sib_IRQ: adc: bad pen state = %d\n", pen_state);
  }
}

static void tpanel_interrupt(int irq, void *dev_id,
			     struct pt_regs * regs)
{
  IntEnable1 &= ~INT1_SIBIRQPOSINT;
  IntClear1 = INT1_SIBIRQPOSINT;

  sib_read_check(BETTY_REG_INT_CLEAR_STATUS,do_sib_IRQ);
}

static int fasync_tpanel(int fd, struct file *filp, int on)
{
  int retval;

  retval = fasync_helper(fd, filp, on, &fasyncptr);
  if (retval < 0)
    return retval;
  return 0;
}

static int open_tpanel(struct inode * inode, struct file * file)
{
  samplebuf_tail = samplebuf_head;

  return 0;
}

static int close_tpanel(struct inode * inode, struct file * file)
{
  fasync_tpanel(-1, file, 0);
#if 0
  if (--tpanel.active)
    return 0;
  // set for standby
  *VR41XX_MPIUINTREG = 0;
  *VR41XX_PIUCNTREG = 0;
  free_irq(VR41XX_IRQ_PIU, NULL);
  *VR41XX_CMUCLKMSK &= ~VR41XX_CMUCLKMSK_MSKPIUPCLK;
#ifdef CONFIG_CPU_VR4181
  // MFK: coordinate with AIU driver
  *VR41XX_CMUCLKMSK &= ~VR41XX_CMUCLKMSK_MSKADUPCLK | VR41XX_CMUCLKMSK_MSKADU18M;
#endif
  MOD_DEC_USE_COUNT;
#endif
  return 0;
}

/*
 *  Read touch panel data.
 *
 *  Data format:
 *  unsigned short state: 0 = pen up
 *                        1 = pen down
 *  unsigned short x:     raw data
 *  unsigned short y:     raw data
 *
 *  No calibration is done on the driver side, that is expected to be done on
 *  the user side.
 */

static ssize_t read_tpanel(struct file * file,
			   char * buffer, size_t count, loff_t *ppos)
{
  unsigned int state, x, y, mx, my, pressure, lost;
  size_t retcnt = 0;

  if (count < 12)
    return -EINVAL;

  // We need to access the circular buffer atomic to
  // anything else that will read or write it

  cli();

  init_waitqueue_head(&tpanel_wait);

  while (samplebuf_head == samplebuf_tail) {
    if (file->f_flags & O_NONBLOCK) {
      sti();
      return -EAGAIN;
    }
    interruptible_sleep_on(&tpanel_wait);
    // sleep_on will sti for us
    if (signal_pending(current))
      return -ERESTARTSYS;
    cli();
  }

  do {
    state = samplebuf_tail->state;
    lost = samplebuf_tail->lost_count;
    x = samplebuf_tail->x;
    y = samplebuf_tail->y;

    mx = samplebuf_tail->mx;
    my = samplebuf_tail->my;

    pressure = samplebuf_tail->pressure;

    samplebuf_tail++;
    if (samplebuf_tail == samplebuf + BUFFSIZE)
      samplebuf_tail = samplebuf;
		
    sti();
		
    if (!access_ok(VERIFY_WRITE, buffer, 12))
      return -EFAULT;

    switch(state){
    case 0:	/* pen up */
      __put_user(0x0000, (short*)(buffer + retcnt));
      __put_user(0, (short*)(buffer + retcnt + 2));
      __put_user(0, (short*)(buffer + retcnt + 4));
      __put_user(0, (short*)(buffer + retcnt + 6));
      __put_user(0, (short*)(buffer + retcnt + 8));
      __put_user(0, (short*)(buffer + retcnt + 10));
      break;
    case 1:	/* pen down */
      __put_user(0x4000, (short*)(buffer + retcnt));
      __put_user(0, (short*)(buffer + retcnt + 2));
      __put_user(0, (short*)(buffer + retcnt + 4));
      __put_user(0, (short*)(buffer + retcnt + 6));
      __put_user(0, (short*)(buffer + retcnt + 8));
      __put_user(0, (short*)(buffer + retcnt + 10));
      break;
    case 2:
	if( lost > 0 ){
	    if( lost > 0xff )
		lost = 0xff;
	    __put_user(0xe000 | lost, (short*)(buffer + retcnt));
	}else
	    __put_user(0xc000,   (short*)(buffer + retcnt));
      __put_user(x,        (short*)(buffer + retcnt + 2));
      __put_user(mx,   (short*)(buffer + retcnt + 4));
      __put_user(y,        (short*)(buffer + retcnt + 6));
      __put_user(my,   (short*)(buffer + retcnt + 8));
      __put_user(pressure, (short*)(buffer + retcnt + 10));
      break;
    }

    retcnt += 12;
    if (retcnt > count - 12)
      return retcnt;

    cli();
  } while (samplebuf_head != samplebuf_tail);

  sti();

  return retcnt;
}

static unsigned int poll_tpanel(struct file *file, poll_table * wait)
{
  int comp;

  poll_wait(file, &tpanel_wait, wait);
  cli();
  comp = (samplebuf_head != samplebuf_tail);
  sti();
  if (comp)
    return POLLIN | POLLRDNORM;
  return 0;
}


static int ioctl_tpanel(struct inode * inode, struct file * file,
			unsigned int cmd, unsigned long arg)
{
  struct scanparam *sparm;
  unsigned int i1, i2;

  sparm = (struct scanparam *)arg;

  switch (cmd) {
  case TPGETSCANPARM:
    i1 = 0;
    i2 = 0;
    if (put_user(i1, &sparm->interval))
      return -EFAULT;
    if (put_user(i2, &sparm->settletime))
      return -EFAULT;
    return 0;
  case TPSETSCANPARM:
    if (get_user(i1, &sparm->interval))
      return -EFAULT;
    if (get_user(i2, &sparm->settletime))
      return -EFAULT;
    __put_user(i1, &sparm->interval);
    __put_user(i2, &sparm->settletime);
    return 0;
#ifdef DEBUG_USE_IOCTL
  case TPGET_ADCWAIT:
      put_user(tpanel_conf.adc_wait, (int*)arg);
      return 0;

  case TPSET_ADCWAIT:
      tpanel_conf.adc_wait_next = tpanel_conf.adc_wait = arg;
      return 0;
  case TPGET_ADCWAITNT:
      put_user(tpanel_conf.adc_wait_next, (int*)arg);
      return 0;
  case TPSET_ADCWAITNT:
      tpanel_conf.adc_wait_next = arg;
      return 0;
  case TPGET_PRESSMIN:
      put_user(tpanel_conf.pressure_min, (int*)arg);
      return 0;
  case TPSET_PRESSMIN:
      tpanel_conf.pressure_min = arg;
      return 0;
  case TPGET_PRESSCHK:
      put_user(tpanel_conf.pressure_check, (int*)arg);
      return 0;
  case TPSET_PRESSCHK:
      tpanel_conf.pressure_check = arg;
      return 0;
  case TPGET_SAMPRATE:
      put_user(tpanel_conf.sample_rate, (int*)arg);
      return 0;
  case TPSET_SAMPRATE:
      tpanel_conf.sample_rate = arg;
      return 0;
#endif
  default:
    return -EINVAL;
  }
}

struct file_operations telios_tpanel_fops = {
  read:           read_tpanel,
  poll:           poll_tpanel,
  ioctl:	  ioctl_tpanel,
  open:           open_tpanel,
  release:        close_tpanel,
  fasync:	  fasync_tpanel,
};

static struct miscdevice telios_tpanel = {
  VR41XX_TPANEL_MINOR, "telios_tpanel", &telios_tpanel_fops
};

static int __init telios_tpanel_init(void)
{
  int i;
  unsigned flags;

  init_waitqueue_head(&tpanel_wait);

  save_and_cli(flags);

  /* SIB (for touchscreen, sound and modem) */
  sib_init();

  IntEnable1 &= ~INT1_SIBIRQPOSINT;
  IntClear1 = INT1_SIBIRQPOSINT;

  sharp_telios_map_bklgt_io();

  /* Set panel to "interrupt mode" */
  if ((i = request_irq(20, tpanel_interrupt, SA_INTERRUPT,
		       "touchpanel", NULL))) {
    printk(KERN_ERR "tpanel: Cannot allocate irq. (%d)\n", i);
    return i;
  }

  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, ~0);
  sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);

  tpanel_interrupt_mode();
  pen_state = MEASURE_PENDOWN;
  
  fasyncptr = NULL;

  restore_flags(flags);

  misc_register(&telios_tpanel);

  printk("touch panel initialized\n");
  return 0;
}

__initcall(telios_tpanel_init);
