/* $Id: e105btns.c,v 1.7 2000/01/24 12:13:49 mfklar Exp $
 *
 * Casio E-10x/50x input button (sub)driver
 *
 * Original driver by Robert Coie, modified and integrated into main VR41xx
 * buttons driver by Michael Klar.
 *
 * Copyright (c) 1999, Robert Coie <rac@intrigue.com>
 * Copyright (c) 2000, Michael Klar <wyldfier@iname.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/ioport.h>
#include <linux/buttons.h>

#include <asm/io.h>
#include <asm/vr41xx.h>

extern void add_button_data(unsigned char);

#define E105_BUTTONS_GPIO	9
#define E105_BUTTONS_IRQ	VR41XX_IRQ_GPIO(E105_BUTTONS_GPIO)
#define E105_BUTTONS_MASK	(1 << E105_BUTTONS_GPIO)
#define E105_BUTTONS_PORT	0xA018
#define E105_BUTTONS_DEV_ID	((void *) 0x45357878)	/* 'E5xx' */

const unsigned short e105_btn_map[] = { BTN_UP, BTN_DOWN, BTN_ACTION,
		 BTN_EXIT, BTN_NORTH, BTN_SOUTH, BTN_EAST, BTN_WEST };

struct e105_buttons_device_t {
	u8		status;
};

static struct e105_buttons_device_t buttons;


static void
e105_buttons_interrupt( int irq, void * dev_id, struct pt_regs * regs )
{
	u8	new_status;
	int	i;

	// flip active level for interrupt to opposite of current pin state
	if (*VR41XX_GIUPIODL & E105_BUTTONS_MASK)
		*VR41XX_GIUINTALSELL &= ~E105_BUTTONS_MASK;
	else
		*VR41XX_GIUINTALSELL |= E105_BUTTONS_MASK;
	
// MFK *** this is the old behavior, with only button press detected, no
// debounce, and only the first button detected if multiple buttons pressed.
// A fuller implimentation is to come.
	new_status = inb(E105_BUTTONS_PORT);
	if( (buttons.status = new_status) != 0 )
		for (i = 0; i < 8; i++)
			if (new_status & (1 << i))
				add_button_data(i + 0x20);
}

void open_e105_buttons( void )
{
	if ( check_region( E105_BUTTONS_PORT, 1 ) )
		return;

	buttons.status = 0;
	// see README.GIU for explaination of why cli/sti
	cli();
	*VR41XX_GIUINTTYPL   &= ~E105_BUTTONS_MASK;
	*VR41XX_GIUINTALSELL |=  E105_BUTTONS_MASK;
#if 0 // shouldn't be needed
	*VR41XX_GIUINTENL    |=  E105_BUTTONS_MASK;
#endif
	sti();

	if ( request_irq( E105_BUTTONS_IRQ, e105_buttons_interrupt,
			 SA_SHIRQ, "E105buttons", E105_BUTTONS_DEV_ID ) );
		return;
	request_region( E105_BUTTONS_PORT, 1, "E105buttons" );
}

void close_e105_buttons( void )
{
	release_region( E105_BUTTONS_PORT, 1 );
	free_irq( E105_BUTTONS_IRQ, E105_BUTTONS_DEV_ID );
}

