/*
 * linux/arch/mips/vr41xx/mq200.c
 *
 * Platform support for the machine which LCD control by MQ200.
 *
 * Copyright (C) 2000 Soichiro Nakanisi
 * Copyright (C) 2000 SATO Kazumi
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <asm/io.h>
#include <asm/system.h>
#include <asm/vr41xx.h>

#if defined(CONFIG_NEC_MOBILEGEAR2_R430)
#define MQ200_LCD_CONTROL	__preg32(KSEG1 + 0x0a60e000)
#define	MQ200_LCD_ON		0x00000001
#elif defined(CONFIG_NEC_MOBILEGEAR2_R530)
#define MQ200_LCD_CONTROL	__preg32(KSEG1 + 0x0a60e000)
#define	MQ200_LCD_ON		0x00000001
#else
#error This machine really use MQ200???
#endif

#define LCDON_PORT	MQ200_LCD_CONTROL
#define	LCDON		MQ200_LCD_ON
#define LIGHTON_PORT	MQ200_LCD_CONTROL
#define	LIGHTON		MQ200_LCD_ON

static int backlight, contrast, lcdpower;

void mq200_init_backlight(void)
{
#ifdef STANDALONE
	backlight = 1;
#else /* STANDALONE */
	/* save boot time status (maybe set by Windows CE) */
	/* but we don't know current value methods, so set constant */
	backlight = 1;
#endif /* STANDALONE */
}

int get_mq200_backlight(void)
{
	return backlight;
}

int mq200_backlight(int n)
{
	backlight = n;
	if(n == 0) {
		;
	} else {
		;
	}
	return 0;
}

void mq200_init_contrast(void)
{
#ifdef STANDALONE
        contrast = 6;
#else /* STANDALONE */
        /* save boot time status (maybe set by Windows CE) */
        /* but we don't know current value methods, so set constant */
        contrast = 6;
#endif /* STANDALONE */
}

int get_mq200_contrast(void)
{
	return contrast;
}

int mq200_contrast(int n)
{
	contrast = n;
	// Turn the lcd on and some contrast.
	return 0;
}

int get_mq200_lcdpower(void)
{
	return lcdpower;
}

int mq200_lcdpower(int n)
{
	int flags;

	lcdpower = n;
	if(n == 0) {
		save_and_cli(flags);
		*LCDON_PORT &=  ~LCDON;
		restore_flags(flags);
	} else {
		save_and_cli(flags);
		*LCDON_PORT |=  LCDON;
		restore_flags(flags);
	}
	return 0;
}

void mq200_setup(void)
{
	lcdpower = 1; /* boot time always on */
	/* backlight & contrast inherit by WinCE */
	mq200_init_backlight();
	mq200_init_contrast();
	mq200_lcdpower(lcdpower);
	mq200_contrast(contrast);
	mq200_backlight(backlight);
}
