/*
 * linux/arch/mips/vr41xx/setup.c
 *
 * VR41xx setup routines
 *
 * Copyright (C) 1999 Bradley D. LaRonde
 * Copyright (C) 1999, 2000 Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/config.h>
#include <linux/console.h>
#include <linux/ide.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/reboot.h>
#include <asm/vr41xx.h>
#include <asm/vr41xx-platdep.h>
#include <asm/io.h>

struct semaphore vr41xx_dma_sem;

// This enables the CF hacks below, to be replaced with real CF driver eventually
#if defined(CONFIG_EVEREX_FREESTYLE)
#define CF_HACK
#else
// If CF doesn't work on your device, try changing this to #define:
#undef CF_HACK
#endif

#ifdef CONFIG_TCIC
#define TCIC_BASE		0x240

#define TCIC_MODE		0x08
#define TCIC_AUX		0x0E

#define TCIC_MODE_PGMMASK	0x1f
#define TCIC_AUX_ILOCK		(6<<5)
#define TCIC_ILOCK_CRESET	0x04
#endif

extern struct ide_ops vr41xx_ide_ops;
extern void vr41xx_restart(char* c);
extern void vr41xx_hibernate(void);
extern void vr41xx_wait(void);

#if defined(CF_HACK) || defined(CONFIG_I82365)
void __init put_cf_reg(unsigned char reg, unsigned char val)
{
#ifdef CONFIG_CPU_VR4181
	*VR41XX_PCCARDINDEX = reg;
	barrier();
	*VR41XX_PCCARDDATA = val;
#else
	outb(reg, 0x3e0);
	outb(val, 0x3e1);
#endif
}
#endif

#if defined(CONFIG_NEC_OSPREY) || defined(CONFIG_IBM_WORKPAD) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R300) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R310) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R430) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R530) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R700) || \
    defined(CONFIG_DOCOMO_SIGMARION) || \
    defined(CONFIG_AGENDA_VR3)
extern void platdep_setup(void);
#else
void platdep_setup(void) {}
#endif

void __init vr41xx_setup(void)
{
	mips_io_port_base = ADJUSTED_PORT_BASE;
	isa_slot_offset = VR41XX_ISAMEM_BASE;

#ifdef CONFIG_BLK_DEV_IDE
	ide_ops = &vr41xx_ide_ops;
#endif

#ifdef CONFIG_FB
	conswitchp = &dummy_con;
#endif
 
	// reset the PC card, and power off
#ifdef CONFIG_I82365
	put_cf_reg(0x03, 0x20);	// socket 0
	put_cf_reg(0x43, 0x20); // socket 1
	put_cf_reg(0x02, 0x00);	// socket 0
	put_cf_reg(0x42, 0x00); // socket 1
#endif
#ifdef CONFIG_TCIC
{
	u_char mode;
	mode = (inb(TCIC_BASE+TCIC_MODE) & TCIC_MODE_PGMMASK) | TCIC_AUX_ILOCK;
	outb(mode, TCIC_BASE+TCIC_MODE);
	outb(TCIC_ILOCK_CRESET, TCIC_BASE+TCIC_AUX);
}
#endif

#ifdef CF_HACK
	// this is a nasty hack to initialize the CF registers, eventually this
	// is to be replaced with a real CF controller driver
 {
	unsigned char i;
	unsigned char cfinitdata[] =  { 0x90, 0x63, 0x00, 0x35, 0x00, 0x55,
		0x76, 0x03, 0x77, 0x03, 0x70, 0x01, 0x78, 0x01, 0x10, 0x00,
		0x1f, 0x00, 0xf0, 0x7f, 0x00, 0x00, 0x20, 0x00, 0xe4, 0x00,
		0xe0, 0x3f, 0x00, 0x02, 0xe5, 0x00, 0xed, 0x00, 0x1b, 0x3f,
		0x00, 0x00, 0xee, 0x00, 0xf6, 0x00, 0x00, 0x00, 0x00, 0x01,
		0xf7, 0x00, 0xff, 0x00, 0x00, 0x00 };

#ifdef CONFIG_CPU_VR4181
	// Enable CF, not key scan.
	*VR41XX_KEYEN = 0;

	// Unmask all CF IRQs for now except for IRQ 3 (which makes it fail)
	// I think the bit sense in reversed from what User Manual states
	*VR41XX_INTMSKREG = 0xdeb0;

	// Set CF wait states.
	*VR41XX_CFG_REG_1 = 1;
#else
#warning "Generic CF hack in use, assuming non-i82365 registers set up prior to boot"
#endif

	put_cf_reg(0x02, 0x10);
	udelay(1);

	for (i = 0x02; i < 0x36; i++)
		put_cf_reg(i, cfinitdata[i-2]);

	// enable the IO and mem windows, now that start and stop values set up
	put_cf_reg(0x06, 0xdf);

	// reset the card
	put_cf_reg(0x03, 0x23);
	udelay(1);
	put_cf_reg(0x03, 0x63);
	mdelay(30);

	// now set up CF config register (this assumes it is a CF card...)
	writeb(0x43, ioremap(0x10200, 1));
 }
#endif

	_machine_restart = vr41xx_restart;
	_machine_halt = vr41xx_hibernate;
	_machine_power_off = vr41xx_hibernate;

	cpu_wait = vr41xx_wait;

	// Insure that vr41xx_dma_sem is initialized as unlocked, even
	// in the case of a failed hibernate/wakeup:
	init_MUTEX(&vr41xx_dma_sem);

	// Do platform-dependent setup.
	// This is mostly stuff that doesn't fit well anywhere else.
	platdep_setup();
}
