/*
 *
 *  linux/drivers/char/tx39uart.h
 *
 *  Toshiba TMPR39xx UART module support.
 *
 *  Copyright (C) 2000  Yoichi Yuasa <yuasa@users.sourceforge.net>
 *
 */

#define TX39UART_NPORTS		2

#define TX39UART_UARTA_IRQ	3
#define TX39UART_UARTB_IRQ	4

/* TX39xx UART Register Base */
#define TX39UART_UARTA_BASE	KSEG1ADDR(0x10c000b0)
#define TX39UART_UARTB_BASE	KSEG1ADDR(0x10c000c8)

/* UART Register Offsets */
#define TX39UART_CTRL1		0x00
#define TX39UART_CTRL2		0x04
#define TX39UART_DMACTRL1	0x08
#define TX39UART_DMACTRL2	0x0c
#define TX39UART_DMACOUNT	0x10
#define TX39UART_RX		0x14
#define TX39UART_TX		0x14

#define TX39UART_UARTA_REG(x)	(*((volatile u32 *)(TX39UART_UARTA_BASE + x)))
#define TX39UART_UARTB_REG(x)	(*((volatile u32 *)(TX39UART_UARTB_BASE + x)))

#define UartA_Ctrl1		TX39UART_UARTA_REG(TX39UART_CTRL1)
#define UartA_Ctrl2		TX39UART_UARTA_REG(TX39UART_CTRL2)
#define UartA_DMActrl1		TX39UART_UARTA_REG(TX39UART_DMACTRL1)
#define UartA_DMActrl2		TX39UART_UARTA_REG(TX39UART_DMACTRL2)
#define UartA_DMAcount		TX39UART_UARTA_REG(TX39UART_DMACOUNT)
#define UartA_Data		TX39UART_UARTA_REG(TX39UART_RX)

#define UartB_Ctrl1		TX39UART_UARTB_REG(TX39UART_CTRL1)
#define UartB_Ctrl2		TX39UART_UARTB_REG(TX39UART_CTRL2)
#define UartB_DMActrl1		TX39UART_UARTB_REG(TX39UART_DMACTRL1)
#define UartB_DMActrl2		TX39UART_UARTB_REG(TX39UART_DMACTRL2)
#define UartB_DMAcount		TX39UART_UARTB_REG(TX39UART_DMACOUNT)
#define UartB_Data		TX39UART_UARTB_REG(TX39UART_RX)

/* UART Control 1 Register */
#define TX39UART_UARTON		BIT(31)
#define TX39UART_EMPTY		BIT(30)
#define TX39UART_PRXHOLDFULL	BIT(29)
#define TX39UART_RXHOLDFULL	BIT(28)
#define TX39UART_ENDMARX 	BIT(15)
#define TX39UART_ENDMATX 	BIT(14)
#define TX39UART_ENBREAKHALT 	BIT(12)
#define TX39UART_PULSEOPT2	BIT(9)
#define TX39UART_PULSEOPT1	BIT(8)
#define TX39UART_DTINVERT	BIT(7)
#define TX39UART_DISTXD		BIT(6)
#define TX39UART_TWOSTOP	BIT(5)
#define TX39UART_LOOPBACK	BIT(4)
#define TX39UART_BIT_7		BIT(3)
#define TX39UART_EVENPARITY	BIT(2)
#define TX39UART_ENPARITY	BIT(1)
#define TX39UART_ENUART		BIT(0)

/* UART DMA Control 1 Register */
#define TX39UART_DMA_START	0xfffffffc

/* UART DMA Control 2 Register */
#define TX39UART_DMA_LENGTH	0x0000ffff

/* UART DMA Count */
#define TX39UART_DMA_COUNT	0x0000ffff

/* UART Transmit Holding Register and UART Receiver Holding Register */
#define TX39UART_BREAK		BIT(8)
#define TX39UART_TX_DATA	0x000000ff
#define TX39UART_RX_DATA	0x000000ff

/* Baud rate definitions for UART control register 2 */
#define TX39UART_BASE_BAUD	576000
#define TX39UART_BAUD_115200	4
#define TX39UART_BAUD_57600	9
#define TX39UART_BAUD_38400	14
#define TX39UART_BAUD_19200	29
#define TX39UART_BAUD_9600	59
#define TX39UART_BAUD_4800	119
#define TX39UART_BAUD_2400	239
#define TX39UART_BAUD_1200	479
#define TX39UART_BAUD_600	959
#define TX39UART_BAUD_300	1919

#define TX39UART_MINOR_START	64
#define TX39UART_RX_THROTTLE	0x0000001

#define TX39UART_CONSOLE_ACTIVE	0x01

#define O_OTHER(tty)    \
      ((O_OLCUC(tty))  ||\
      (O_ONLCR(tty))   ||\
      (O_OCRNL(tty))   ||\
      (O_ONOCR(tty))   ||\
      (O_ONLRET(tty))  ||\
      (O_OFILL(tty))   ||\
      (O_OFDEL(tty))   ||\
      (O_NLDLY(tty))   ||\
      (O_CRDLY(tty))   ||\
      (O_TABDLY(tty))  ||\
      (O_BSDLY(tty))   ||\
      (O_VTDLY(tty))   ||\
      (O_FFDLY(tty)))

#define I_OTHER(tty)    \
      ((I_INLCR(tty))  ||\
      (I_IGNCR(tty))   ||\
      (I_ICRNL(tty))   ||\
      (I_IUCLC(tty))   ||\
      (L_ISIG(tty)))

#define TX39UART_MAGIC		0x12345678

struct tx39uart_port {
        struct gs_port          gs;
        int                     line;
        u32                     base;
        unsigned int            irq;
	u32			flags;
	int			x_char;
};
