/*
 * Toshiba PLUM2 PCMCIA Controller support.
 *
 * Copyright (C) 2000 Yoichi Yuasa <yuasa@users.sourceforge.net>
 * All rights reserved
 */

#ifndef __PLUM2_PCMCIA_H
#define __PLUM2_PCMCIA_H

/* 
 * Register definitions for the Toshiba PLUM2 PCMCIA Controller
 */
#define PLUM2_PCC1REG_BASE	KSEG1ADDR(0x10805000)
#define PLUM2_PCC2REG_BASE	KSEG1ADDR(0x10805800)

/* Register Table */
#define PLUM2_PCC_ID		0x000	/* Identification and Revision */
#define PLUM2_PCC_STATUS	0x004	/* Interface Status */
#define PLUM2_PCC_POWERCTRL	0x008	/* Power Control */
#define PLUM2_PCC_INTCTRL	0x00c	/* Interrupt and General Control */
#define PLUM2_PCC_CSC		0x010	/* Card Status Changes */
#define PLUM2_PCC_CSCINTCTRL	0x014	/* Card Status Changes Interrupt Control */
#define PLUM2_PCC_WINENABLE	0x018	/* Window Enable */
#define PLUM2_PCC_IO_CTRL	0x01c	/* I/O Window Control */
#define PLUM2_PCC_IO0_START	0x020	/* I/O Window 0 Start Address */
#define PLUM2_PCC_IO0_STOP	0x024	/* I/O Window 0 Stop Address */
#define PLUM2_PCC_IO0_ADDRCTRL	0x028	/* I/O Window 0 Address Control */
#define PLUM2_PCC_IO1_ADDRCTRL	0x038	/* I/O Window 1 Address Control */
#define PLUM2_PCC_MEM0_START	0x040	/* Memory Window 0 Start Address */
#define PLUM2_PCC_MEM0_STOP	0x044	/* Memory Window 0 Stop Address */
#define PLUM2_PCC_MEM0_OFFSET	0x048	/* Memory Window 0 Offset Address */
#define PLUM2_PCC_MEM0_CTRL	0x04c	/* Memory Window 0 Control */
#define PLUM2_PCC_ADDGENCTRL	0x058	/* Additional General Control */
#define PLUM2_PCC_GCTRL		0x078	/* Global Control */
#define PLUM2_PCC_CARDTIMING	0x0ec	/* PCMCIA Card Timing Register */
#define PLUM2_PCC_RESERVED	0x0f0	/* Reserved Register */
#define PLUM2_PCC_FUNCCTRL	0x0f8	/* PCMCIA Function Control Register */
#define PLUM2_PCC_SPMODE	0x0fc	/* Special Mode Register */
#define PLUM2_PCC_SLOTCTRL	0x100	/* Slot Control Register */
#define PLUM2_PCC_BUFOFF	0x104	/* Buffer Off Register */
#define PLUM2_PCC_DETECTMODE	0x108	/* Card Detect Mode Register */
#define PLUM2_PCC_CARDPOWER	0x10c	/* Card Power Control Register(Common) */

#define PLUM2_PCC_IO_START(x)	(PLUM2_PCC_IO0_START + (x << 4))
#define PLUM2_PCC_IO_STOP(x)	(PLUM2_PCC_IO0_STOP + (x << 4))
#define PLUM2_PCC_IO_ADDRCTRL(x) (PLUM2_PCC_IO0_ADDRCTRL + (x << 4))
#define PLUM2_PCC_MEM_START(x)	(PLUM2_PCC_MEM0_START + (x << 5))
#define PLUM2_PCC_MEM_STOP(x)	(PLUM2_PCC_MEM0_STOP + (x << 5))
#define PLUM2_PCC_MEM_OFFSET(x)	(PLUM2_PCC_MEM0_OFFSET + (x << 5))
#define PLUM2_PCC_MEM_CTRL(x)	(PLUM2_PCC_MEM0_CTRL + (x << 5))

#define PLUM2_PCC_IO_MAX	1
#define PLUM2_PCC_MEM_MAX	4

/* Idetification and Revision */
#define PLUM2_PCC_REVISION	(BIT(3)|BIT(2)|BIT(1)|BIT(0))
#define PLUM2_PCC_CARDID	(BIT(7)|BIT(6))

/* Interface Status Register */
#define PLUM2_PCC_STSCHG	BIT(0)
#define PLUM2_PCC_BVD1		BIT(0)
#define PLUM2_PCC_BVD2		BIT(1)
#define PLUM2_PCC_BINAUDIO	BIT(1)
#define PLUM2_PCC_CD		(BIT(3)|BIT(2))
#define PLUM2_PCC_WRITEPROTECT	BIT(4)
#define PLUM2_PCC_READY		BIT(5)
#define PLUM2_PCC_INTREQ	BIT(5)
#define PLUM2_PCC_POWERACTIVE	BIT(6)

/* Power Control Register */
#define PLUM2_PCC_CARDPOWENABLE	BIT(4)
#define PLUM2_PCC_AUTOPOWER	BIT(5)
#define PLUM2_PCC_DISABLERESET	BIT(6)
#define PLUM2_PCC_OUTPUTENABLE	BIT(7)
//#define PLUM2_PCC_POWERENABLE	(BIT(7)|BIT(4))

#define PLUM2_PCC_VCC_5V	BIT(3)
#define PLUM2_PCC_VCC_3V	BIT(2)
#define PLUM2_PCC_VCC_0V	(BIT(3)|BIT(2))
#define PLUM2_PCC_VCC_MASK	(BIT(3)|BIT(2))

#define PLUM2_PCC_VPP1_0V	BIT(1)
#define PLUM2_PCC_VPP1_VCC	BIT(0)
#define PLUM2_PCC_VPP1_HIZ	(BIT(1)|BIT(0))
#define PLUM2_PCC_VPP1_MASK	(BIT(1)|BIT(0))

/* Interrupt and General Control Register */
#define PLUM2_PCC_CARDTYPE	BIT(5)
#define PLUM2_PCC_IOCARD	BIT(5)
#define PLUM2_PCC_MEMCARD	0
#define PLUM2_PCC_CARDRESET	BIT(6)

/* Card Status Change Register and Card Status Change Interrupt Control Register */
#define PLUM2_PCC_BATTDEAD	BIT(0)
#define PLUM2_PCC_BATTWARN	BIT(1)
#define PLUM2_PCC_CSCREADY	BIT(2)
#define PLUM2_PCC_DETECT	BIT(3)

/* Window Enable Register */
#define PLUM2_PCC_IO0_ENABLE	BIT(6)
#define PLUM2_PCC_IO1_ENABLE	BIT(7)
#define PLUM2_PCC_IO_ENABLE(x)	(PLUM2_PCC_IO0_ENABLE << x)

#define PLUM2_PCC_MEM0_ENABLE	BIT(0)
#define PLUM2_PCC_MEM1_ENABLE	BIT(1)
#define PLUM2_PCC_MEM2_ENABLE	BIT(2)
#define PLUM2_PCC_MEM3_ENABLE	BIT(3)
#define PLUM2_PCC_MEM4_ENABLE	BIT(4)
#define PLUM2_PCC_MEM_ENABLE(x)	(PLUM2_PCC_MEM0_ENABLE << x)

/* I/O Window Control Register */
#define PLUM2_PCC_IO_16BIT(x)	(BIT(0) << (x << 2))
#define PLUM2_PCC_IO_IOIS16S(x)	(BIT(1) << (x << 2))
#define PLUM2_PCC_IO_0WS(x)	(BIT(2) << (x << 2))
#define PLUM2_PCC_IO_WS(x)	(BIT(3) << (x << 2))
#define PLUM2_PCC_IO_MASK(x)	((BIT(0)|BIT(1)|BIT(2)|BIT(3)) << (x << 2))

/* I/O Window */
#define PLUM2_PCC_IO_WINDOWS	2
#define PLUM2_PCC_IO_START_MAX	0xffff
#define PLUM2_PCC_IO_STOP_MAX	0xffff

/* I/O Window Address Control Register */
#define PLUM2_PCC_IO_AREA1	0
#define PLUM2_PCC_IO_AREA2	BIT(0)

/* Memory Window */
#define PLUM2_PCC_MEM_WINDOWS	5
#define PLUM2_PCC_MEM_MASK	0x3fff
#define PLUM2_PCC_MEM_SHIFT	12
#define PLUM2_PCC_MEM_SIZE	(1UL << PLUM2_PCC_MEM_SHIFT)

#define PLUM2_PCC_MEM_START_MAX	0x3fffff
#define PLUM2_PCC_MEM_STOP_MAX	0x3fffff
#define PLUM2_PCC_MEM_SPEED_MAX	1000
#define PLUM2_PCC_CARDSTART_MAX	0x03ffffff

#define PLUM2_PCC_MEM_8M	0x00800000
#define PLUM2_PCC_MEM_16M	0x01000000
#define PLUM2_PCC_MEM_32M	0x02000000
#define PLUM2_PCC_MEM_64M	0x04000000
#define PLUM2_PCC_MEM_SIZE_MAX	PLUM2_PCC_MEM_64M

/* Memory Window Control Register */
#define PLUM2_PCC_MEM_0WS	BIT(6)
#define PLUM2_PCC_MEM_16BIT	BIT(7)
#define PLUM2_PCC_MEM_WS0	BIT(15)
#define PLUM2_PCC_MEM_WS1	BIT(16)
#define PLUM2_PCC_MEM_WS	(BIT(16)|BIT(15))
#define PLUM2_PCC_MEM_REGACT	BIT(22)
#define PLUM2_PCC_MEM_WRPROT	BIT(23)
#define PLUM2_PCC_MEM_MAPCTRL	(BIT(25)|BIT(24))
#define PLUM2_PCC_MEM_MAP1	0
#define PLUM2_PCC_MEM_MAP2	BIT(24)
#define PLUM2_PCC_MEM_MAP3	BIT(25)
#define PLUM2_PCC_MEM_MAP4	(BIT(25)|BIT(24))

#define PLUM2_PCC_IO1_BASE	0x6c600000
#define PLUM2_PCC_IO1_SIZE	0x00100000
#define PLUM2_PCC_IO2_BASE	0x6c700000
#define PLUM2_PCC_IO2_SIZE	0x00100000

#define PLUM2_PCC_MEM1_BASE	0x6c800000
#define PLUM2_PCC_MEM1_SIZE	0x00800000
#define PLUM2_PCC_MEM2_BASE	0x6d000000
#define PLUM2_PCC_MEM2_SIZE	0x01000000
#define PLUM2_PCC_MEM3_BASE	0x6e000000
#define PLUM2_PCC_MEM3_SIZE	0x02000000
#define PLUM2_PCC_MEM4_BASE	0x70000000
#define PLUM2_PCC_MEM4_SIZE	0x04000000

/* PCMCIA Function Control Register */
#define PLUM2_PCC_POWCTRLBUF_EN	BIT(6)
#define PLUM2_PCC_CARDTIMING_EN	BIT(3)
#define PLUM2_PCC_VSS_ENABLE	BIT(1)
#define PLUM2_PCC_3VPOWER_EN	BIT(0)

/* Slot Control Register */
#define PLUM2_PCC_SLOT_ON	BIT(7)
#define PLUM2_PCC_ID_LOCK	BIT(5)
#define PLUM2_PCC_ID_WP		BIT(4)

/* Card Power Control Register */
#define PLUM2_PCC_CARD_STARTUP	0x00
#define PLUM2_PCC_CARD_SHUTDOWN	0x01

/* Global Control Register */
#define PLUM2_PCC_EWB_CSC_INT	0x04

/* Additional General Control */
#define PLUM2_PCC_VSENCE2	BIT(7)
#define PLUM2_PCC_VSENCE1	BIT(6)
#define PLUM2_PCC_DETECT_INT	BIT(5)
#define PLUM2_PCC_REMOVAL_RST	BIT(1)
#endif
