/*
	linux/drivers/sound/teliossnd.c

	Sound driver for SHARP Telios
	(don't work. debugging now!!)

        Copyright (C) 2001 AXE, Inc.  (e-mail: <fuku@axe-inc.co.jp>)

        Based on drivers/sound/vr41xx.c
	Copyright (C) 2000 by Hiroshi Kawashima <kawashima@iname.com>

	This is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
	Version 2 (June 1991). See the "COPYING" file distributed with this
	software for more info.
*/

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/config.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/sound.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>

#include <asm/tx3920.h>
#include <asm/plum2.h>
#include <asm/betty.h>

#include <linux/soundcard.h>

#include <linux/delay.h>
#include <asm/delay.h>

// #define DEBUG	1

#ifdef MODULE
static int sq_unit = -1;
static int mixer_unit = -1;
static int state_unit = -1;
static int irq_installed = 0;
#endif /* MODULE */
static char **sound_buffers = NULL;

#ifdef CONFIG_SOUND_TELIOS
static char **telios_dma_bufs = NULL;
static char *telios_sound_buffer = NULL;
#endif


/*** Some declarations *******************************************************/
/*==== Copied from old dmasound.h ====*/
/*
 * Minor numbers for the sound driver.
 *
 * Unfortunately Creative called the codec chip of SB as a DSP. For this
 * reason the /dev/dsp is reserved for digitized audio use. There is a
 * device for true DSP processors but it will be called something else.
 * In v3.0 it's /dev/sndproc but this could be a temporary solution.
 */

#define SND_NDEVS	256		/* Number of supported devices */
#define SND_DEV_CTL	0		/* Control port /dev/mixer */
#define SND_DEV_SEQ	1		/* Sequencer output /dev/sequencer (FM
							   synthesizer and MIDI output) */
#define SND_DEV_MIDIN	2	/* Raw midi access */
#define SND_DEV_DSP		3	/* Digitized voice /dev/dsp */
#define SND_DEV_AUDIO	4	/* Sparc compatible /dev/audio */
#define SND_DEV_DSP16	5	/* Like /dev/dsp but 16 bits/sample */
#define SND_DEV_STATUS	6	/* /dev/sndstat */
/* #7 not in use now. Was in 2.4. Free for use after v3.0. */
#define SND_DEV_SEQ2	8	/* /dev/sequencer, level 2 interface */
#define SND_DEV_SNDPROC 9	/* /dev/sndproc for programmable devices */
#define SND_DEV_PSS	SND_DEV_SNDPROC

#define DSP_DEFAULT_SPEED	8000

#define ON		1
#define OFF		0

#define MAX_AUDIO_DEV	5
#define MAX_MIXER_DEV	2
#define MAX_SYNTH_DEV	3
#define MAX_MIDI_DEV	6
#define MAX_TIMER_DEV	3
/*====================================*/

#ifdef CONFIG_SOUND_TELIOS
#  define	DMASND_TELIOS		6
#endif

#define MAX_CATCH_RADIUS		10
#define MIN_BUFFERS			4

#ifdef  CONFIG_SOUND_TELIOS
#  define MIN_BUFSIZE 			8	
#  define MAX_BUFSIZE			16
#endif

static int catchRadius = 0;
/* Real buffer size is bufSize * 1024 byte */
#ifdef  CONFIG_SOUND_TELIOS
static int numBufs = 128, bufSize = 8;
#endif

MODULE_PARM(catchRadius, "i");
MODULE_PARM(numBufs, "i");
MODULE_PARM(bufSize, "i");
MODULE_PARM(numreadBufs, "i");
MODULE_PARM(readbufSize, "i");

#define arraysize(x)	(sizeof(x)/sizeof(*(x)))
#define min(x, y)	((x) < (y) ? (x) : (y))
#define le2be16(x)	(((x)<<8 & 0xff00) | ((x)>>8 & 0x00ff))
#define le2be16dbl(x)	(((x)<<8 & 0xff00ff00) | ((x)>>8 & 0x00ff00ff))

#define IOCTL_IN(arg, ret) \
	do { int error = get_user(ret, (int *)(arg)); \
		if (error) return error; \
	} while (0)
#define IOCTL_OUT(arg, ret)	ioctl_return((int *)(arg), ret)


static void read_af_ioport(void)
{
    unsigned long value;
    value = SIBSubFrame0Status;
    ct_set(value, "NR");
}

static void read_af_ioport1(void)
{
    unsigned long value;
    value = SIBSubFrame0Status;
    ct_set(value, "N1");
}

static void read_af_ioport2(void)
{
    unsigned long value;
    value = SIBSubFrame0Status;
    ct_set(value, "N2");
}

/*** Some low level helpers **************************************************/


/*** Translations ************************************************************/

/*** Machine definitions *****************************************************/


typedef struct {
	int type;
	void *(*dma_alloc)(unsigned int, int);
	void (*dma_free)(void *, unsigned int);
	int (*irqinit)(void);
#ifdef MODULE
	void (*irqcleanup)(void);
#endif /* MODULE */
	void (*init)(void);
	void (*silence)(void);
	int (*setFormat)(int);
	int (*setVolume)(int);
	int (*setBass)(int);
	int (*setTreble)(int);
	int (*setGain)(int);
	void (*play)(void);
} MACHINE;


/*** Low level stuff *********************************************************/


typedef struct {
	int format;		/* AFMT_* */
	int stereo;		/* 0 = mono, 1 = stereo */
	int size;		/* 8/16 bit*/
	int speed;		/* speed */
} SETTINGS;

typedef struct {
	ssize_t (*ct_ulaw)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_alaw)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_s8)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_u8)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_s16be)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_u16be)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_s16le)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_u16le)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
} TRANS;

struct sound_settings {
	MACHINE mach;		/* machine dependent things */
	SETTINGS hard;		/* hardware settings */
	SETTINGS soft;		/* software settings */
	SETTINGS dsp;		/* /dev/dsp default settings */
	TRANS *trans;		/* supported translations */
	int volume_left;	/* volume (range is machine dependent) */
	int volume_right;
	int bass;		/* tone (range is machine dependent) */
	int treble;
	int gain;
	int minDev;		/* minor device number currently open */
};

static struct sound_settings sound;

#ifdef CONFIG_SOUND_TELIOS 
static int TeliosIrqInit(void);
#ifdef MODULE
static void TeliosIrqCleanUp(void);
#endif /* MODULE */
static void TeliosSilence(void);
static void TeliosInit(void);
static int TeliosSetFormat(int format);
static int TeliosSetVolume(int volume); 
/* static int Vr41xxSetTreble(int treble); */
static void telios_sq_play_next_frame(int index);
static void TeliosPlay(void);
static void telios_sq_interrupt(int irq, void *dummy, struct pt_regs *fp);
#endif /* CONFIG_SOUND_VR41XX */


/*** Mid level stuff *********************************************************/


static void sound_silence(void);
static void sound_init(void);
static int sound_set_format(int format);
static int sound_set_speed(int speed);
static int sound_set_stereo(int stereo);
static int sound_set_volume(int volume);
static ssize_t sound_copy_translate(const u_char *userPtr,
				    size_t userCount,
				    u_char frame[], ssize_t *frameUsed,
				    ssize_t frameLeft);
/*
 * /dev/mixer abstraction
 */

struct sound_mixer {
    int busy;
    int modify_counter;
};

static struct sound_mixer mixer;

/*
 * Sound queue stuff, the heart of the driver
 */

struct sound_queue {
	int max_count, block_size;
	char **buffers;
	int max_active;

	/* it shouldn't be necessary to declare any of these volatile */
	int front, rear, count;
	int rear_size;
	/*
	 *	The use of the playing field depends on the hardware
	 *
	 *	Atari, PMac: The number of frames that are loaded/playing
	 *
	 *	Amiga: Bit 0 is set: a frame is loaded
	 *	       Bit 1 is set: a frame is playing
	 */
	int active;
	wait_queue_head_t action_queue, open_queue, sync_queue;
	int open_mode;
	int busy, syncing;
};

static struct sound_queue sq;

#define sq_block_address(i)	(sq.buffers[i])
#define SIGNAL_RECEIVED	(signal_pending(current))
#define NON_BLOCKING(open_mode)	(open_mode & O_NONBLOCK)
#define ONE_SECOND	HZ	/* in jiffies (100ths of a second) */
#define NO_TIME_LIMIT	0xffffffff
#define SLEEP(queue, time_limit) \
	interruptible_sleep_on_timeout(&queue, (time_limit));
#define WAKE_UP(queue)	(wake_up_interruptible(&queue))

/*
 * /dev/sndstat
 */

struct sound_state {
	int busy;
	char buf[512];
	int len, ptr;
};

static struct sound_state state;

/*** Common stuff ********************************************************/

static loff_t sound_lseek(struct file *file, long long offset, int orig);
static inline int ioctl_return(int *addr, int value)
{
	if (value < 0)
		return(value);

	return put_user(value, addr);
}


/*** Config & Setup **********************************************************/


void dmasound_init(void);
void dmasound_setup(char *str, int *ints);


/*** Translations ************************************************************/


/* ++TeSche: radically changed for new expanding purposes...
 *
 * These two routines now deal with copying/expanding/translating the samples
 * from user space into our buffer at the right frequency. They take care about
 * how much data there's actually to read, how much buffer space there is and
 * to convert samples into the right frequency/encoding. They will only work on
 * complete samples so it may happen they leave some bytes in the input stream
 * if the user didn't write a multiple of the current sample size. They both
 * return the number of bytes they've used from both streams so you may detect
 * such a situation. Luckily all programs should be able to cope with that.
 *
 * I think I've optimized anything as far as one can do in plain C, all
 * variables should fit in registers and the loops are really short. There's
 * one loop for every possible situation. Writing a more generalized and thus
 * parameterized loop would only produce slower code. Feel free to optimize
 * this in assembler if you like. :)
 *
 * I think these routines belong here because they're not yet really hardware
 * independent, especially the fact that the Falcon can play 16bit samples
 * only in stereo is hardcoded in both of them!
 *
 * ++geert: split in even more functions (one per format)
 */

#ifdef CONFIG_SOUND_TELIOS

static inline void telios_fill(u_int d, u_int *p, int n)
{
	while (n--)
		*p++ = d;
}

static ssize_t telios_ct_s8(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;
	u_int *p = (u_int *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 1, frameLeft >> 3);
		uUsed = count * 2;
		fUsed = count * 8;
	} else {
		count = min(userCount, frameLeft >> 3);
		uUsed = count;
		fUsed = count * 8;
	}
	while (count > 0) {
		u_char data;
		u_int d;

		if (get_user(data, userPtr++))
			return -EFAULT;

		d = (data << 24) & 0xff000000; 	// 8bit -> 32 bit

		*p++ = d;

		if (sound.hard.stereo) {
			// Translate stereo -> mono
			if (get_user(data, userPtr++))
				return -EFAULT;

			d = (data << 24) & 0xff000000; 	// 8bit -> 32 bit
		}

		*p++ = d;

		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	telios_fill(0x00, p, (frameLeft - fUsed) >> 2);

	return (uUsed);
}


static ssize_t telios_ct_u8(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;

	u_int  *p = (u_int *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 1, frameLeft >> 3);
		uUsed = count * 2;
		fUsed = count * 8;
	} else {
		count = min(userCount, frameLeft >> 3);
		uUsed = count;
		fUsed = count * 8;
	}
	while (count > 0) {
		u_char data;
		u_int d;

		if (get_user(data, userPtr++))
			return -EFAULT;

		data ^= 0x80;
		d = (data << 24) & 0xffff0000; 	// 8bit -> 32 bit

		*p++ = d;

		if (sound.hard.stereo) {
			// Translate stereo -> mono
			if (get_user(data, userPtr++))
				return -EFAULT;

			data ^= 0x80;
			d = (data << 24) & 0xffff0000; 	// 8bit -> 32 bit
		}

		*p++ = d;

		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	telios_fill(0x0, p, (frameLeft - fUsed) >> 2);

	return (uUsed);
}

static ssize_t telios_ct_s16le(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	// uUsed & fUsed is byte count!!
	ssize_t count, uUsed, fUsed;
	u_int  *p = (u_int *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
	        count = min(userCount >> 2, frameLeft >> 3);
		uUsed = count * 4;
		fUsed = count * 8;
	} else {
		count = min(userCount >> 1, frameLeft >> 3);
		uUsed = count * 2;
		fUsed = count * 8;
	}
	if (sound.hard.stereo) 
	while (count > 0) {
		u_short data;
		u_int   d;

		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;

		d = (data << 16) & 0xffff0000; // 16bit -> 32 bit
		*p++ = d;
		
		if (sound.hard.stereo) {
		    if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;

		    d = (data << 16) & 0xffff0000; 	// 16bit -> 32 bit
		}

		*p++ = (u_int) d;

		count--;
	}

	*frameUsed += fUsed;
	// Clear rest of frame

	telios_fill(0x0, p, (frameLeft - fUsed) >> 2);
	return (uUsed);
}

static ssize_t telios_ct_u16le(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;
	u_int  *p = (u_int *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 2, frameLeft >> 3);
		uUsed = count * 4;
		fUsed = count * 8;
	} else {
		count = min(userCount >> 1, frameLeft >> 3);
		uUsed = count * 2;
		fUsed = count * 8;
	}
	while (count > 0) {
		u_short data;
		u_int   d;

		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;
		data ^= 0x8000;
		d = (data << 16) & 0xffffc000; // 16bit -> 32 bit
		*p++ = d;
		
		if (sound.hard.stereo) {
		    if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;
		    data ^= 0x8000; 
		    d = (data << 16) & 0xffffc000; 	// 16bit -> 32 bit
		}

		*p++ = d;

		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	telios_fill(0x0, p, (frameLeft - fUsed) >> 2);

	return(uUsed);
}

/* Convert Big Endian short -> Little Endian short */
static inline u_short be2le(u_short d)
{
	return (((d << 8) & 0xff00) + ((d >>8) & 0x00ff));
}

static ssize_t telios_ct_s16be(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;
	u_short  *p = (u_short *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 2, frameLeft >> 3);
		uUsed = count * 4;
		fUsed = count * 8;
	} else {
		count = min(userCount >> 1, frameLeft >> 3);
		uUsed = count * 2;
		fUsed = count * 8;
	}
	while (count > 0) {
		u_short data;
		u_int   d;

		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;

		data = be2le(data) ^ 0x8000;

		d = (data << 16) & 0xffff0000; // 16bit -> 32 bit
		*p++ = d;
		
		if (sound.hard.stereo) {
		    if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;
		    data = be2le(data) ^ 0x8000;
		    d = (data << 16) & 0xffff0000; 	// 16bit -> 32 bit
		}

		*p++ = d;

		count--;

	}
	*frameUsed += fUsed;
	// Clear rest of frame
	telios_fill(0x0, p, (frameLeft - fUsed) >> 2);

	return(uUsed);
}

static ssize_t telios_ct_u16be(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;
	u_short  *p = (u_short *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 2, frameLeft >> 3);
		uUsed = count * 4;
		fUsed = count * 8;
	} else {
		count = min(userCount >> 1, frameLeft >> 3);
		uUsed = count * 2;
		fUsed = count * 8;
	}
	while (count > 0) {
	        u_short data;
		u_int   d;

		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;

		data = be2le(data);

		d = (data << 16) & 0xffff0000; // 16bit -> 32 bit
		*p++ = d;
		
		if (sound.hard.stereo) {
		    if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;
		    data = be2le(data);
		    d = (data << 16) & 0xffff0000; 	// 16bit -> 32 bit
		}

		*p++ = d;

		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	telios_fill(0x0, p, (frameLeft - fUsed) >> 2);
	return (uUsed);
}

#endif /* CONFIG_SOUND_TELIOS */

#ifdef CONFIG_SOUND_TELIOS
static TRANS transTelios = {
	NULL, NULL, telios_ct_s8, telios_ct_u8,
	telios_ct_s16be, telios_ct_u16be, telios_ct_s16le, telios_ct_u16le
};
#endif /* CONFIG_SOUND_TELIOS */


/*** Low level stuff *********************************************************/

#ifdef CONFIG_SOUND_TELIOS

#define BETTY_IO_DATA_IODAT0   BIT(0) 
#define BETTY_IO_DATA_IODAT1   BIT(1) 
#define BETTY_IO_DATA_IODAT2   BIT(2) 
#define BETTY_IO_DATA_IODAT3   BIT(3) 
#define BETTY_IO_DATA_IODAT4   BIT(4) 
#define BETTY_IO_DATA_IODAT5   BIT(5) 
#define BETTY_IO_DATA_IODAT6   BIT(6) 
#define BETTY_IO_DATA_IODAT7   BIT(7) 
#define BETTY_IO_DATA_IODAT8   BIT(8) 
#define BETTY_IO_DATA_IODAT9   BIT(9) 

#define IRQ_CHI1_0INT 18
#define IRQ_CHI0_5INT 17

static void telios_set_rate(void)
{
        printk("telios_set_rate: %d\n", sound.soft.speed);
	sib_read_check(BETTY_REG_IO_PORT_DATA, read_af_ioport1);
	switch (sound.soft.speed) {
	case 8000:
	  sib_write_check(BETTY_REG_IO_PORT_DATA,
			  BETTY_IO_DATA_IODAT7 |
			  BETTY_IO_DATA_IODAT6 |
			  BETTY_IO_DATA_IODAT1);
		break;
	case 44100:
	  sib_write_check(BETTY_REG_IO_PORT_DATA,
			  BETTY_IO_DATA_IODAT7 |
			  BETTY_IO_DATA_IODAT6 |
			  BETTY_IO_DATA_IODAT4);
		break;
	case 22050:
	  sib_write_check(BETTY_REG_IO_PORT_DATA,
			  BETTY_IO_DATA_IODAT7 |
			  BETTY_IO_DATA_IODAT6 |
			  BETTY_IO_DATA_IODAT3);
		break;
	case 11025:
	  sib_write_check(BETTY_REG_IO_PORT_DATA,
			  BETTY_IO_DATA_IODAT7 |
			  BETTY_IO_DATA_IODAT6 |
			  BETTY_IO_DATA_IODAT2);
		break;
	default:
	  sib_write_check(BETTY_REG_IO_PORT_DATA,
			  BETTY_IO_DATA_IODAT7 |
			  BETTY_IO_DATA_IODAT6 |
			  BETTY_IO_DATA_IODAT1);
	  //		printk("DMA sound: Illegal sound rate %d\n", sound.soft.speed);
		break;
	}
	sib_read_check(BETTY_REG_IO_PORT_DATA, read_af_ioport2);
}

static void telios_enable_irq(void)
{
  IntEnable1 |= INT1_CHI1_0INT|INT1_CHI0_5INT;
}

static void telios_disable_irq(void)
{
  IntEnable1 &= ~(INT1_CHI1_0INT|INT1_CHI0_5INT);
}

static void telios_clear_irq(void)
{
  IntClear1 = INT1_CHI1_0INT|INT1_CHI0_5INT;
}

static void telios_enable_dma(void)
{
#if 0
	// Enable DMA in DCU
	*VR41XX_DMARSTREG |= 0x0001;	// Release DMA Reset
	barrier();
	*VR41XX_DMAMSKREG |= 0x0004;	// Enable Audio Output DMA
	barrier();
	*VR41XX_DMASENREG |= 0x0001;	// Enable DMA Sequencer
	barrier();
	// But DMA not started yet, need 'AUISEN := 1'
#endif
}

int playing = 0;

static void telios_disable_dma(void)
{
#if 0
	*VR41XX_DMAMSKREG &= 0xfffb;	// Disable Audio Output DMA
	barrier();
#endif
	CHISize &= ~CHI_ENDMATXCHI;
	playing = 0;
}

static void telios_suspend_dma(void)
{
#if 0
	*VR41XX_DMAMSKREG &= 0xfffb;	// Disable Audio Output DMA
#endif 
	CHISize &= ~CHI_ENDMATXCHI;
	playing = 0;
}

static void telios_resume_dma(void)
{
#if 0
  	*VR41XX_DMAMSKREG |= 0x0004;	// Enable Audio Output DMA
#endif
	if(!playing){
	  CHISize = (0xfff << CHI_CHISIZE_SHIFT) | CHI_ENDMATXCHI;
	  playing = 1;
#if 0
	  {
	      int i;
	      unsigned flags;

	      save_and_cli(flags);
	      for(i=0; i<1000*20; i++)
		  udelay(1000);
	      restore_flags(flags);
	{
	    u_int *buf = (u_int *)telios_sound_buffer;
	    int i, l, r;

	    l = 0; r = 0;
	    for( i=0; i< (16<<10)/sizeof(u_int); i+=2 ){
		if( ((i / 2) / 32) & 1 ){
		if( buf[i] != 0xc0000000 )
			ct_set(i,"NM");
		}else{
		if( buf[i] != 0x40000000 )
			ct_set(i,"NM");
		}
		if( ((i / 2) / 16) & 1 ){
		if( buf[i+1] != 0xc0000000 )
			ct_set(i,"NM");
		}else{
		if( buf[i+1] != 0x40000000 )
			ct_set(i,"NM");
		}
	    }
	}
	  }
#endif
	}
}

static void telios_speaker_on(void)
{
#if 0
    /* Speaker power ON via GPIO it's machine dependent */
	VR41XX_ENABLE_SPEAKER();
	barrier();
#endif
}

static void telios_disable_sound(void)
{
	telios_disable_dma();
	telios_clear_irq();
	// vr41xx_disable_irq();
#if 0
    /* Speaker operation disable */
    *VR41XX_SEQREG  = 0x0000;	// AUISEN := 0
    barrier();

    /* Speaker power OFF via GPIO it's machine dependent */
	VR41XX_DISABLE_SPEAKER();
	barrier();

    /* DAC vref Off */
    *VR41XX_SCNTREG &= 0x7fff;
    barrier();

    /* Disable AIU clock in CMU */
    *VR41XX_CMUCLKMSK &= 0xfffb;
    barrier();
#endif
}		

static void telios_kick_dma(void)
{
    /* conversion rate */
	telios_set_rate();

	/* Enable DMA */
	telios_enable_dma();

#if 0
    /* DAC Vref on */
	*VR41XX_SCNTREG |= 0x8000;
    barrier();

	/* Wait for Vref stablization */
	// udelay(10);
	mdelay(50);	// 50 msec is enough ?

	// Speaker Power On
	vr41xx_speaker_on();

    // Speaker operation enable
    *VR41XX_SEQREG  = 0x0001;	// AUISEN := 1

    barrier();
#endif
}

#endif /* CONFIG_SOUND_TELIOS */


#ifdef CONFIG_SOUND_TELIOS
/*
 * VR41XX
 */
static int __init TeliosIrqInit(void)
{
	int err;
	u_char *tmp;

	sib_write_check(BETTY_REG_IO_PORT_DIRECTION,0x03ff);

	/* turn off DMA for audio channels */
	telios_disable_dma();
	telios_clear_irq();
	telios_disable_sound();
	/* Register interrupt handler. */
	tmp = kmalloc (16 << 10, GFP_DMA);
	if(!tmp){
	        printk("Can't allocate memory for sound buffer!\n");
                return 0;
	}

	telios_sound_buffer = (u_char *)((u_long) tmp | KSEG1);

	telios_fill(0x0, telios_sound_buffer,  16 << 8);

	ct_set(77,"NI");
	{
	    u_int *buf = (u_int *)telios_sound_buffer;
	    int i, l, r;

	    l = 0; r = 0;
	    for( i=0; i< (16<<10)/sizeof(u_int); i+=2 ){
		if( ((i / 2) / 32) & 1 )
		    buf[i] = 0xc0000000;
		else
		    buf[i] = 0x40000000;
		if( ((i / 2) / 16) & 1 )
		    buf[i+1] = 0xc0000000;
		else
		    buf[i+1] = 0x40000000;
	    }
	}

	{
	    u_int *buf = (u_int *)telios_sound_buffer;
	    int i, l, r;

	    l = 0; r = 0;
	    for( i=0; i< (16<<10)/sizeof(u_int); i+=2 ){
		if( ((i / 2) / 32) & 1 )
		    buf[i] = 0xc0000000;
		else
		    buf[i] = 0x40000000;
		if( ((i / 2) / 16) & 1 )
		    buf[i+1] = 0xc0000000;
		else
		    buf[i+1] = 0x40000000;
	    }
	}

	//	printk("telios_sound_buffer=%08x(%d), size=%d\n",
	//	       telios_sound_buffer, telios_sound_buffer,  16 << 10);

	if ((err = request_irq(IRQ_CHI0_5INT, telios_sq_interrupt, SA_INTERRUPT,
			       "DMA sound1", telios_sq_interrupt))) {
		return(0);
	}

	if ((err = request_irq(IRQ_CHI1_0INT, telios_sq_interrupt, SA_INTERRUPT,
			       "DMA sound2", telios_sq_interrupt))) {
		free_irq(IRQ_CHI0_5INT, telios_sq_interrupt);
		return(0);
	}

	return(1);
}

#ifdef MODULE
static void TeliosIrqCleanUp(void)
{
	/* turn off DMA for audio channels */
	telios_disable_dma();
	telios_disable_irq();
	telios_clear_irq();
	telios_disable_sound();
	/* release the interrupt */
	free_irq(IRQ_CHI0_5INT, telios_sq_interrupt);
	free_irq(IRQ_CHI1_0INT, telios_sq_interrupt);
}
#endif /* MODULE */
#endif /* CONFIG_SOUND_TELIOS */


#ifdef CONFIG_SOUND_TELIOS
static void TeliosSilence(void)
{
	/* Disable sound & DMA */
	telios_disable_sound();
	telios_disable_irq();
}
#endif /* CONFIG_SOUND_TELIOS */

#ifdef CONFIG_SOUND_TELIOS
static char telios_first_dma_kick = 1;

static void TeliosInit(void)
{
        u_int addr;

	sound.hard = sound.soft;
	sound.trans = &transTelios;

	CHIControl &= ~CHI_ENCHI;

	ClockControl &= ~( CLK_CHICLKDIR | CLK_ENCHIMCLK );
	
	CHIControl = CHI_CHIFSDIR |
	  ((3 << CHI_CHIFSWIDTH_SHIFT) & CHI_CHIFSWIDTH_MASK) |
	  ((3 << CHI_CHINCHAN_SHIFT)   & CHI_CHINCHAN_MASK)   |
	  ((1 << CHI_CHITXBOFF_SHIFT)  & CHI_CHITXBOFF_MASK)  |
	  CHI_TXMSBFIRST | CHI_CHITXFSPOL |
	  CHI_CHITXEN | CHI_ENCHI
	  | CHI_CHIFSEDGE ;

	ct_set(CHIControl, "NC");

	CHIPointerEnable = 
	  CHI_CHITXPTRB3EN |CHI_CHITXPTRB2EN |CHI_CHITXPTRB1EN |CHI_CHITXPTRB0EN |
	  CHI_CHITXPTRA3EN |CHI_CHITXPTRA2EN |CHI_CHITXPTRA1EN |CHI_CHITXPTRA0EN;

        CHITransmitPointerA = 0x03020100;
        CHITransmitPointerB = 0x03020100;

	addr = PHYSADDR(telios_sound_buffer);
	addr = (u_int) addr & 0xfffffffc;
	CHITXStart = addr;
}
#endif /* CONFIG_SOUND_VR41XX */


#ifdef CONFIG_SOUND_TELIOS
static int TeliosSetFormat(int format)
{
	int size;

	/* Vr41xx sound DMA supports only 8bit (expanded to 10bit mode */
	switch (format) {
	case AFMT_QUERY:
		return(sound.soft.format);
	case AFMT_MU_LAW:
		size = 8;
		break;
	case AFMT_A_LAW:
		size = 8;
		break;
	case AFMT_U8:
		size = 8;
		break;
	case AFMT_S8:
		size = 8;
		break;
	case AFMT_S16_BE:
		size = 16;
		break;
	case AFMT_U16_BE:
		size = 16;
		break;
	case AFMT_S16_LE:
		size = 16;
		break;
	case AFMT_U16_LE:
		size = 16;
		break;
	default: /* :-) */
#ifdef DEBUG
		printk("default: AFMT_S8\n");
#endif
		size = 8;
		format = AFMT_S8;
	}

	sound.soft.format = format;
	sound.soft.size = size;
	if (sound.minDev == SND_DEV_DSP) {
		sound.dsp.format = format;
		sound.dsp.size = sound.soft.size;
	}

	return(format);
}
#endif /* CONFIG_SOUND_VR41XX */


#ifdef CONFIG_SOUND_TELIOS
#endif /* CONFIG_SOUND_TELIOS */


#ifdef CONFIG_SOUND_TELIOS

#define TELIOS_PLAY_LOADED		1
#define TELIOS_PLAY_PLAYING		2
#define TELIOS_PLAY_MASK		3

int half_point = 1;

static void telios_sq_play_next_frame(int index)
{
	u_char *start,*dest;
	u_long size;
	u_int addr;

	/* used by Vr41xxPlay() if all doubts whether there really is something
	 * to be played are already wiped out.
	 */

	start = (u_char *)sq_block_address(sq.front);
	ct_set(sq.front,"fr");
	//	ct_set(half_point,"hp");
	//size = (sq.count == index ? sq.rear_size : sq.block_size)>>1;
#if 1
	dest = (u_char *) telios_sound_buffer + (half_point ? 0 : 8192);
	ct_set(dest, "ac");
	memcpy(dest, start, 8192);
#else
	telios_suspend_dma();
	addr = PHYSADDR(start);
	addr = (u_int) addr & 0xfffffffc;
	CHITXStart = addr;
#endif 
	sq.front = (sq.front+1) % sq.max_count;
	//ct_set(sq.front,"fn");
	sq.active |= TELIOS_PLAY_LOADED;
	
	//	CHISize |= ((0xfff << CHI_CHISIZE_SHIFT) & CHI_CHISIZE_MASK);
	telios_resume_dma();

	telios_enable_irq();
	telios_clear_irq();
}

#endif /* CONFIG_SOUND_VR41XX */


#ifdef CONFIG_SOUND_TELIOS
static void TeliosPlay(void)
{
	// fixme
	int minframes = 1;

	telios_disable_irq();

	//	ct_set(sq.active, "st");
	if (sq.active & TELIOS_PLAY_LOADED) {
		/* There's already a frame loaded */
	        telios_enable_irq();
		//ct_set(00, "pl");
		return;
	}

	if (sq.active & TELIOS_PLAY_PLAYING){
		/* Increase threshold: frame 1 is already being played */
		minframes = 2;
		//		ct_set(00, "pp");
	}

	if (sq.count < minframes) {
		/* Nothing to do */
	  	telios_enable_irq();
		return;
	}

	if (sq.count <= minframes && sq.rear_size < sq.block_size && !sq.syncing) {
		/* hmmm, the only existing frame is not
		 * yet filled and we're not syncing?
		 */
	        telios_enable_irq();
		return;
	}

	telios_sq_play_next_frame(minframes);
}
#endif /* CONFIG_SOUND_VR41XX */

#ifdef CONFIG_SOUND_TELIOS
static void telios_sq_interrupt(int irq, void *dummy, struct pt_regs *fp)
{
	int minframes = 1;
	static int irq_old;

	static int high_old, low_old, high, low;

	telios_disable_irq();
	telios_clear_irq();

	tx3920_read_rtc(high,low);
	//	ct_set( low - low_old, "r1");

	low_old = low;
	high_old = high;

        if(irq == IRQ_CHI0_5INT)       half_point = 1;
	else if(irq == IRQ_CHI1_0INT)  half_point = 0;

	if(irq == irq_old){
	  ct_set(00,"er");
	}

	irq_old = irq;

	if (!sq.active) {
		/* Playing was interrupted and sq_reset() has already cleared
		 * the sq variables, so better don't do anything here.
		 */
		WAKE_UP(sq.sync_queue);
		telios_clear_irq();
		return;
	}

	if (sq.active & TELIOS_PLAY_PLAYING) {
		/* We've just finished a frame */
		sq.count--;
		ct_set(99,"sp");
	}

	if (sq.active & TELIOS_PLAY_LOADED){
		/* Increase threshold: frame 1 is already being played */
		minframes = 2;
	}

	/* Shift the flags */
       	sq.active = (sq.active << 1) & TELIOS_PLAY_MASK;

	if (!sq.active) {
		/* No frame is playing, disable audio DMA */
		telios_clear_irq();
		ct_set(00, "em");
		telios_suspend_dma();
		return;
	}

	if (sq.count >= minframes) {
		/* Try to play the next frame */
	        TeliosPlay();
	}

	if (!sq.active) {
		/* Nothing to play anymore.
		   Wake up a process waiting for audio output to drain. */
		WAKE_UP(sq.sync_queue);
		ct_set(00, "sq");
	}

	tx3920_read_rtc(high,low);
	//	ct_set( low - low_old, "r2");

}

#endif /* CONFIG_SOUND_VR41XX */


/*** Machine definitions *****************************************************/

#ifdef CONFIG_SOUND_TELIOS
static MACHINE machTelios = {
	DMASND_TELIOS,		// int type
	NULL,			// void *dma_alloc(uint, int)
	NULL,			// void  dma_free(void *, uint)
	TeliosIrqInit,		// void  irqinit(void)
#ifdef MODULE
	TeliosIrqCleanUp,	// void  irqcleanup(void)
#endif /* MODULE */
	TeliosInit,		// void  init(void)
        TeliosSilence,		// void  silence(void)
	TeliosSetFormat,	// int   setFormat(int)
	TeliosSetVolume,	// int   setVolume(int)
	NULL,			// int   setBass(int)
	NULL,			// int   setTreble(int)
	NULL,			// int   setGain(int)
	TeliosPlay		// void  play(void)
};
#endif /* CONFIG_SOUND_TELIOS */

/*** Mid level stuff *********************************************************/


static void sound_silence(void)
{
	/* update hardware settings one more */
	(*sound.mach.init)();

	(*sound.mach.silence)();
}


static void sound_init(void)
{
	(*sound.mach.init)();
}


static int sound_set_format(int format)
{
	return(*sound.mach.setFormat)(format);
}


static int sound_set_speed(int speed)
{
	if (speed < 0)
		return(sound.soft.speed);

	sound.soft.speed = speed;
	(*sound.mach.init)();
	if (sound.minDev == SND_DEV_DSP)
		sound.dsp.speed = sound.soft.speed;

	//	telios_set_rate();

	return(sound.soft.speed);
}


static int sound_set_stereo(int stereo)
{
	if (stereo < 0)
		return(sound.soft.stereo);
	stereo = !!stereo;    /* should be 0 or 1 now */

	sound.soft.stereo = stereo;
	if (sound.minDev == SND_DEV_DSP)
		sound.dsp.stereo = stereo;
	(*sound.mach.init)();

	return(stereo);
}

#ifdef CONFIG_SOUND_VR41XX
	// fixme
#endif
#ifdef CONFIG_SOUND_TELIOS
#define VOLUME_VOXWARE_TO_TELIOS(v) \
        (((v) < 0) ? 0 : ((v) > 100) ? 32 : ((v) * 32)/100)
#define VOLUME_TELIOS_TO_VOXWARE(v) ((v)*100/32)

static int TeliosSetVolume(int volume)
{
    sound.volume_left = VOLUME_VOXWARE_TO_TELIOS(volume & 0xff);
    sound.volume_right = VOLUME_VOXWARE_TO_TELIOS((volume & 0xff00) >> 8);

    Plum2_DispLcdLevel = (sound.volume_left + sound.volume_right) >> 1;

    return VOLUME_TELIOS_TO_VOXWARE(sound.volume_left) |
      (VOLUME_TELIOS_TO_VOXWARE(sound.volume_right) << 8);

}
//#else
static int sound_set_volume(int volume)
{
	return(*sound.mach.setVolume)(volume);
}
#endif


static ssize_t sound_copy_translate(const u_char *userPtr,
				    size_t userCount,
				    u_char frame[], ssize_t *frameUsed,
				    ssize_t frameLeft)
{
	ssize_t
		(*ct_func)(const u_char *, size_t, u_char *, ssize_t *, ssize_t) = NULL;

	switch (sound.soft.format) {
	case AFMT_MU_LAW:
		ct_func = sound.trans->ct_ulaw;
		break;
	case AFMT_A_LAW:
		ct_func = sound.trans->ct_alaw;
		break;
	case AFMT_S8:
		ct_func = sound.trans->ct_s8;
		break;
	case AFMT_U8:
		ct_func = sound.trans->ct_u8;
		break;
	case AFMT_S16_BE:
		ct_func = sound.trans->ct_s16be;
		break;
	case AFMT_U16_BE:
		ct_func = sound.trans->ct_u16be;
		break;
	case AFMT_S16_LE:
		ct_func = sound.trans->ct_s16le;
		break;
	case AFMT_U16_LE:
		ct_func = sound.trans->ct_u16le;
		break;
	}
	if (ct_func)
		return ct_func(userPtr, userCount, frame, frameUsed, frameLeft);
	else
		return 0;
}


/*
 * /dev/mixer abstraction
 */


#define RECLEVEL_VOXWARE_TO_GAIN(v) \
	((v) < 0 ? 0 : (v) > 100 ? 15 : (v) * 3 / 20)
#define RECLEVEL_GAIN_TO_VOXWARE(v) (((v) * 20 + 2) / 3)


static int mixer_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	mixer.busy = 1;
	return 0;
}


static int mixer_release(struct inode *inode, struct file *file)
{
	mixer.busy = 0;
	MOD_DEC_USE_COUNT;

#ifdef CONFIG_SOUND_TELIOS
	telios_disable_sound();
#endif
	return 0;
}


static int mixer_ioctl(struct inode *inode, struct file *file, u_int cmd,
		       u_long arg)
{
        int data;

	if (_SIOC_DIR(cmd) & _SIOC_WRITE)
	    mixer.modify_counter++;
	if (cmd == OSS_GETVERSION)
	    return IOCTL_OUT(arg, SOUND_VERSION);
	switch (sound.mach.type) {
#ifdef CONFIG_SOUND_TELIOS
	case DMASND_TELIOS:
		switch (cmd) {
		case SOUND_MIXER_INFO: {
		    mixer_info info;
		    strncpy(info.id, "Telios", sizeof(info.id));
		    strncpy(info.name, "Telios", sizeof(info.name));
		    info.name[sizeof(info.name)-1] = 0;
		    info.modify_counter = mixer.modify_counter;
		    copy_to_user_ret((int *)arg, &info, sizeof(info), -EFAULT);
		    return 0;
		}
		case SOUND_MIXER_READ_DEVMASK:
		  return IOCTL_OUT(arg, SOUND_MASK_VOLUME);

#if 0 
			return IOCTL_OUT(arg, 0);	// fixme
#endif
		case SOUND_MIXER_READ_RECMASK:
			return IOCTL_OUT(arg, 0);
		case SOUND_MIXER_READ_STEREODEVS:
			return IOCTL_OUT(arg, 0);
		case SOUND_MIXER_READ_VOLUME:
		  return IOCTL_OUT(arg,
				   VOLUME_TELIOS_TO_VOXWARE(sound.volume_left) |
				   (VOLUME_TELIOS_TO_VOXWARE(sound.volume_right) <<
 8));
#if 0
		  return IOCTL_OUT(arg, 0);	// fixme
#endif
		case SOUND_MIXER_WRITE_VOLUME:
#if 0
			return IOCTL_OUT(arg, 0);	// fixme
#endif
			IOCTL_IN(arg, data);
			return IOCTL_OUT(arg, sound_set_volume(data));

		case SOUND_MIXER_READ_TREBLE:
			return IOCTL_OUT(arg, 0);	// fixme
		case SOUND_MIXER_WRITE_TREBLE:
			return IOCTL_OUT(arg, 0);	// fixme
		}
		break;

#endif /* CONFIG_SOUND_VR41XX */
	}

	return -EINVAL;
}


static struct file_operations mixer_fops =
{
	llseek:  sound_lseek,
	ioctl:   mixer_ioctl,
	open:    mixer_open,
	release: mixer_release
};

static void __init mixer_init(void)
{
#ifndef MODULE
	int mixer_unit;
#endif
	mixer_unit = register_sound_mixer(&mixer_fops, -1);
	if (mixer_unit < 0)
		return;

	mixer.busy = 0;
	sound.treble = 0;
	sound.bass = 0;
	switch (sound.mach.type) {
	case DMASND_TELIOS:
		sound.volume_left = 64;
		sound.volume_right = 64;
		// nothing to do ?? fixme
		break;
	}
}


/*
 * Sound queue stuff, the heart of the driver
 */


static int sq_allocate_buffers(void)
{
	int i;
	u_char **t;

	if (sound_buffers)
		return 0;
	//	printk("soud_buff size = %x\n", numBufs * sizeof(char *));
	sound_buffers = kmalloc (numBufs * sizeof(char *), GFP_KERNEL);
	//	printk("soud_buff = %08x\n", sound_buffers);
	if (!sound_buffers)
		return -ENOMEM;

	//	printk("t size = %x\n", numBufs * sizeof(char *));
	t = kmalloc (numBufs * sizeof(char *), GFP_KERNEL);
	if (!t)
		return -ENOMEM;
	//	printk("t = %8x\n", t);

	for (i = 0; i < numBufs; i++) {
	  //	  printk("t[%d] size = %x\n", i, bufSize << 10);
                t[i] = kmalloc (bufSize << 10, GFP_DMA);
		//	  printk("t[%d]  = %08x\n", i, t[i]);

		if (!t[i]/*sound_buffers[i]*/) {
		  while (i--)
			  kfree(t[i]);
			  //sound.mach.dma_free (sound_buffers[i], bufSize << 10);
			  //kfree(sound_buffers[i]);

			kfree (t);
			kfree (sound_buffers);
			sound_buffers = 0;
			return -ENOMEM;
		}

		sound_buffers[i] = (u_char *)((u_long) t[i] | KSEG1);
		//		printk("sound_buffers[%d] = %x \n", i, sound_buffers[i]);
	}

	kfree(t);
	return 0;
}


static void sq_release_buffers(void)
{
	int i;

	if (sound_buffers) {
		for (i = 0; i < numBufs; i++)
		  //			sound.mach.dma_free (sound_buffers[i], bufSize << 10);
		  kfree((u_char *)
			(PHYSADDR((u_long)sound_buffers[i]) | KSEG0));

		kfree (sound_buffers);
		sound_buffers = 0;
	}

}

static void sq_setup(int numBufs, int bufSize, char **write_buffers)
{
	sq.max_count = numBufs;
	sq.max_active = numBufs;
	sq.block_size = bufSize;
	sq.buffers = write_buffers;

	sq.front = sq.count = 0;
	sq.rear = -1;
	sq.syncing = 0;
	sq.active = 0;

}



static void sq_play(void)
{
	(*sound.mach.play)();
}


/* ++TeSche: radically changed this one too */

static ssize_t sq_write(struct file *file, const char *src, size_t uLeft,
			loff_t *ppos)
{
	ssize_t uWritten = 0;
	u_char *dest;
	ssize_t uUsed, bUsed, bLeft;

	ct_set(sq.count, "si");

	/* ++TeSche: Is something like this necessary?
	 * Hey, that's an honest question! Or does any other part of the
	 * filesystem already checks this situation? I really don't know.
	 */
	if (uLeft == 0)
		return 0;
	/* The interrupt doesn't start to play the last, incomplete frame.
	 * Thus we can append to it without disabling the interrupts! (Note
	 * also that sq.rear isn't affected by the interrupt.)
	 */
	if (sq.count > 0 && (bLeft = sq.block_size-sq.rear_size) > 0) {
	        dest = sq_block_address(sq.rear);
		bUsed = sq.rear_size;
		uUsed = sound_copy_translate(src, uLeft, dest, &bUsed, bLeft);
		if (uUsed <= 0)
			return uUsed;
		src += uUsed;
		uWritten += uUsed;
		uLeft -= uUsed;
		sq.rear_size = bUsed;
	}

	do {

		while (sq.count == sq.max_active) {
		  ct_set(sq.count, "sf");
			/* All frames are in use.  -- kei */
			sq_play();
			if (NON_BLOCKING(sq.open_mode))
				return uWritten > 0 ? uWritten : -EAGAIN;
			SLEEP(sq.action_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)
				return uWritten > 0 ? uWritten : -EINTR;
		}

		/* Here, we can avoid disabling the interrupt by first
		 * copying and translating the data, and then updating
		 * the sq variables. Until this is done, the interrupt
		 * won't see the new frame and we can work on it
		 * undisturbed.
		 */

		dest = sq_block_address((sq.rear+1) % sq.max_count);
		bUsed = 0;
		bLeft = sq.block_size;
		uUsed = sound_copy_translate(src, uLeft, dest, &bUsed, bLeft);
		if (uUsed <= 0)
			break;
		src += uUsed;
		uWritten += uUsed;
		uLeft -= uUsed;
		if (bUsed) {
			// Need to be atomic -- kei
		  //disable_irq(IRQ_CHI1_0INT);
		    	//disable_irq(IRQ_CHI0_5INT);
    		        telios_disable_irq();
			ct_set(sq.rear, "sr");
		        sq.rear = (sq.rear+1) % sq.max_count;
			sq.rear_size = bUsed;
			sq.count++;
			telios_enable_irq();
		    	//enable_irq(IRQ_CHI1_0INT);
		    	//enable_irq(IRQ_CHI0_5INT);
		}
	} while (bUsed);   /* uUsed may have been 0 */      
	ct_set(sq.count, "so");
	sq_play();
	ct_set(uLeft, "ul");
	ct_set(uWritten, "uw");
	return uUsed < 0? uUsed: uWritten;
}

/***********/


static int sq_open(struct inode *inode, struct file *file)
{
	int rc = 0;

	MOD_INC_USE_COUNT;
	if (file->f_mode & FMODE_WRITE) {
		if (sq.busy) {
			rc = -EBUSY;
			if (NON_BLOCKING(file->f_flags))
				goto err_out;
			rc = -EINTR;
			while (sq.busy) {
				SLEEP(sq.open_queue, ONE_SECOND);
				if (SIGNAL_RECEIVED)
					goto err_out;
			}
		}
		sq.busy = 1; /* Let's play spot-the-race-condition */

		if (sq_allocate_buffers()) goto err_out_nobusy;
		printk("allocated\n");
		sq_setup(numBufs, bufSize<<10,sound_buffers);
		sq.open_mode = file->f_mode;
	}
	sound.minDev = MINOR(inode->i_rdev) & 0x0f;
	sound.soft = sound.dsp;
	sound.hard = sound.dsp;
	sound_init();
	if ((MINOR(inode->i_rdev) & 0x0f) == SND_DEV_AUDIO) {
		sound_set_speed(8000);
		sound_set_stereo(0);
		sound_set_format(AFMT_MU_LAW);
	}

	return 0;

err_out_nobusy:
	if (file->f_mode & FMODE_WRITE) {
		sq.busy = 0;
		WAKE_UP(sq.open_queue);
	}
	printk("err_out_nobusy\n");
err_out:
	MOD_DEC_USE_COUNT;

	return rc;
}


static void sq_reset(void)
{
	sound_silence();
	sq.active = 0;
	sq.count = 0;
	sq.front = (sq.rear+1) % sq.max_count;
}


static int sq_fsync(struct file *filp, struct dentry *dentry)
{
	int rc = 0;

	sq.syncing = 1;
	sq_play();	/* there may be an incomplete frame waiting */

	while (sq.active) {
		SLEEP(sq.sync_queue, ONE_SECOND);
		if (SIGNAL_RECEIVED) {
			/* While waiting for audio output to drain, an
			 * interrupt occurred.  Stop audio output immediately
			 * and clear the queue. */
		  ct_set(sq.front,"rs");
			sq_reset();
			rc = -EINTR;
			break;
		}
	}

	sq.syncing = 0;
	return rc;
}

static int sq_release(struct inode *inode, struct file *file)
{
	int rc = 0;

	if (sq.busy)
		rc = sq_fsync(file, file->f_dentry);
	sound.soft = sound.dsp;
	sound.hard = sound.dsp;
	sound_silence();

	sq_release_buffers();
	MOD_DEC_USE_COUNT;

	/* There is probably a DOS atack here. They change the mode flag. */
	/* XXX add check here */
	if (file->f_mode & FMODE_WRITE) {
		sq.busy = 0;
		WAKE_UP(sq.open_queue);
	}

	/* Wake up a process waiting for the queue being released.
	 * Note: There may be several processes waiting for a call
	 * to open() returning. */

	return rc;
}


static int sq_ioctl(struct inode *inode, struct file *file, u_int cmd,
		    u_long arg)
{
	u_long fmt;
	int data;
	int size, nbufs;

	switch (cmd) {
	case SNDCTL_DSP_RESET:
		sq_reset();
		return 0;
	case SNDCTL_DSP_POST:
	case SNDCTL_DSP_SYNC:
		return sq_fsync(file, file->f_dentry);

		/* ++TeSche: before changing any of these it's
		 * probably wise to wait until sound playing has
		 * settled down. */
	case SNDCTL_DSP_SPEED:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_speed(data));
	case SNDCTL_DSP_STEREO:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_stereo(data));
	case SOUND_PCM_WRITE_CHANNELS:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_stereo(data-1)+1);
	case SNDCTL_DSP_SETFMT:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_format(data));
	case SNDCTL_DSP_GETFMTS:
		fmt = 0;
		if (sound.trans) {
			if (sound.trans->ct_ulaw)
				fmt |= AFMT_MU_LAW;
			if (sound.trans->ct_alaw)
				fmt |= AFMT_A_LAW;
			if (sound.trans->ct_s8)
				fmt |= AFMT_S8;
			if (sound.trans->ct_u8)
				fmt |= AFMT_U8;
			if (sound.trans->ct_s16be)
				fmt |= AFMT_S16_BE;
			if (sound.trans->ct_u16be)
				fmt |= AFMT_U16_BE;
			if (sound.trans->ct_s16le)
				fmt |= AFMT_S16_LE;
			if (sound.trans->ct_u16le)
				fmt |= AFMT_U16_LE;
		}
		return IOCTL_OUT(arg, fmt);
	case SNDCTL_DSP_GETBLKSIZE:
		size = sq.block_size
			* sound.soft.size * (sound.soft.stereo + 1)
			/ (sound.hard.size * (sound.hard.stereo + 1));
		return IOCTL_OUT(arg, size);
	case SNDCTL_DSP_SUBDIVIDE:
		break;
	case SNDCTL_DSP_SETFRAGMENT:
		if (sq.count || sq.active || sq.syncing)
			return -EINVAL;
		IOCTL_IN(arg, size);
		nbufs = size >> 16;
		if (nbufs < 2 || nbufs > numBufs)
			nbufs = numBufs;
		size &= 0xffff;
		if (size >= 8 && size <= 29) {
			size = 1 << size;
			size *= sound.hard.size * (sound.hard.stereo + 1);
			size /= sound.soft.size * (sound.soft.stereo + 1);
			if (size > (bufSize << 10))
				size = bufSize << 10;
		} else
			size = bufSize << 10;
		sq_setup(numBufs, size, sound_buffers);
		sq.max_active = nbufs;
		return 0;

	default:
		return mixer_ioctl(inode, file, cmd, arg);
	}
	return -EINVAL;
}



static struct file_operations sq_fops =
{
	llseek:  sound_lseek,
	write:   sq_write,
	ioctl:   sq_ioctl,
	open:    sq_open,
	release: sq_release
};


static void __init sq_init(void)
{
#ifndef MODULE
	int sq_unit;
#endif
	sq_unit = register_sound_dsp(&sq_fops, -1);
	if (sq_unit < 0)
		return;

	init_waitqueue_head(&sq.action_queue);
	init_waitqueue_head(&sq.open_queue);
	init_waitqueue_head(&sq.sync_queue);

	sq.busy = 0;

	/* whatever you like as startup mode for /dev/dsp,
	 * (/dev/audio hasn't got a startup mode). note that
	 * once changed a new open() will *not* restore these!
	 */
	sound.dsp.format = AFMT_U8;
	sound.dsp.stereo = 0;
	sound.dsp.size = 8;

	/* set minimum rate possible without expanding */
	switch (sound.mach.type) {
	case DMASND_TELIOS:
		sound.dsp.speed = 8000;
		break;
	}

	/* before the first open to /dev/dsp this wouldn't be set */
	sound.soft = sound.dsp;
	sound.hard = sound.dsp;

	sound_silence();
}

/*
 * /dev/sndstat
 */


/* state.buf should not overflow! */

static int state_open(struct inode *inode, struct file *file)
{
	char *buffer = state.buf, *mach = "";
	int len = 0;

	if (state.busy)
		return -EBUSY;

	MOD_INC_USE_COUNT;
	state.ptr = 0;
	state.busy = 1;

	switch (sound.mach.type) {
	case DMASND_TELIOS:
		mach = "Telios ";
		break;
	}
	len += sprintf(buffer+len, "%sDMA sound driver:\n", mach);

	len += sprintf(buffer+len, "\tsound.format = 0x%x", sound.soft.format);
	switch (sound.soft.format) {
	case AFMT_MU_LAW:
		len += sprintf(buffer+len, " (mu-law)");
		break;
	case AFMT_A_LAW:
		len += sprintf(buffer+len, " (A-law)");
		break;
	case AFMT_U8:
		len += sprintf(buffer+len, " (unsigned 8 bit)");
		break;
	case AFMT_S8:
		len += sprintf(buffer+len, " (signed 8 bit)");
		break;
	case AFMT_S16_BE:
		len += sprintf(buffer+len, " (signed 16 bit big)");
		break;
	case AFMT_U16_BE:
		len += sprintf(buffer+len, " (unsigned 16 bit big)");
		break;
	case AFMT_S16_LE:
		len += sprintf(buffer+len, " (signed 16 bit little)");
		break;
	case AFMT_U16_LE:
		len += sprintf(buffer+len, " (unsigned 16 bit little)");
		break;
	}
	len += sprintf(buffer+len, "\n");
	len += sprintf(buffer+len, "\tsound.speed = %dHz (phys. %dHz)\n",
		       sound.soft.speed, sound.hard.speed);
	len += sprintf(buffer+len, "\tsound.stereo = 0x%x (%s)\n",
		       sound.soft.stereo, sound.soft.stereo ? "stereo" : "mono");
	switch (sound.mach.type) {
	}
	len += sprintf(buffer+len, "\tsq.block_size = %d sq.max_count = %d"
		       " sq.max_active = %d\n",
		       sq.block_size, sq.max_count, sq.max_active);
	len += sprintf(buffer+len, "\tsq.count = %d sq.rear_size = %d\n", sq.count,
		       sq.rear_size);
	len += sprintf(buffer+len, "\tsq.active = %d sq.syncing = %d\n",
		       sq.active, sq.syncing);
	state.len = len;
	return 0;
}


static int state_release(struct inode *inode, struct file *file)
{
	state.busy = 0;
	MOD_DEC_USE_COUNT;
	return 0;
}


static ssize_t state_read(struct file *file, char *buf, size_t count,
			  loff_t *ppos)
{
	int n = state.len - state.ptr;
	if (n > count)
		n = count;
	if (n <= 0)
		return 0;
	if (copy_to_user(buf, &state.buf[state.ptr], n))
		return -EFAULT;
	state.ptr += n;
	return n;
}


static struct file_operations state_fops =
{
	llseek:  sound_lseek,
	read:    state_read,
	open:    state_open,
	release: state_release
};


static void __init state_init(void)
{
#ifndef MODULE
	int state_unit;
#endif
	state_unit = register_sound_special(&state_fops, SND_DEV_STATUS);
	if (state_unit < 0)
		return;
	state.busy = 0;
}


/*** Common stuff ********************************************************/

static loff_t sound_lseek(struct file *file, long long offset, int orig)
{
	return -ESPIPE;
}


/*** Config & Setup **********************************************************/

void __init dmasound_init(void)
{
	int has_sound = 0;
	sound.mach = machTelios;
	has_sound = 1;

	if (!has_sound)
		return;

	/* Set up sound queue, /dev/audio and /dev/dsp. */
	ct_set(CHIControl, "NC");
	CHIControl &= ~CHI_ENCHI;

	sib_read_check(BETTY_REG_IO_PORT_DATA, read_af_ioport);

	sib_write_check(BETTY_REG_IO_PORT_DATA,
			BETTY_IO_DATA_IODAT9 |
			BETTY_IO_DATA_IODAT7 |
			BETTY_IO_DATA_IODAT6 |
			BETTY_IO_DATA_IODAT4);

	sib_read_check(BETTY_REG_IO_PORT_DATA, read_af_ioport2);

	/* Set default settings. */
	sq_init();

	/* Set up /dev/sndstat. */
	state_init();

	/* Set up /dev/mixer. */
	mixer_init();
	if (!sound.mach.irqinit()) {
		printk(KERN_ERR "DMA sound driver: Interrupt initialization failed\n");
		return;
	}
#ifdef MODULE
	irq_installed = 1;
#endif

	printk(KERN_INFO "DMA sound driver installed, using %d buffers of %dk.\n",
	       numBufs, bufSize);

	return;
}


#define MAXARGS		8	/* Should be sufficient for now */

void __init dmasound_setup(char *str, int *ints)
{
	/* check the bootstrap parameter for "dmasound=" */

	switch (ints[0]) {
	case 3:
		if ((ints[3] < 0) || (ints[3] > MAX_CATCH_RADIUS))
			printk("dmasound_setup: illegal catch radius, using default = %d\n", catchRadius);
		else
			catchRadius = ints[3];
		/* fall through */
	case 2:
		if (ints[1] < MIN_BUFFERS)
			printk("dmasound_setup: illegal number of buffers, using default = %d\n", numBufs);
		else
			numBufs = ints[1];
		if (ints[2] < MIN_BUFSIZE || ints[2] > MAX_BUFSIZE)
			printk("dmasound_setup: illegal buffer size, using default = %d\n", bufSize);
		else
			bufSize = ints[2];
		break;
	case 0:
		break;
	default:
		printk("dmasound_setup: illegal number of arguments\n");
	}
}


#ifdef MODULE

int init_module(void)
{
	dmasound_init();
	return 0;
}


void cleanup_module(void)
{
	if (irq_installed) {
		sound_silence();
		sound.mach.irqcleanup();
	}

	sq_release_buffers();

	if (mixer_unit >= 0)
		unregister_sound_mixer(mixer_unit);
	if (state_unit >= 0)
		unregister_sound_special(state_unit);
	if (sq_unit >= 0)
		unregister_sound_dsp(sq_unit);
}

#endif /* MODULE */

#ifndef MODULE
__initcall(dmasound_init);
#endif

#undef DEBUG

/* vim:set ts=4:set sw=4: */
