#ifndef __ASM_MIPS_PLUM2_H__
#define __ASM_MIPS_PLUM2_H__

#include <asm/addrspace.h>

#define PLUM2_REGISTER_BASE	(KSEG1 + 0x10800000)

#ifndef _LANGUAGE_ASSEMBLY
#define PLUM2_REG_AT(x)		(*((volatile u32 *)(PLUM2_REGISTER_BASE + x)))
#else
#define PLUM2_REG_AT(x)		(PLUM2_REGISTER_BASE + x)
#endif

#ifndef BIT
#define BIT(x)			(1 << x)
#endif

#define Plum2_McsEnable		PLUM2_REG_AT(0x700c)

#define PLUM2_RMCS_ENABLE	BIT(0)

/*
 * Display Controller
 */

#define PLUM2_DISP_CTRL_BASE	0x10801000
#define PLUM2_DISP_LCD_BASE	0x6c400000
#define PLUM2_DISP_CRT_BASE	0x6c401000
#define PLUM2_DISP_VRAM_BASE	0x6c000000

#define Plum2_DispIntEnable	PLUM2_REG_AT(0x1000)
#define Plum2_DispIntStatus	PLUM2_REG_AT(0x1004)
#define Plum2_DispBufCtrl	PLUM2_REG_AT(0x1008)
#define Plum2_DispVramCtrl	PLUM2_REG_AT(0x100c)
#define Plum2_DispVramRefresh	PLUM2_REG_AT(0x1010)
#define Plum2_DispLcdClock	PLUM2_REG_AT(0x1014)
#define Plum2_DispCrtClock	PLUM2_REG_AT(0x1018)
#define Plum2_DispPllCtrl	PLUM2_REG_AT(0x101c)

#define Plum2_DispLcdCtrl	PLUM2_REG_AT(0x1040)
#define Plum2_DispStnCtrl	PLUM2_REG_AT(0x1044)
#define Plum2_DispLcdLevel	PLUM2_REG_AT(0x1048)
#define Plum2_DispLcdLuminance	PLUM2_REG_AT(0x104c)

#define Plum2_DispDstnDitherBuf	PLUM2_REG_AT(0x1050)
#define Plum2_DispDstnOffscrBuf	PLUM2_REG_AT(0x1054)

#define Plum2_DispDacCtrl	PLUM2_REG_AT(0x1060)
#define Plum2_DispCrtBorder	PLUM2_REG_AT(0x1064)
#define Plum2_DispPaletteSnoop	PLUM2_REG_AT(0x1068)

#define Plum2_DispGraphicTest	PLUM2_REG_AT(0x107c)

#define Plum2_DispLcdHTotal	PLUM2_REG_AT(0x1080)
#define Plum2_DispLcdHStart	PLUM2_REG_AT(0x1084)
#define Plum2_DispLcdHSync	PLUM2_REG_AT(0x1088)
#define Plum2_DispLcdHBlank	PLUM2_REG_AT(0x108c)
#define Plum2_DispLcdHPixel	PLUM2_REG_AT(0x1090)

#define Plum2_DispLcdVTotal	PLUM2_REG_AT(0x1094)
#define Plum2_DispLcdVStart	PLUM2_REG_AT(0x1098)
#define Plum2_DispLcdVSync	PLUM2_REG_AT(0x109c)
#define Plum2_DispLcdVBlank	PLUM2_REG_AT(0x10a0)

#define Plum2_DispLcdLine	PLUM2_REG_AT(0x10a8)
#define Plum2_DispLcdIntLine	PLUM2_REG_AT(0x10ac)
#define Plum2_DispLcdPolarity	PLUM2_REG_AT(0x10b0)

#define Plum2_DispLcdCtrlTest	PLUM2_REG_AT(0x10bc)
#define Plum2_DispLcdBufSelect	PLUM2_REG_AT(0x10c0)
#define Plum2_DispLcdStartAddr0	PLUM2_REG_AT(0x10c4)
#define Plum2_DispLcdStartAddr1 PLUM2_REG_AT(0x10c8)
#define Plum2_DispLcdVramPitch1	PLUM2_REG_AT(0x10cc)
#define Plum2_DispLcdVramPitch2	PLUM2_REG_AT(0x10d0)
#define Plum2_DispLcdVramOffset	PLUM2_REG_AT(0x10d4)
#define Plum2_DispLcdLowOffest	PLUM2_REG_AT(0x10d8)
#define Plum2_DispLcdMode	PLUM2_REG_AT(0x10dc)

/*
 * Smart Media Controller
 */

/*
 * USB Host Controller
 */


/*
 * PCMCIA Controller
 */

#define PLUM2_PCMCIA_SOCKETS	2

extern void plum2pcc_pccard_shutdown(void);
extern void plum2pcc_pccard_startup(void);


/*
 * Power and Clock Control
 */

#define Plum2_PowerCtrl		PLUM2_REG_AT(0x7000)
#define Plum2_ClockCtrl		PLUM2_REG_AT(0x7004)

#define PLUM2_POWER_LCDDSP	BIT(0)
#define PLUM2_POWER_LCDPWR	BIT(1)
#define PLUM2_POWER_BKLIGHT	BIT(2)
#define PLUM2_POWER_IO5PWR	BIT(3)
#define PLUM2_POWER_EXTPW0	BIT(4)
#define PLUM2_POWER_EXTPW1	BIT(5)
#define PLUM2_POWER_EXTPW2	BIT(6)
#define PLUM2_POWER_LCDEN	BIT(8)
#define PLUM2_POWER_IO5EN	BIT(9)
#define PLUM2_POWER_USBEN	BIT(10)

/*
 * Interrupt Controller
 */

#define PLUM2_IRQ_BASE		32
#define PLUM2_IRQ_MAX		8

#define PLUM2_CSC_IRQ		32
#define PLUM2_PCC1_IRQ		33
#define PLUM2_PCC2_IRQ		34
#define PLUM2_DISP_IRQ		35
#define PLUM2_USB_IRQ		36
#define PLUM2_USBWAKE_IRQ	37
#define PLUM2_SM_IRQ		38
#define PLUM2_EXT_IRQ		39

#define PLUM2_INTSTATUS		0x8000

#define Plum2_IntStatus		PLUM2_REG_AT(PLUM2_INTSTATUS)

#define PLUM2_PCC_INT		BIT(0)
#define PLUM2_CSC_INT		(BIT(2)|BIT(1))
#define PLUM2_C1SC_INT		BIT(1)
#define PLUM2_C2SC_INT		BIT(2)
#define PLUM2_DISP_INT		BIT(3)
#define PLUM2_USB_INT		BIT(4)
#define PLUM2_USBWAKE_INT	BIT(5)
#define PLUM2_SM_INT		BIT(6)
#define PLUM2_EXT_INT		BIT(7)

/* Interrupt Enable Register */
#define PLUM2_INTENABLE		0x8010

#define Plum2_IntEnable		PLUM2_REG_AT(PLUM2_INTENABLE)

#define PLUM2_INT_ENABLE	BIT(0)

/* External */
#define PLUM2_EXT_INTSTATUS_RAW	0x8100
#define PLUM2_EXT_INTSTATUS	0x8104
#define PLUM2_EXT_INTCLEAR	0x8104
#define PLUM2_EXT_INTENABLE	0x8110

#define Plum2_ExtIntStatusRaw	PLUM2_REG_AT(PLUM2_EXT_INTSTATUS_RAW)
#define Plum2_ExtIntStatus	PLUM2_REG_AT(PLUM2_EXT_INTSTATUS)
#define Plum2_ExtIntClear	PLUM2_REG_AT(PLUM2_EXT_INTCLEAR)
#define Plum2_ExtIntEnable	PLUM2_REG_AT(PLUM2_EXT_INTENABLE)

#define PLUM2_EXT_CLEAR_MASK	BIT(0)
#define PLUM2_EXT_ENABLE_MASK	(BIT(8)|BIT(0))

/* PCMCIA */
#define PLUM2_PCC_INTSTATUS_RAW	0x8200
#define PLUM2_PCC_INTSTATUS	0x8204
#define PLUM2_PCC_INTCLEAR	0x8200
#define PLUM2_PCC_INTENABLE	0x8210
#define PLUM2_PCC_CLKSELECT	0x8220

#define Plum2_PccIntStatusRaw	PLUM2_REG_AT(PLUM2_PCC_INTSTATUS_RAW)
#define Plum2_PccIntStatus	PLUM2_REG_AT(PLUM2_PCC_INTSTATUS)
#define Plum2_PccIntClear	PLUM2_REG_AT(PLUM2_PCC_INTCLEAR)
#define Plum2_PccIntEnable	PLUM2_REG_AT(PLUM2_PCC_INTENABLE)
#define Plum2_PccClkSelect	PLUM2_REG_AT(PLUM2_PCC_CLKSELECT)

#define PLUM2_PCC1_INT		(BIT(1)|BIT(0))
#define PLUM2_PCC2_INT		(BIT(3)|BIT(2))

#define PLUM2_PCC1_INT_CLEAR		(BIT(1)|BIT(0))
#define PLUM2_PCC1_INT_ENABLE		(BIT(1)|BIT(0))
#define PLUM2_PCC1_INT_INPUT_ENABLE	(BIT(9)|BIT(8))
#define PLUM2_PCC2_INT_CLEAR		(BIT(3)|BIT(2))
#define PLUM2_PCC2_INT_ENABLE		(BIT(3)|BIT(2))
#define PLUM2_PCC2_INT_INPUT_ENABLE	(BIT(11)|BIT(10))

/* USB */
#define PLUM2_USB_INTENABLE	0x8310

#define Plum2_UsbIntEnable	PLUM2_REG_AT(PLUM2_USB_INTENABLE)

#define PLUM2_USB_ENABLE_MASK	BIT(1)
#define PLUM2_USBW_ENABLE_MASK	BIT(0)

/* Smart Media */
#define PLUM2_SM_INTENABLE	0x8410

#define Plum2_SmIntEnable	PLUM2_REG_AT(PLUM2_SM_INTENABLE)

#define PLUM2_SM_ENABLE_MASK	BIT(0)


#endif /* __ASM_MIPS_PLUM2_H */
