#ifndef __ASM_MIPS_TX3920_H__
#define __ASM_MIPS_TX3920_H__

#include <asm/addrspace.h>
 
/*
 *	TX3920 Processor core system address map
 */

#define TX39_SYSADDR_DRAM_BANK00	0x00000000
#define TX39_SYSADDR_DRAM_BANK10	0x02000000
#define TX39_SYSADDR_DRAM_BANK01	0x04000000
#define TX39_SYSADDR_DRAM_BANK11	0x06000000
#define TX39_SYSADDR_DRAM_BANK_SIZE	0x02000000

#define TX39_SYSADDR_CARD1		0x08000000
#define TX39_SYSADDR_CARD2		0x0c000000
#define TX39_SYSADDR_CARD_SIZE		0x04000000

#define TX39_SYSADDR_CS1		0x10000000
#define TX39_SYSADDR_CS2		0x10400000
#define TX39_SYSADDR_CS3		0x10800000
#define TX39_SYSADDR_CS_SIZE		0x00400000

#define TX39_SYSADDR_FUNCTION_REG	0x10c00000

#define TX39_SYSADDR_CS0		0x11000000
#define TX39_SYSADDR_CS0_SIZE		0x0f000000

#define TX39_SYSADDR_SDRAM_BANK0_MODE		0x10e00000
#define TX39_SYSADDR_SDRAM_BANK1_MODE		0x10f00000
#define TX39_SYSADDR_SDRAM_BANK_MODE_SIZE	0x00100000

#define TX39_SYSADDR_CARD_MEM1		0x64000000
#define TX39_SYSADDR_CARD_MEM2		0x68000000
#define TX39_SYSADDR_CARD_MEM_SIZE	0x04000000

#define TX39_SYSADDR_MCS0		0x6c000000
#define TX39_SYSADDR_MCS1		0x70000000
#define TX39_SYSADDR_MCS_SIZE		0x04000000

#define TX39_USERADDR_CS1		0x58000000
#define TX39_USERADDR_CS2		0x5c000000
#define TX39_USERADDR_CS3		0x60000000
#define TX39_USERADDR_CS_SIZE		0x04000000

/*
 *	General macro definitions
 */

#define REGISTER_BASE		(KSEG1 + TX39_SYSADDR_FUNCTION_REG)

#ifndef _LANGUAGE_ASSEMBLY
#define REG_AT(x)	(*((volatile u32 *)(REGISTER_BASE + x)))
#else
#define REG_AT(x)	(REGISTER_BASE + x)
#endif

#define BIT(x)	(1 << (x))


/******************************************************************************
*
*	BIU Module
*
******************************************************************************/

/* BIU Registers */
#define MemConfig0		REG_AT(0x000)
#define MemConfig1		REG_AT(0x004)
#define MemConfig2		REG_AT(0x008)
#define MemConfig3		REG_AT(0x00c)
#define MemConfig4		REG_AT(0x010)
#define MemConfig5		REG_AT(0x014)
#define MemConfig6		REG_AT(0x018)
#define MemConfig7		REG_AT(0x01c)
#define MemConfig8		REG_AT(0x020)

/* Memory config register 0 */
#define MEM0_ENDCLKOUTTRI	BIT(30)
#define MEM0_DISDQMINIT		BIT(29)
#define MEM0_ENSDRAMRD		BIT(28)
#define MEM0_SHOWDINO		BIT(27)
#define MEM0_ENRMAP2		BIT(26)
#define MEM0_ENRMAP1		BIT(25)
#define MEM0_ENWRINPAGE		BIT(24)
#define MEM0_ENCS3USER		BIT(23)
#define MEM0_ENCS2USER		BIT(22)
#define MEM0_ENCS1USER		BIT(21)
#define MEM0_ENCS1DRAM		BIT(20)
#define MEM0_BANK1CONF		(BIT(19) | BIT(18))
#define MEM0_BANK0CONF		(BIT(17) | BIT(16))
#define MEM0_ROWSEL1		(BIT(15) | BIT(14))
#define MEM0_ROWSEL0		(BIT(13) | BIT(12))
#define MEM0_COLSEL1		(BIT(11) | BIT(10) | BIT(9) | BIT(8))
#define MEM0_COLSEL0		(BIT(7) | BIT(6) | BIT(5) | BIT(4))
#define MEM0_CS3SIZE		BIT(3)
#define MEM0_CS2SIZE		BIT(2)
#define MEM0_CS1SIZE		BIT(1)
#define MEM0_CS0SIZE		BIT(0)

#define MEM0_BANK1CONF_SHIFT	18
#define MEM0_BANK0CONF_SHIFT	16
#define MEM0_ROWSEL1_SHIFT	14
#define MEM0_ROWSEL0_SHIFT	12
#define MEM0_COLSEL1_SHIFT	8
#define MEM0_COLSEL0_SHIFT	4

/* Memory config register 1 */
#define MEM1_C48MPLLON		BIT(30)
#define MEM1_ENMCS1BE		BIT(29)
#define MEM1_ENMCS0BE		BIT(28)
#define MEM1_ENMCS1ACC		BIT(27)
#define MEM1_ENMCS0ACC		BIT(26)
#define MEM1_BCLKDIV		(BIT(25) | BIT(24) | BIT(23))
#define MEM1_ENBCLK		BIT(22)
#define MEM1_ENMCS1PAGE		BIT(21)
#define MEM1_ENMCS0PAGE		BIT(20)
#define MEM1_ENMCS1WAIT		BIT(19)
#define MEM1_ENMCS0WAIT		BIT(18)
#define MEM1_MCS1_32		BIT(17)
#define MEM1_MCS0_32		BIT(16)
#define MEM1_MCS1ACCVAL1	(BIT(15) | BIT(14) | BIT(13) | BIT(12))
#define MEM1_MCS1ACCVAL2	(BIT(11) | BIT(10) | BIT(9) | BIT(8))
#define MEM1_MCS0ACCVAL1	(BIT(7) | BIT(6) | BIT(5) | BIT(4))
#define MEM1_MCS0ACCVAL2	(BIT(3) | BIT(2) | BIT(1) | BIT(0))

/* Memory config register 2 */
#define MEM2_CS3ACCVAL1		(BIT(31) | BIT(30) | BIT(29) | BIT(28))
#define MEM2_CS3ACCVAL2		(BIT(27) | BIT(26) | BIT(25) | BIT(24))
#define MEM2_CS2ACCVAL1		(BIT(23) | BIT(22) | BIT(21) | BIT(20))
#define MEM2_CS2ACCVAL2		(BIT(19) | BIT(18) | BIT(17) | BIT(16))
#define MEM2_CS1ACCVAL1		(BIT(15) | BIT(14) | BIT(13) | BIT(12))
#define MEM2_CS1ACCVAL2		(BIT(11) | BIT(10) | BIT(9) | BIT(8))
#define MEM2_CS0ACCVAL1		(BIT(7) | BIT(6) | BIT(5) | BIT(4))
#define MEM2_CS0ACCVAL2		(BIT(3) | BIT(2) | BIT(1) | BIT(0))

/* Memory config register 3 */
#define MEM3_CARD2ACCVAL	(BIT(31) | BIT(30) | BIT(29) | BIT(28))
#define MEM3_CARD1ACCVAL	(BIT(27) | BIT(26) | BIT(25) | BIT(24))
#define MEM3_CARD2IOACCVAL	(BIT(23) | BIT(22) | BIT(21) | BIT(20))
#define MEM3_CARD1IOACCVAL	(BIT(19) | BIT(18) | BIT(17) | BIT(16))
#define MEM3_ENCS3PAGE		BIT(11)
#define MEM3_ENCS2PAGE		BIT(10)
#define MEM3_ENCS1PAGE		BIT(9)
#define MEM3_ENCS0PAGE		BIT(8)
#define MEM3_CARD2WAITEN	BIT(7)
#define MEM3_CARD1WAITEN	BIT(6)
#define MEM3_CARD2IOEN		BIT(5)
#define MEM3_CARD1IOEN		BIT(4)
#define MEM3_CARD2_8SEL		BIT(3)
#define MEM3_CARD1_8SEL		BIT(2)

/* Memory config register 4 */
#define MEM4_ENBANK1EDO		BIT(31)
#define MEM4_ENBANK0EDO		BIT(30)
#define MEM4_ENARB		BIT(29)
#define MEM4_DISSNOOP		BIT(28)
#define MEM4_CLRWRBUSERRINT	BIT(27)
#define MEM4_ENBANK1OPT		BIT(26)
#define MEM4_ENBANK0OPT		BIT(25)
#define MEM4_ENWATCH		BIT(24)
#define MEM4_WATCHTIMEVAL	(BIT(23) | BIT(22) | BIT(21) | BIT(20))
#define MEM4_MEMPOWERDOWN	BIT(16)
#define MEM4_ENRFSH1		BIT(15)
#define MEM4_ENRFSH0		BIT(14)
#define MEM4_RFSHVAL1		(BIT(13) | BIT(12) | BIT(11) | BIT(10) | BIT(9) | BIT(8))
#define MEM4_RFSHVAL0		(BIT(5) | BIT(4) | BIT(3) | BIT(2) | BIT(1) | BIT(0))

/* Memory config register 5 */
#define MEM5_STARTVAL2		0xfffffe00
#define MEM5_MASK2		(BIT(3) | BIT(2) | BIT(1) | BIT(0) | 

/* Memory config register 6 */
#define MEM6_STARTVAL1		0xfffffe00
#define MEM6_MASK1		(BIT(3) | BIT(2) | BIT(1) | BIT(0) | 

/* Memory config register 7 */
#define MEM7_RMAPADD2		0xfffffe00

/* Memory config register 8 */
#define MEM8_RMAPADD1		0xfffffe00


/******************************************************************************
*
*	Clock Module
*
******************************************************************************/

/* Clock registers */
#define ClockControl		REG_AT(0x1c0)

/* Clock Control Register */
#define CLK_CHICLKDIV		0xff000000
#define CLK_CHIMCLKSEL		BIT(21)
#define CLK_CHICLKDIR		BIT(20)
#define CLK_ENCHIMCLK		BIT(19)
#define CLK_ENVIDCLK		BIT(18)
#define CLK_ENSPICLK		BIT(16)
#define CLK_ENTIMERCLK		BIT(15)
#define CLK_SIBMCLKDIR		BIT(13)
#define CLK_ENSIBMCLK		BIT(11)
#define CLK_SIBMCLKDIV_MASK	(BIT(10) | BIT(9) | BIT(8))
#define CLK_CSERSEL		BIT(7)
#define CLK_CSERDIV_MASK	(BIT(6) | BIT(5) | BIT(4))
#define CLK_ENCSERCLK		BIT(3)
#define CLK_ENIRCLK		BIT(2)
#define CLK_EN_UART_A		BIT(1)
#define CLK_EN_UART_B		BIT(0)

#define CLK_CHICLKDIV_SHIFT	24
#define CLK_SIBMCLKDIV_SHIFT	8
#define CLK_CSERDIV_SHIFT	4


/******************************************************************************
*
*	Interrupt Module
*
******************************************************************************/

#define TX39INT_REGS		8

#define INT1_REG		0
#define INT2_REG		1
#define INT3_REG		2
#define INT4_REG		3
#define INT5_REG		4
#define INT6_REG		5
#define INT7_REG		6
#define INT8_REG		7

/* Interrupt Status Registers */
#define IntStatus1		REG_AT(0x100)
#define IntStatus2		REG_AT(0x104)
#define IntStatus3		REG_AT(0x108)
#define IntStatus4		REG_AT(0x10c)
#define IntStatus5		REG_AT(0x110)
#define IntStatus6		REG_AT(0x114)
#define IntStatus7		REG_AT(0x130)
#define IntStatus8		REG_AT(0x138)

/* Interrupt Clear Registers */
#define IntClear1		REG_AT(0x100)
#define IntClear2		REG_AT(0x104)
#define IntClear3		REG_AT(0x108)
#define IntClear4		REG_AT(0x10c)
#define IntClear5		REG_AT(0x110)
#define IntClear7		REG_AT(0x130)
#define IntClear8		REG_AT(0x138)

/* Interrupt Enable Registers */
#define IntEnable1		REG_AT(0x118)
#define IntEnable2		REG_AT(0x11c)
#define IntEnable3		REG_AT(0x120)
#define IntEnable4		REG_AT(0x124)
#define IntEnable5		REG_AT(0x128)
#define IntEnable6		REG_AT(0x12c)
#define IntEnable7		REG_AT(0x134)
#define IntEnable8		REG_AT(0x13c)

/* Interrupt Status Register 1 */
#define INT1_CHI0_5INT		BIT(29)
#define INT1_CHI1_0INT		BIT(28)
#define INT1_CHIDMACNTINT	BIT(27)
#define INT1_CHININTA		BIT(26)
#define INT1_CHININTB		BIT(25)
#define INT1_CHIACTINT		BIT(24)
#define INT1_CHIERRINT		BIT(23)
#define INT1_SND0_5INT		BIT(22)
#define INT1_SND1_0INT		BIT(21)
#define INT1_TEL0_5INT		BIT(20)
#define INT1_TEL1_0INT		BIT(19)
#define INT1_SNDDMACNTINT	BIT(18)
#define INT1_TELDMACNTINT	BIT(17)
#define INT1_LSNDCLIPINT	BIT(16)
#define INT1_RSNDCLIPINT	BIT(15)
#define INT1_VALSNDPOSINT	BIT(14)
#define INT1_VALSNDNEGINT	BIT(13)
#define INT1_VALTELPOSINT	BIT(12)
#define INT1_VALTELNEGINT	BIT(11)
#define INT1_SNDININT		BIT(10)
#define INT1_TELININT		BIT(9)
#define INT1_SIBSF0INT		BIT(8)
#define INT1_SIBSF1INT		BIT(7)
#define INT1_SIBIRQPOSINT	BIT(6)
#define INT1_SIBIRQNEGINT	BIT(5)

/* Interrupt Status Register 2 */
#define INT2_UARTARXINT		BIT(31)
#define INT2_UARTARXOVERRUNINT	BIT(30)
#define INT2_UARTAFRAMEERRINT	BIT(29)
#define INT2_UARTABREAKINT	BIT(28)
#define INT2_UARTAPARITYERRINT	BIT(27)
#define INT2_UARTATXINT		BIT(26)
#define INT2_UARTATXOVERRUNINT	BIT(25)
#define INT2_UARTAEMPTYINT	BIT(24)
#define INT2_UARTADMAFULLINT	BIT(23)
#define INT2_UARTADMAHALFINT	BIT(22)

#define INT2_UARTBRXINT		BIT(21)
#define INT2_UARTBRXOVERRUNINT	BIT(20)
#define INT2_UARTBFRAMEERRINT	BIT(29)
#define INT2_UARTBBREAKINT	BIT(18)
#define INT2_UARTBPARITYERRINT	BIT(17)
#define INT2_UARTBTXINT		BIT(16)
#define INT2_UARTBTXOVERRUNINT	BIT(15)
#define INT2_UARTBEMPTYINT	BIT(14)
#define INT2_UARTBDMAFULLINT	BIT(13)
#define INT2_UARTBDMAHALFINT	BIT(12)

#define INT2_UARTRXINT		(BIT(31) | BIT(21))
#define INT2_UARTRXOVERRUNINT	(BIT(30) | BIT(20))
#define INT2_UARTFRAMEERRINT	(BIT(29) | BIT(19))
#define INT2_UARTBREAKINT	(BIT(28) | BIT(18))
#define INT2_UARTPARITYERRINT	(BIT(27) | BIT(17))
#define INT2_UARTTXINT		(BIT(26) | BIT(16))
#define INT2_UARTTXOVERRUNINT	(BIT(25) | BIT(15))
#define INT2_UARTEMPTYINT	(BIT(24) | BIT(14))
#define INT2_UARTDMAFULLINT	(BIT(23) | BIT(13))
#define INT2_UARTDMAHALFINT	(BIT(22) | BIT(12))

#define INT2_UARTA_RXINT	(BIT(31) | BIT(30) | BIT(29) | BIT(28) | BIT(27))
#define INT2_UARTA_TXINT	(BIT(26) | BIT(25) | BIT(24))
#define INT2_UARTA_DMAINT	(BIT(23) | BIT(22))

#define INT2_UARTB_RXINT	(BIT(21) | BIT(20) | BIT(19) | BIT(18) | BIT(17))
#define INT2_UARTB_TXINT	(BIT(16) | BIT(15) | BIT(14))
#define INT2_UARTB_DMAINT	(BIT(13) | BIT(12))

#define INT2_UARTA_RDI		BIT(31)
#define INT2_UARTA_RLSI		(BIT(30) | BIT(29) | BIT(28) | BIT(27))
#define INT2_UARTA_THRI		(BIT(26) | BIT(25) | BIT(24))

#define INT2_UARTB_RDI		BIT(21)
#define INT2_UARTB_RLSI		(BIT(20) | BIT(19) | BIT(18) | BIT(17))
#define INT2_UARTB_THRI		(BIT(16) | BIT(15) | BIT(14))

#define INT2_UARTA_INT		(INT2_UARTA_RXINT | INT2_UARTA_TXINT | \
				 INT2_UARTA_DMAINT)
#define INT2_UARTB_INT		(INT2_UARTB_RXINT | INT2_UARTB_TXINT | \
				 INT2_UARTB_DMAINT)

#define INT2_UART_RDI(x)	((x == 0) ? INT2_UARTA_RDI : INT2_UARTB_RDI)
#define INT2_UART_RLSI(x)	((x == 0) ? INT2_UARTA_RLSI : INT2_UARTB_RLSI)
#define INT2_UART_THRI(x)	((x == 0) ? INT2_UARTA_THRI : INT2_UARTB_THRI)

#define INT2_UART_RXINT(x)	((x == 0) ? INT2_UARTA_RXINT : INT2_UARTB_RXINT)
#define INT2_UART_TXINT(x)	((x == 0) ? INT2_UARTA_TXINT : INT2_UARTB_TXINT)
#define INT2_UART_INT(x)	((x == 0) ? INT2_UARTA_INT : INT2_UARTB_INT)

#define INT2_UART_OFFSET(x)	((x == 0) ? 22 : 12)

/* Interrupt Enable Register 3 */
#define INT3_MFIOPOSINT(n)	BIT(n)

/* Interrupt Enable Register 4 */
#define INT4_MFIONEGINT(n)	BIT(n)

/* Interrupt Status Register 5 */
#define INT5_RTCINT		BIT(31)
#define INT5_ALARMINT		BIT(30)
#define INT5_PERINT		BIT(29)
#define INT5_STPTIMERINT	BIT(28)
#define INT5_POSPWRINT		BIT(27)
#define INT5_NEGPWRINT		BIT(26)
#define INT5_POSPWROKINT	BIT(25)
#define INT5_NEGPWROKINT	BIT(24)
#define INT5_POSONBUTNINT	BIT(23)
#define INT5_NEGONBUTNINT	BIT(22)
#define INT5_SPIBUFAVAILINT	BIT(21)
#define INT5_SPIERRINT		BIT(20)
#define INT5_SPIRCVINT		BIT(19)
#define INT5_SPIEMPTYINT	BIT(18)
#define INT5_IRCONSMINT		BIT(17)
#define INT5_CARSTINT		BIT(16)
#define INT5_POSCARINT		BIT(15)
#define INT5_NEGCARINT		BIT(14)

#define INT5_PWROKINT		(INT5_POSPWROKINT | INT5_NEGPWROKINT)

#define INT5_NEGINT_SHIFT 0
#define INT5_POSINT_SHIFT 7

/* Interrupt Status Register 6 */
#define INT6_IRQHIGH		BIT(31)
#define INT6_IRQLOW		BIT(30)
#define INT6_INTVECT		(BIT(5) | BIT(4) | BIT(3) | BIT(2))

#define INTVECT_SHIFT		2
#define INTVECT(x)		((x & INT6_INTVECT) >> INTVECT_SHIFT)

/* Interrupt Enable Register 6 */
#define INT6_GLOBALEN		BIT(18)
#define INT6_PWROKINT		BIT(15)
#define INT6_ALARMINT		BIT(14)
#define INT6_PERINT	 	BIT(13)
#define INT6_UARTARXINT		BIT(12)
#define INT6_UARTBRXINT		BIT(12)
#define INT6_MFIOPOSINT		BIT(11)
#define INT6_IOPOSINT56		BIT(10)
#define INT6_MFIONEGINT		BIT(9)
#define INT6_IONEGINT56		BIT(8)


/* Interrupt Enable Register 8 */
#define INT8_IONEGINT0		BIT(0)
#define INT8_IONEGINT1		BIT(1)
#define INT8_IONEGINT2		BIT(2)
#define INT8_IONEGINT3		BIT(3)
#define INT8_IONEGINT4		BIT(4)
#define INT8_IONEGINT5		BIT(5)
#define INT8_IONEGINT6		BIT(6)
#define INT8_IONEGINT7		BIT(7)
#define INT8_IONEGINT8		BIT(8)
#define INT8_IONEGINT9		BIT(9)
#define INT8_IONEGINT10		BIT(10)
#define INT8_IONEGINT11		BIT(11)
#define INT8_IONEGINT12		BIT(12)
#define INT8_IONEGINT13		BIT(13)
#define INT8_IONEGINT14		BIT(14)
#define INT8_IONEGINT15		BIT(15)
#define INT8_IOPOSINT0		BIT(16)
#define INT8_IOPOSINT1		BIT(17)
#define INT8_IOPOSINT2		BIT(18)
#define INT8_IOPOSINT3		BIT(19)
#define INT8_IOPOSINT4		BIT(20)
#define INT8_IOPOSINT5		BIT(21)
#define INT8_IOPOSINT6		BIT(22)
#define INT8_IOPOSINT7		BIT(23)
#define INT8_IOPOSINT8		BIT(24)
#define INT8_IOPOSINT9		BIT(25)
#define INT8_IOPOSINT10		BIT(26)
#define INT8_IOPOSINT11		BIT(27)
#define INT8_IOPOSINT12		BIT(28)
#define INT8_IOPOSINT13		BIT(29)
#define INT8_IOPOSINT14		BIT(30)
#define INT8_IOPOSINT15		BIT(31)


/******************************************************************************
*
*	IO Module
*
******************************************************************************/

/* IO registers */
#define IOControl		REG_AT(0x180)
#define IODataInOut		REG_AT(0x19c)
#define MFIOOutput		REG_AT(0x184)
#define MFIODirection		REG_AT(0x188)
#define MFIOInput		REG_AT(0x18c)
#define MFIOSelect		REG_AT(0x190)
#define IOPowerDown		REG_AT(0x194)
#define MFIOPowerDown		REG_AT(0x198)


/******************************************************************************
*
*	Power Module
*
******************************************************************************/

/* Power registers */
#define PowerControl		REG_AT(0x1C4)

/* Power Control Register */
#define PWR_ONBUTN		BIT(31)
#define PWR_PWRINT		BIT(30)
#define PWR_PWROK		BIT(29)
#define PWR_PWROKNMI		BIT(28)
#define PWR_SLOWBUS		BIT(26)
#define PWR_STPTIMERVAL		(BIT(15) | BIT(14) | BIT(13) | BIT(12))
#define PWR_ENSTPTIMER		BIT(11)
#define PWR_ENFORCESHUTDWN	BIT(10)
#define PWR_FORCESHUTDWN	BIT(9)
#define PWR_FORCESHUTDWNOCC	BIT(8)
#define PWR_SELC2MS		BIT(7)
#define PWR_WARMSTART		BIT(6)
#define PWR_BPDBVCC3		BIT(5)
#define PWR_STOPCPU		BIT(4)
#define PWR_DBNCONBUTN		BIT(3)
#define PWR_COLDSTART		BIT(2)
#define PWR_PWRCS		BIT(1)
#define PWR_VCCON		BIT(0)

#define PWR_STPTIMERVAL_SHIFT	12


/******************************************************************************
*
*	Timer Module
*
******************************************************************************/

/* Timer registers */
#define RTChigh			REG_AT(0x140)
#define RTClow			REG_AT(0x144)
#define AlarmHigh		REG_AT(0x148)
#define AlarmLow		REG_AT(0x14c)
#define TimerControl		REG_AT(0x150)
#define PeriodTimer		REG_AT(0x154)

/* RTC Register */
#define RTC_HIGH_MASK		0x000007ff
#define RTC_PERSEC_COUNT	32768		/* RTC = 32.768 kHz */

#define tx3920_read_rtc(high,low)				\
do{u32 __temp_low__, __temp_high__;				\
    low  = RTClow;					\
    high = RTChigh & RTC_HIGH_MASK;				\
    do{								\
	__temp_low__  = low;			\
	__temp_high__ = high;					\
	low  = RTClow;						\
	high = RTChigh & RTC_HIGH_MASK;				\
    }while( low != __temp_low__ || high != __temp_high__ );	\
}while(0)

/* Timer Control Register */
#define TIM_FREEZEPRE		BIT(7)
#define TIM_FREEZERTC		BIT(6)
#define TIM_FREEZETIMER		BIT(5)
#define TIM_ENPERTIMER		BIT(4)
#define TIM_RTCCLR		BIT(3)
#define TIM_ENRTCTST		BIT(0)

/* Periodic Timer Register */
#ifdef CONFIG_SHARP_TELIOS
#define PER_TIMER_COUNT		(129024000 / 4 / 32 / HZ - 1)
				/* CPU core clock = 129.024MHz (maybe) */
#else
#define PER_TIMER_COUNT		11520

/*
 * For a system clock frequency of 36.864MHz, the timer counts at one tick
 * every 868nS (ie CLK/32). Therefore 1152 counts gives a 1mS interval
 */
#endif

/******************************************************************************
*
*	CHI Module
*
******************************************************************************/

/* Register locations */

#define CHIControl             REG_AT(0x1d8)
#define CHIPointerEnable       REG_AT(0x1dc)
#define CHIRecivePointerA      REG_AT(0x1e0)
#define CHIRecivePointerB      REG_AT(0x1e4)
#define CHITransmitPointerA    REG_AT(0x1e8)
#define CHITransmitPointerB    REG_AT(0x1ec)
#define CHISize                REG_AT(0x1f0)
#define CHIRXStart             REG_AT(0x1f4)
#define CHITXStart             REG_AT(0x1f8)
#define CHIRXHolding           REG_AT(0x1fc)
#define CHITXHolding           REG_AT(0x1fc)

/* CHI Contorl Register */

#define CHI_CHILOOP               BIT(29)
#define CHI_CHIENTEST             BIT(28)
#define CHI_CHIFSDIR              BIT(27)
#define CHI_CHIFSWIDTH_MASK       (BIT(26) | BIT(25))
#define CHI_CHIFSWIDTH_SHIFT      25
#define CHI_CHINCHAN_MASK         0x01f00000
#define CHI_CHINCHAN_SHIFT        20
#define CHI_CHITXBOFF_MASK        0x000f0000
#define CHI_CHITXBOFF_SHIFT       16
#define CHI_CHIRXBOFF_MASK        0x0000f000
#define CHI_CHIRXBOFF_SHIFT       12
#define CHI_TXMSBFIRST            BIT(11)
#define CHI_RXMSBFIRST            BIT(10)
#define CHI_CHIRXFSPOL            BIT(9)
#define CHI_CHITXFSPOL            BIT(8)
#define CHI_CHIRXEDGE             BIT(7)
#define CHI_CHITXEDGE             BIT(6)
#define CHI_CHIFSEDGE             BIT(5)
#define CHI_CHITXFSEDGE           BIT(4)
#define CHI_CHICLK2XMODE          BIT(3)
#define CHI_CHIRXEN               BIT(2)
#define CHI_CHITXEN               BIT(1)
#define CHI_ENCHI                 BIT(0)

/* CHI Pointer Enable Register */

#define CHI_CHITXPTRB3EN          BIT(31)
#define CHI_CHITXPTRB2EN          BIT(30)
#define CHI_CHITXPTRB1EN          BIT(29)
#define CHI_CHITXPTRB0EN          BIT(28)
#define CHI_CHITXPTRA3EN          BIT(27)
#define CHI_CHITXPTRA2EN          BIT(26)
#define CHI_CHITXPTRA1EN          BIT(25)
#define CHI_CHITXPTRA0EN          BIT(24)
#define CHI_CHIRXPTRB3EN          BIT(23)
#define CHI_CHIRXPTRB2EN          BIT(22)
#define CHI_CHIRXPTRB1EN          BIT(21)
#define CHI_CHIRXPTRB0EN          BIT(20)
#define CHI_CHIRXPTRA3EN          BIT(19)
#define CHI_CHIRXPTRA2EN          BIT(18)
#define CHI_CHIRXPTRA1EN          BIT(17)
#define CHI_CHIRXPTRA0EN          BIT(16)


/* CHI Receive Pointer A Register */

#define CHI_CHIRXPTRA3_MASK       0x0f000000
#define CHI_CHIRXPTRA3_SHIFT      24
#define CHI_CHIRXPTRA2_MASK       0x000f0000
#define CHI_CHIRXPTRA2_SHIFT      16
#define CHI_CHIRXPTRA1_MASK       0x00000f00
#define CHI_CHIRXPTRA1_SHIFT      8
#define CHI_CHIRXPTRA0_MASK       0x0000000f
#define CHI_CHIRXPTRA0_SHIFT      0

/* CHI Receive Pointer A Register */

#define CHI_CHIRXPTRB3_MASK       0x0f000000
#define CHI_CHIRXPTRB3_SHIFT      24
#define CHI_CHIRXPTRB2_MASK       0x000f0000
#define CHI_CHIRXPTRB2_SHIFT      16
#define CHI_CHIRXPTRB1_MASK       0x00000f00
#define CHI_CHIRXPTRB1_SHIFT      8
#define CHI_CHIRXPTRB0_MASK       0x0000000f
#define CHI_CHIRXPTRB0_SHIFT      0

/* CHI Transmit Pointer A Register */

#define CHI_CHITXPTRA3_MASK       0x0f000000
#define CHI_CHITXPTRA3_SHIFT      24
#define CHI_CHITXPTRA2_MASK       0x000f0000
#define CHI_CHITXPTRA2_SHIFT      16
#define CHI_CHITXPTRA1_MASK       0x00000f00
#define CHI_CHITXPTRA1_SHIFT      8
#define CHI_CHITXPTRA0_MASK       0x0000000f
#define CHI_CHITXPTRA0_SHIFT      0

/* CHI Transmit Pointer A Register */

#define CHI_CHIDMAPTR_MASK        0x03fc0000
#define CHI_CHIDMAPTR_SHIFT       18
#define CHI_CHIBUFF1TIME          BIT(15)
#define CHI_CHIDMALOOP            BIT(14)
#define CHI_CHISIZE_MASK          0x000003fc
#define CHI_CHISIZE_SHIFT         2
#define CHI_ENDMARXCHI            BIT(1)
#define CHI_ENDMATXCHI            BIT(0)

/* CHI RX Start Register */

#define CHI_CHIRXSTART_MASK       0xfffffffc
#define CHI_CHIRXSTART_SHIFT      2

/* CHI TX Start Register */

#define CHI_CHITXSTART_MASK       0xfffffffc
#define CHI_CHITXSTART_SHIFT      2


/******************************************************************************
*
*      SIB (Serial Interconnect Bus) Module
*
******************************************************************************/

/* Register locations */

#define SIBSize	      	        REG_AT(0x060)
#define SIBSoundRXStart	      	REG_AT(0x064)
#define SIBSoundTXStart         REG_AT(0x068)
#define SIBTelecomRXStart       REG_AT(0x06C)
#define SIBTelecomTXStart       REG_AT(0x070)
#define SIBControl              REG_AT(0x074)
#define SIBSoundTXRXHolding     REG_AT(0x078)
#define SIBTelecomTXRXHolding   REG_AT(0x07C)
#define SIBSubFrame0Control     REG_AT(0x080)
#define SIBSubFrame1Control     REG_AT(0x084)
#define SIBSubFrame0Status      REG_AT(0x088)
#define SIBSubFrame1Status      REG_AT(0x08C)
#define SIBDMAControl           REG_AT(0x090)

/* SIB Size Register */

#define SIB_SNDSIZE_MASK        0x3ffc0000
#define SIB_SNDSIZE_SHIFT       18
#define SIB_TELSIZE_MASK        0x00003ffc
#define SIB_TELSIZE_SHIFT       2

/* SIB Control Register */

#define SIB_SIBIRQ              BIT(31)
#define SIB_ENCNTTEST           BIT(30)
#define SIB_ENDMATEST           BIT(29)
#define SIB_SNDMONO             BIT(28)
#define SIB_RMONOSNDIN          BIT(27)
#define SIB_SIBSCLKDIV_MASK     (BIT(26) | BIT(25) | BIT(24))
#define SIB_SIBSCLKDIV_SHIFT    24
#define SIB_TEL16               BIT(23)
#define SIB_TELFSDIV_MASK       0x007f0000
#define SIB_TELFSDIV_SHIFT      16
#define SIB_SND16               BIT(15)
#define SIB_SNDFSDIV_MASK       0x00007f00
#define SIB_SNDFSDIV_SHIFT      8
#define SIB_SELTELSF1           BIT(7)
#define SIB_SELSNDSF1           BIT(6)
#define SIB_ENTEL               BIT(5)
#define SIB_ENSND               BIT(4)
#define SIB_SIBLOOP             BIT(3)
#define SIB_ENSF1               BIT(2)
#define SIB_ENSF0               BIT(1)
#define SIB_ENSIB               BIT(0)

/* SIB Frame Format (SIBSubFrame0Status and SIBSubFrame1Status) */

#define SIB_REGISTER_EXT        BIT(31)  /* Must be zero */
#define SIB_ADDRESS_MASK        0x78000000
#define SIB_ADDRESS_SHIFT       27
#define SIB_WRITE               BIT(26)
#define SIB_AUD_VALID           BIT(17)
#define SIB_TEL_VALID           BIT(16)
#define SIB_DATA_MASK           0x00ff
#define SIB_DATA_SHIFT          0

/* SIB DMA Control Register */

#define SIB_SNDBUFF1TIME        BIT(31)
#define SIB_SNDDMALOOP          BIT(30)
#define SIB_SNDDMAPTR_MASK      0x3ffc0000
#define SIB_SNDDMAPTR_SHIFT     18
#define SIB_ENDMARXSND          BIT(17)
#define SIB_ENDMATXSND          BIT(16)
#define SIB_TELBUFF1TIME        BIT(15)
#define SIB_TELDMALOOP          BIT(14)
#define SIB_TELDMAPTR_MASK      0x00003ffc
#define SIB_TELDMAPTR_SHIFT     2
#define SIB_ENDMARXTEL          BIT(1)
#define SIB_ENDMATXTEL          BIT(0)

/******************************************************************************
*
*      SPI Module
*
******************************************************************************/

/* Register locations */

#define SPIControl     	        REG_AT(0x160)
#define SPITransmitHolding      REG_AT(0x164)
#define SPIReceiveHolding       REG_AT(0x164)

/* SIB Control Register */

#define SPI_SPION		BIT(17)
#define SPI_EMPTY		BIT(16)
#define SPI_DELAYVAL_MASK       0x0000f000
#define SPI_DELAYVAL_SHIFT      12
#define SPI_BAUDRATE_MASK       0x00000f00
#define SPI_BAUDRATE_SHIFT      8
#define SPI_PHAPOL		BIT(5)
#define SPI_CLKPOL		BIT(4)
#define SPI_WORD		BIT(2)
#define SPI_LSB			BIT(1)
#define SPI_ENSPI		BIT(0)

#endif /* __TX3920_H__ */
