/*
 * kernel/axmon.h
 *
 * Copyright (C) 2001 AXE, Inc.
 */

#ifndef _AXMON_H
#define _AXMON_H
#ifdef CONFIG_AXE_MONITOR
#ifdef CONFIG_AXE_MONITOR_SCHED_LOG

#include <linux/sched.h>

#define AXMON_TICK_RANGE_MAX 15
#define AXMON_HIST_NUM 16
#define AXMON_PERIOD_SEC 10
#define AXMON_PERIOD (HZ * (AXMON_PERIOD_SEC))

struct axmon_pf_hist{
    unsigned long tick_count;
    unsigned long num[AXMON_TICK_RANGE_MAX+1];
};

extern struct axmon_pf_hist axmon_pf_hist[AXMON_HIST_NUM];
extern struct axmon_pf_hist axmon_pf_hist_snap[AXMON_HIST_NUM];
extern int axmon_pf_hist_ptr;
extern unsigned long axmon_pf_last_rotate;

extern unsigned long axmon_pf_counter_max, axmon_pf_counter_pid;
extern int axmon_misc_enable;

static inline void axmon_pf_rotate(unsigned long tick_count)
{
    int i;
    axmon_pf_hist[axmon_pf_hist_ptr].tick_count = tick_count;
    axmon_pf_hist_ptr = (axmon_pf_hist_ptr + 1) % AXMON_HIST_NUM;
    for( i=0; i<=AXMON_TICK_RANGE_MAX; i++ )
	axmon_pf_hist[axmon_pf_hist_ptr].num[i] = 0;
}

static inline void axmon_pf_set(int ocupied_tick)
{
    if( ocupied_tick < AXMON_TICK_RANGE_MAX )
	axmon_pf_hist[axmon_pf_hist_ptr].num[ocupied_tick] ++;
    else
	axmon_pf_hist[axmon_pf_hist_ptr].num[AXMON_TICK_RANGE_MAX] ++;
}

void axmon_pf_snap(void);
void axmon_pf_print(void);

/* ******************************************************************** */

#define AXMON_TRACE_EV_TO_READY    0
#define AXMON_TRACE_EV_TO_RUN      1
#define AXMON_TRACE_EV_RUN2READY   2
#define AXMON_TRACE_EV_RUN2WAIT    3
#define AXMON_TRACE_EV_TO_WAIT     4

void axmon_set_trace(struct task_struct *tgt, int ev);
void axmon_check_trace(struct task_struct *prev, struct task_struct *next);

/* ******************************************************************** */
void axmon_pswitch(struct task_struct *prev, struct task_struct *next);

#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */
#endif /* CONFIG_AXE_MONITOR */
#endif /* _AXMON_H */
